"use strict";

#feature-id    Utilities > FITSFileManager

#feature-info Copy and move files based on FITS keys.<br/>\
   Written by Jean-Marc Lugrin (c) 2012,2013.

// ==================================================================================================
// The complete source code is hosted at https://bitbucket.org/bitli/fitsfilemanager
// with test scripts
// ==================================================================================================

// Copyright (c) 2012-2013 Jean-Marc Lugrin
// Copyright (c) 2003-2012 Pleiades Astrophoto S.L.
//
// Redistribution and use in both source and binary forms, with or without
// modification, is permitted provided that the following conditions are met:
//
// 1. All redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//
// 2. All redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//
// 3. Neither the names "PixInsight" and "Pleiades Astrophoto", nor the names
//    of their contributors, may be used to endorse or promote products derived
//    from this software without specific prior written permission. For written
//    permission, please contact info@pixinsight.com.
//
// 4. All products derived from this software, in any form whatsoever, must
//    reproduce the following acknowledgment in the end-user documentation
//    and/or other materials provided with the product:
//
//    "This product is based on software from the PixInsight project, developed
//    by Pleiades Astrophoto and its contributors (http://pixinsight.com/)."
//
//    Alternatively, if that is where third-party acknowledgments normally
//    appear, this acknowledgment must be reproduced in the product itself.
//
// THIS SOFTWARE IS PROVIDED BY PLEIADES ASTROPHOTO AND ITS CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL PLEIADES ASTROPHOTO OR ITS
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, BUSINESS
// INTERRUPTION; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; AND LOSS OF USE,
// DATA OR PROFITS) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
// ****************************************************************************

// Base on FITSkey_0.06 of Nikolay but completely rewritten with another approach
// Thanks to Nikolay for sharing the original FITSKey

// NOTE : The parameters have their own version number in FITSFileManger-parameters.jsh
// Version created by makeRelease on Sun Feb 10 2013 10:40:11 GMT+0100
#define VERSION "1.0"
#define TITLE     "FITSFileManager"

// --- Debugging control ----------------
// Set to false when doing hasardous developments...
#define EXECUTE_COMMANDS true

// Debug supprt is in the module PJSR-logging
// Tracing - define DEBUG if you define any other DEBUG_xxx
//#define DEBUG
//#define DEBUG_EVENTS
//#define DEBUG_SHOW_FITS
//#define DEBUG_FITS
//#define DEBUG_VARS
//#define DEBUG_COLUMNS
// ------------------------------------





// FITS keyword implementation is not yet fully OK and could not be implemented as may be part of PI 1.8
// #define IMPLEMENTS_FITS_EXPORT

// Padding format
#define FFM_COUNT_PAD 4




// Change log
// 2012-08-27 - 0.1 - Initial Version
// 2012-09-11 - 0.5 - Significant enhancements
//     Code refactoring, speedups
//     Save/restore parameters
//     Corrected mapping of files in tree and list if not sorted as loaded,
//     added refresh button because there is no onSort event,
//     default sort is ascending on FileName
//     Added button remove all
//     Added help label
//     Use SectionBar
//     Added optional indicator to accept missing values '?' and default value
//     Check for missing key values, show message
//     Source file list is refreshed after a move
//     Supressed Export FITS keys as incompletely implemented and may be integrated in 1.8
// 2012-09-26 - 0.6 - Many enhancements
//     Use TreeBox instead of TextBox
//     Added button to check/uncheck boxes
//     List types in keyword table
//     More dynamic layout
//     Added predefined templates and regexps
//     Added copy via FITS load/save with added KEYWORD
//     Show conversion definitions
// 2012-11-11 - 0.7 - Bug correction, keyword enhancements
//     Corrected bug on display of FITS keyword in image table
//     Add the input &extension if the output file has no extension
//     Added &object as a synthethic keyword
//     Added &night as an experimental keyword
//     Added alternate FITS keywords for creation of synthetic keywords
//     Load HIERARCH FITS keywords
// 2012-11-19 - 0.8 - Refactor, bug corrections, more configurability
//     Refactored FITS keyword loading (separate js file), parameters (separate file) and quite some code
//     Correcting error of selection of file in FITSKeyword window
//     Corrected error of list entry reported by Vicent
//     Document &kw:present?absent;
//     Show alternate FITS keyword in 'Remapping' section
//     Allow selection of visibility of synthetic variable in inputFile table
//     Accept FITS keywords as variables, clean FITS kewyword value to make clean file names
//     Removed the &object; variable as this can now be done with &OBJECT;
//     Added predefiend named configuration to select kewyword mappign and conversions,
//     two configurations predefined (DEFAULT and CAHA)
//     Reworked conversion rules for filters and types, support back references in right hand side.
//     Enhance unquoting of string (respect FITS standard)
//     Handle cursor and console during file move/copy operations
//     Additional tests
//     Various presentation enhancements

// 2012-12-28 - 0.9
//     Very large rewrite of configuration mecanism
// 2013-02-06 - 1.0
//     Bug corrections on selection,
//    made minimum dialog size smaller (hopefully support screens down to approx 1200 x 760)
//    added shopw/hide fullPath on inputTreeBox
//    added exit button and confirmation for remove all

// TODO
// Make cleaning of FITS keyword values used in file name configurable
// Support optional reformatting of file name as valid PI identifier
// Add mark of sequence of text to ignore if missing variable value (in parentheses for example)
// Enhance control of ordering (support for ordering of non string values if possible)
// Add a way to use directory of source file as variable  &filedir, &filedirparent for template matching and group names
// Create a log file to record the operations
// Support export of FITS keywords as CSV or tab delimite
// Request confirmation for move (or move and copy)
// Possibility to add FITS keywords to copied files (to replace erroneous values or add missing ones)
// Possibility to check FITS header for suspect values and consistency (for example with size)
// May be allow to preview or open a selected file
// Possibly alternate configuration depending on IMAGETYP
// Batch mode (non GUI)
// Normalize directory (remove .., redundant /)
// Check # of images in file (when using load image), or use direct writes to update fits headers
// Enhance 'night' with accepting other dates and possibly midnight offset
// Management of alternate source keywords for synthethic keywords, possibly default values in keyword mapping
// Make creation of HISTORY/ORIGFILE keyword optional, make name of ORIGFILE key configurable
// Support date formatting
// Allow choice of 'cleaned up' strings (make optiona lremove of special chraracters)

// See http://heasarc.gsfc.nasa.gov/docs/software/ftools/fitsverify/




//================================================================
// include "PJSR-logging.jsh"
//================================================================
// PJSR-logging.jsh

// This file is part of FITSFileManager, see copyrigh in FITSFileManager.js


// Support debug and log methods

// HTML characters for the console



// Define the loggin module
var Log = (function() {

   // --- private variables and methods for pretty print ----------------
   // From the web, made a little bit more robust
   var maxDepth = 20;

   function pp(object, depth, embedded) {

    var cx = /[\u0000\u00ad\u0600-\u0604\u070f\u17b4\u17b5\u200c-\u200f\u2028-\u202f\u2060-\u206f\ufeff\ufff0-\uffff]/g,
        escapable = /[\\\"\x00-\x1f\x7f-\x9f\u00ad\u0600-\u0604\u070f\u17b4\u17b5\u200c-\u200f\u2028-\u202f\u2060-\u206f\ufeff\ufff0-\uffff]/g,
        gap,
        indent,
        meta = {    // table of character substitutions
            '\b': '\\b',
            '\t': '\\t',
            '\n': '\\n',
            '\f': '\\f',
            '\r': '\\r',
            '"' : '\\"',
            '\\': '\\\\'
        },
        rep;


    function quote(string) {
      // If the string contains no control characters, no quote characters, and no
      // backslash characters, then we can safely slap some quotes around it.
      // Otherwise we must also replace the offending characters with safe escape
      // sequences.

        escapable.lastIndex = 0;
        return escapable.test(string) ? '"' + string.replace(escapable, function (a) {
            var c = meta[a];
            return typeof c === 'string'
                ? c
                : '\\u' + ('0000' + a.charCodeAt(0).toString(16)).slice(-4);
        }) + '"' : '"' + string + '"';
    }

   typeof(depth) == "number" || (depth = 0)

   // Limit depth to avoid recursion issues
   if (depth>maxDepth) return "...";


   typeof(embedded) == "boolean" || (embedded = false)
   var newline = false
   var spacer = function(depth) { var spaces = ""; for (var i=0;i<depth;i++) { spaces += "  "}; return spaces }
   var pretty = ""
   if (      typeof(object) === "undefined" ) { pretty += "undefined" }
   else if ( typeof(object) === "boolean" ||
             typeof(object) === "number" ) {    pretty += object.toString() }
   else if ( typeof(object) === "string" ) {    pretty +=  quote(object) }
   // Avoid error if a function is part of an object
   else if ( typeof(object) === "object" && object.constructor.name === "RegExp") {    pretty +=  object.toString() }
   else if ( typeof(object) === "function" ) {    pretty +=  "function ... (" + object.length + ")" }
   else if (        object  === null) {         pretty += "null" }
   else if ( object instanceof(Array) ) {
      if ( object.length > 0 ) {
         if (embedded) { newline = true }
         var content = "";
         for each (var item in object) { content += pp(item, depth+1) + ",\n" + spacer(depth+1) }
         content = content.replace(/,\n\s*$/, "").replace(/^\s*/,"")
         pretty += "[ " + content + "\n" + spacer(depth) + "]"
      } else { pretty += "[]" }
   }
   else if (typeof(object) == "object") {
      if ( Object.keys(object).length > 0 ){
         if (embedded) { newline = true }
         var content = ""
         for (var key in object) {
         content += spacer(depth + 1) + key.toString() + ": " + pp(object[key], depth+2, true) + ",\n"
         }
         content = content.replace(/,\n\s*$/, "").replace(/^\s*/,"")
         pretty += "{ " + content + "\n" + spacer(depth) + "}"
      } else { pretty += "{}"}
   }
   else { pretty += object.toString() }
   return ((newline ? "\n" + spacer(depth) : "") + pretty)
}



   // --- private variables and methods for debug ---------------------------
   var debugOn = true;

   // To escape HTML characters for the console
   var escapeMap = { '"': '&quot;', '&': '&amp;', '<': '&lt;', '>': '&gt;' };
   var escapeHTML = function (text) {
        return text.replace(/[\"&<>]/g, function (a) { return escapeMap[a]; });
    };

   // --- public properties and methods ---------------------------------------
   return {
      // Debug log the arguments to the console (separated by spaces) if debug log is active
      // null and undefined are transformed in string
      // arrays are shown as array (first level only)
      // Example:
      //    Log.debug("My list:", [1,2,3,4], "is", null);
      // TODO Support more formatting as needed
      debug: function() {
         var str, arg, i;
         if (debugOn) {
            //var str =  Array.prototype.slice.call(arguments).join("");
            str = "";
            for (i = 0; i<arguments.length; i++) {
               if (i>0) { str += " "};
               arg = arguments[i];
               if (typeof arg === "undefined") {
                  str += "undefined";
               } else if (arg === null) {
                  str += "null";
               } else if (Array.isArray(arg)) {
                  str += "["+arg.toString()+"]"; // Should handle recursively, limit depth, ...
               } else {
                  str += arg.toString();
               }
            }
            Console.writeln(escapeHTML(str));
            Console.flush();
         }
      },
      pp: pp,
   }

})();

#ifdef DEBUG
var debug = Log.debug;
#endif

// Example:
//Log.debug("My list:", [1,2,3,4], "is", null);








//==== end of include ============================================

//================================================================
// include "FITSFileManager-json.jsh"
//================================================================
/*
    json_parse_state.js
    2012-06-01

    Public Domain.

    NO WARRANTY EXPRESSED OR IMPLIED. USE AT YOUR OWN RISK.

    This file creates a json_parse function.

        json_parse(text, reviver)
            This method parses a JSON text to produce an object or array.
            It can throw a SyntaxError exception.

            The optional reviver parameter is a function that can filter and
            transform the results. It receives each of the keys and values,
            and its return value is used instead of the original value.
            If it returns what it received, then the structure is not modified.
            If it returns undefined then the member is deleted.

            Example:

            // Parse the text. Values that look like ISO date strings will
            // be converted to Date objects.

            myData = json_parse(text, function (key, value) {
                var a;
                if (typeof value === 'string') {
                    a =
/^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2}(?:\.\d*)?)Z$/.exec(value);
                    if (a) {
                        return new Date(Date.UTC(+a[1], +a[2] - 1, +a[3], +a[4],
                            +a[5], +a[6]));
                    }
                }
                return value;
            });

    This is a reference implementation. You are free to copy, modify, or
    redistribute.

    This code should be minified before deployment.
    See http://javascript.crockford.com/jsmin.html

    USE YOUR OWN COPY. IT IS EXTREMELY UNWISE TO LOAD CODE FROM SERVERS YOU DO
    NOT CONTROL.
*/

/*jslint regexp: false*/

/*members "", "\"", ",", "\/", ":", "[", "\\", "]", acomma, avalue, b,
    call, colon, container, exec, f, false, firstavalue, firstokey,
    fromCharCode, go, hasOwnProperty, key, length, n, null, ocomma, okey,
    ovalue, pop, prototype, push, r, replace, slice, state, t, test, true,
    value, "{", "}"
*/

var json_parse = (function () {
    "use strict";

// This function creates a JSON parse function that uses a state machine rather
// than the dangerous eval function to parse a JSON text.

    var state,      // The state of the parser, one of
                    // 'go'         The starting state
                    // 'ok'         The final, accepting state
                    // 'firstokey'  Ready for the first key of the object or
                    //              the closing of an empty object
                    // 'okey'       Ready for the next key of the object
                    // 'colon'      Ready for the colon
                    // 'ovalue'     Ready for the value half of a key/value pair
                    // 'ocomma'     Ready for a comma or closing }
                    // 'firstavalue' Ready for the first value of an array or
                    //              an empty array
                    // 'avalue'     Ready for the next value of an array
                    // 'acomma'     Ready for a comma or closing ]
        stack,      // The stack, for controlling nesting.
        container,  // The current container object or array
        key,        // The current key
        value,      // The current value
        escapes = { // Escapement translation table
            '\\': '\\',
            '"': '"',
            '/': '/',
            't': '\t',
            'n': '\n',
            'r': '\r',
            'f': '\f',
            'b': '\b'
        },
        string = {   // The actions for string tokens
            go: function () {
                state = 'ok';
            },
            firstokey: function () {
                key = value;
                state = 'colon';
            },
            okey: function () {
                key = value;
                state = 'colon';
            },
            ovalue: function () {
                state = 'ocomma';
            },
            firstavalue: function () {
                state = 'acomma';
            },
            avalue: function () {
                state = 'acomma';
            }
        },
        number = {   // The actions for number tokens
            go: function () {
                state = 'ok';
            },
            ovalue: function () {
                state = 'ocomma';
            },
            firstavalue: function () {
                state = 'acomma';
            },
            avalue: function () {
                state = 'acomma';
            }
        },
        action = {

// The action table describes the behavior of the machine. It contains an
// object for each token. Each object contains a method that is called when
// a token is matched in a state. An object will lack a method for illegal
// states.

            '{': {
                go: function () {
                    stack.push({state: 'ok'});
                    container = {};
                    state = 'firstokey';
                },
                ovalue: function () {
                    stack.push({container: container, state: 'ocomma', key: key});
                    container = {};
                    state = 'firstokey';
                },
                firstavalue: function () {
                    stack.push({container: container, state: 'acomma'});
                    container = {};
                    state = 'firstokey';
                },
                avalue: function () {
                    stack.push({container: container, state: 'acomma'});
                    container = {};
                    state = 'firstokey';
                }
            },
            '}': {
                firstokey: function () {
                    var pop = stack.pop();
                    value = container;
                    container = pop.container;
                    key = pop.key;
                    state = pop.state;
                },
                ocomma: function () {
                    var pop = stack.pop();
                    container[key] = value;
                    value = container;
                    container = pop.container;
                    key = pop.key;
                    state = pop.state;
                }
            },
            '[': {
                go: function () {
                    stack.push({state: 'ok'});
                    container = [];
                    state = 'firstavalue';
                },
                ovalue: function () {
                    stack.push({container: container, state: 'ocomma', key: key});
                    container = [];
                    state = 'firstavalue';
                },
                firstavalue: function () {
                    stack.push({container: container, state: 'acomma'});
                    container = [];
                    state = 'firstavalue';
                },
                avalue: function () {
                    stack.push({container: container, state: 'acomma'});
                    container = [];
                    state = 'firstavalue';
                }
            },
            ']': {
                firstavalue: function () {
                    var pop = stack.pop();
                    value = container;
                    container = pop.container;
                    key = pop.key;
                    state = pop.state;
                },
                acomma: function () {
                    var pop = stack.pop();
                    container.push(value);
                    value = container;
                    container = pop.container;
                    key = pop.key;
                    state = pop.state;
                }
            },
            ':': {
                colon: function () {
                    if (Object.hasOwnProperty.call(container, key)) {
                        throw new SyntaxError('Duplicate key "' + key + '"');
                    }
                    state = 'ovalue';
                }
            },
            ',': {
                ocomma: function () {
                    container[key] = value;
                    state = 'okey';
                },
                acomma: function () {
                    container.push(value);
                    state = 'avalue';
                }
            },
            'true': {
                go: function () {
                    value = true;
                    state = 'ok';
                },
                ovalue: function () {
                    value = true;
                    state = 'ocomma';
                },
                firstavalue: function () {
                    value = true;
                    state = 'acomma';
                },
                avalue: function () {
                    value = true;
                    state = 'acomma';
                }
            },
            'false': {
                go: function () {
                    value = false;
                    state = 'ok';
                },
                ovalue: function () {
                    value = false;
                    state = 'ocomma';
                },
                firstavalue: function () {
                    value = false;
                    state = 'acomma';
                },
                avalue: function () {
                    value = false;
                    state = 'acomma';
                }
            },
            'null': {
                go: function () {
                    value = null;
                    state = 'ok';
                },
                ovalue: function () {
                    value = null;
                    state = 'ocomma';
                },
                firstavalue: function () {
                    value = null;
                    state = 'acomma';
                },
                avalue: function () {
                    value = null;
                    state = 'acomma';
                }
            }
        };

    function debackslashify(text) {

// Remove and replace any backslash escapement.

        return text.replace(/\\(?:u(.{4})|([^u]))/g, function (a, b, c) {
            return b ? String.fromCharCode(parseInt(b, 16)) : escapes[c];
        });
    }

    return function (source, reviver) {

// A regular expression is used to extract tokens from the JSON text.
// The extraction process is cautious.

        var r,          // The result of the exec method.
            tx = /^[\x20\t\n\r]*(?:([,:\[\]{}]|true|false|null)|(-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?)|"((?:[^\r\n\t\\\"]|\\(?:["\\\/trnfb]|u[0-9a-fA-F]{4}))*)")/;

// Set the starting state.

        state = 'go';

// The stack records the container, key, and state for each object or array
// that contains another object or array while processing nested structures.

        stack = [];

// If any error occurs, we will catch it and ultimately throw a syntax error.

        try {

// For each token...

            for (;;) {
                r = tx.exec(source);
                if (!r) {
                    break;
                }

// r is the result array from matching the tokenizing regular expression.
//  r[0] contains everything that matched, including any initial whitespace.
//  r[1] contains any punctuation that was matched, or true, false, or null.
//  r[2] contains a matched number, still in string form.
//  r[3] contains a matched string, without quotes but with escapement.

                if (r[1]) {

// Token: Execute the action for this state and token.

                    action[r[1]][state]();

                } else if (r[2]) {

// Number token: Convert the number string into a number value and execute
// the action for this state and number.

                    value = +r[2];
                    number[state]();
                } else {

// String token: Replace the escapement sequences and execute the action for
// this state and string.

                    value = debackslashify(r[3]);
                    string[state]();
                }

// Remove the token from the string. The loop will continue as long as there
// are tokens. This is a slow process, but it allows the use of ^ matching,
// which assures that no illegal tokens slip through.

                source = source.slice(r[0].length);
            }

// If we find a state/token combination that is illegal, then the action will
// cause an error. We handle the error by simply changing the state.

        } catch (e) {
            state = e;
        }

// The parsing is finished. If we are not in the final 'ok' state, or if the
// remaining source contains anything except whitespace, then we did not have
//a well-formed JSON text.

        if (state !== 'ok' || /[^\x20\t\n\r]/.test(source)) {
            throw state instanceof SyntaxError ? state : new SyntaxError('JSON');
        }

// If there is a reviver function, we recursively walk the new structure,
// passing each name/value pair to the reviver function for possible
// transformation, starting with a temporary root object that holds the current
// value in an empty key. If there is not a reviver function, we simply return
// that value.

        return typeof reviver === 'function' ? (function walk(holder, key) {
            var k, v, value = holder[key];
            if (value && typeof value === 'object') {
                for (k in value) {
                    if (Object.prototype.hasOwnProperty.call(value, k)) {
                        v = walk(value, k);
                        if (v !== undefined) {
                            value[k] = v;
                        } else {
                            delete value[k];
                        }
                    }
                }
            }
            return reviver.call(holder, key, value);
        }({'': value}, '')) : value;
    };
}());

/*
    json.js
    2012-10-08

    Public Domain

    No warranty expressed or implied. Use at your own risk.

    This file has been superceded by http://www.JSON.org/json2.js

    See http://www.JSON.org/js.html

    This code should be minified before deployment.
    See http://javascript.crockford.com/jsmin.html

    USE YOUR OWN COPY. IT IS EXTREMELY UNWISE TO LOAD CODE FROM SERVERS YOU DO
    NOT CONTROL.

    This file adds these methods to JavaScript:

        object.toJSONString(whitelist)
            This method produce a JSON text from a JavaScript value.
            It must not contain any cyclical references. Illegal values
            will be excluded.

            The default conversion for dates is to an ISO string. You can
            add a toJSONString method to any date object to get a different
            representation.

            The object and array methods can take an optional whitelist
            argument. A whitelist is an array of strings. If it is provided,
            keys in objects not found in the whitelist are excluded.



    This file creates a global JSON object containing two methods: stringify
    and parse.

        JSON.stringify(value, replacer, space)
            value       any JavaScript value, usually an object or array.

            replacer    an optional parameter that determines how object
                        values are stringified for objects. It can be a
                        function or an array of strings.

            space       an optional parameter that specifies the indentation
                        of nested structures. If it is omitted, the text will
                        be packed without extra whitespace. If it is a number,
                        it will specify the number of spaces to indent at each
                        level. If it is a string (such as '\t' or '&nbsp;'),
                        it contains the characters used to indent at each level.

            This method produces a JSON text from a JavaScript value.

            When an object value is found, if the object contains a toJSON
            method, its toJSON method will be called and the result will be
            stringified. A toJSON method does not serialize: it returns the
            value represented by the name/value pair that should be serialized,
            or undefined if nothing should be serialized. The toJSON method
            will be passed the key associated with the value, and this will be
            bound to the object holding the key.

            For example, this would serialize Dates as ISO strings.

                Date.prototype.toJSON = function (key) {
                    function f(n) {
                        // Format integers to have at least two digits.
                        return n < 10 ? '0' + n : n;
                    }

                    return this.getUTCFullYear()   + '-' +
                         f(this.getUTCMonth() + 1) + '-' +
                         f(this.getUTCDate())      + 'T' +
                         f(this.getUTCHours())     + ':' +
                         f(this.getUTCMinutes())   + ':' +
                         f(this.getUTCSeconds())   + 'Z';
                };

            You can provide an optional replacer method. It will be passed the
            key and value of each member, with this bound to the containing
            object. The value that is returned from your method will be
            serialized. If your method returns undefined, then the member will
            be excluded from the serialization.

            If the replacer parameter is an array of strings, then it will be
            used to select the members to be serialized. It filters the results
            such that only members with keys listed in the replacer array are
            stringified.

            Values that do not have JSON representations, such as undefined or
            functions, will not be serialized. Such values in objects will be
            dropped; in arrays they will be replaced with null. You can use
            a replacer function to replace those with JSON values.
            JSON.stringify(undefined) returns undefined.

            The optional space parameter produces a stringification of the
            value that is filled with line breaks and indentation to make it
            easier to read.

            If the space parameter is a non-empty string, then that string will
            be used for indentation. If the space parameter is a number, then
            the indentation will be that many spaces.

            Example:

            text = JSON.stringify(['e', {pluribus: 'unum'}]);
            // text is '["e",{"pluribus":"unum"}]'


            text = JSON.stringify(['e', {pluribus: 'unum'}], null, '\t');
            // text is '[\n\t"e",\n\t{\n\t\t"pluribus": "unum"\n\t}\n]'

            text = JSON.stringify([new Date()], function (key, value) {
                return this[key] instanceof Date ?
                    'Date(' + this[key] + ')' : value;
            });
            // text is '["Date(---current time---)"]'


        JSON.parse(text, reviver)
            This method parses a JSON text to produce an object or array.
            It can throw a SyntaxError exception.

            The optional reviver parameter is a function that can filter and
            transform the results. It receives each of the keys and values,
            and its return value is used instead of the original value.
            If it returns what it received, then the structure is not modified.
            If it returns undefined then the member is deleted.

            Example:

            // Parse the text. Values that look like ISO date strings will
            // be converted to Date objects.

            myData = JSON.parse(text, function (key, value) {
                var a;
                if (typeof value === 'string') {
                    a =
/^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2}(?:\.\d*)?)Z$/.exec(value);
                    if (a) {
                        return new Date(Date.UTC(+a[1], +a[2] - 1, +a[3], +a[4],
                            +a[5], +a[6]));
                    }
                }
                return value;
            });

            myData = JSON.parse('["Date(09/09/2001)"]', function (key, value) {
                var d;
                if (typeof value === 'string' &&
                        value.slice(0, 5) === 'Date(' &&
                        value.slice(-1) === ')') {
                    d = new Date(value.slice(5, -1));
                    if (d) {
                        return d;
                    }
                }
                return value;
            });


    This is a reference implementation. You are free to copy, modify, or
    redistribute.
*/

/*jslint evil: true, regexp: true, unparam: true */

/*members "", "\b", "\t", "\n", "\f", "\r", "\"", JSON, "\\", apply,
    call, charCodeAt, getUTCDate, getUTCFullYear, getUTCHours,
    getUTCMinutes, getUTCMonth, getUTCSeconds, hasOwnProperty, join,
    lastIndex, length, parse, parseJSON, prototype, push, replace, slice,
    stringify, test, toJSON, toJSONString, toString, valueOf
*/


// Create a JSON object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.

if (typeof JSON !== 'object') {
    JSON = {};
}

(function () {
    'use strict';

    function f(n) {
        // Format integers to have at least two digits.
        return n < 10 ? '0' + n : n;
    }

    if (typeof Date.prototype.toJSON !== 'function') {

        Date.prototype.toJSON = function (key) {

            return isFinite(this.valueOf()) ?
                this.getUTCFullYear()     + '-' +
                f(this.getUTCMonth() + 1) + '-' +
                f(this.getUTCDate())      + 'T' +
                f(this.getUTCHours())     + ':' +
                f(this.getUTCMinutes())   + ':' +
                f(this.getUTCSeconds())   + 'Z' : null;
        };

        String.prototype.toJSON      =
            Number.prototype.toJSON  =
            Boolean.prototype.toJSON = function (key) {
                return this.valueOf();
            };
    }

    var cx = /[\u0000\u00ad\u0600-\u0604\u070f\u17b4\u17b5\u200c-\u200f\u2028-\u202f\u2060-\u206f\ufeff\ufff0-\uffff]/g,
        escapable = /[\\\"\x00-\x1f\x7f-\x9f\u00ad\u0600-\u0604\u070f\u17b4\u17b5\u200c-\u200f\u2028-\u202f\u2060-\u206f\ufeff\ufff0-\uffff]/g,
        gap,
        indent,
        meta = {    // table of character substitutions
            '\b': '\\b',
            '\t': '\\t',
            '\n': '\\n',
            '\f': '\\f',
            '\r': '\\r',
            '"' : '\\"',
            '\\': '\\\\'
        },
        rep;


    function quote(string) {

// If the string contains no control characters, no quote characters, and no
// backslash characters, then we can safely slap some quotes around it.
// Otherwise we must also replace the offending characters with safe escape
// sequences.

        escapable.lastIndex = 0;
        return escapable.test(string) ? '"' + string.replace(escapable, function (a) {
            var c = meta[a];
            return typeof c === 'string' ? c :
                '\\u' + ('0000' + a.charCodeAt(0).toString(16)).slice(-4);
        }) + '"' : '"' + string + '"';
    }


    function str(key, holder) {

// Produce a string from holder[key].

        var i,          // The loop counter.
            k,          // The member key.
            v,          // The member value.
            length,
            mind = gap,
            partial,
            value = holder[key];

// If the value has a toJSON method, call it to obtain a replacement value.

        if (value && typeof value === 'object' &&
                typeof value.toJSON === 'function') {
            value = value.toJSON(key);
        }

// If we were called with a replacer function, then call the replacer to
// obtain a replacement value.

        if (typeof rep === 'function') {
            value = rep.call(holder, key, value);
        }

// What happens next depends on the value's type.

        switch (typeof value) {
        case 'string':
            return quote(value);

        case 'number':

// JSON numbers must be finite. Encode non-finite numbers as null.

            return isFinite(value) ? String(value) : 'null';

        case 'boolean':
        case 'null':

// If the value is a boolean or null, convert it to a string. Note:
// typeof null does not produce 'null'. The case is included here in
// the remote chance that this gets fixed someday.

            return String(value);

// If the type is 'object', we might be dealing with an object or an array or
// null.

        case 'object':

// Due to a specification blunder in ECMAScript, typeof null is 'object',
// so watch out for that case.

            if (!value) {
                return 'null';
            }

// Make an array to hold the partial results of stringifying this object value.

            gap += indent;
            partial = [];

// Is the value an array?

            if (Object.prototype.toString.apply(value) === '[object Array]') {

// The value is an array. Stringify every element. Use null as a placeholder
// for non-JSON values.

                length = value.length;
                for (i = 0; i < length; i += 1) {
                    partial[i] = str(i, value) || 'null';
                }

// Join all of the elements together, separated with commas, and wrap them in
// brackets.

                v = partial.length === 0 ? '[]' : gap ?
                    '[\n' + gap + partial.join(',\n' + gap) + '\n' + mind + ']' :
                    '[' + partial.join(',') + ']';
                gap = mind;
                return v;
            }

// If the replacer is an array, use it to select the members to be stringified.

            if (rep && typeof rep === 'object') {
                length = rep.length;
                for (i = 0; i < length; i += 1) {
                    k = rep[i];
                    if (typeof k === 'string') {
                        v = str(k, value);
                        if (v) {
                            partial.push(quote(k) + (gap ? ': ' : ':') + v);
                        }
                    }
                }
            } else {

// Otherwise, iterate through all of the keys in the object.

                for (k in value) {
                    if (Object.prototype.hasOwnProperty.call(value, k)) {
                        v = str(k, value);
                        if (v) {
                            partial.push(quote(k) + (gap ? ': ' : ':') + v);
                        }
                    }
                }
            }

// Join all of the member texts together, separated with commas,
// and wrap them in braces.

            v = partial.length === 0 ? '{}' : gap ?
                '{\n' + gap + partial.join(',\n' + gap) + '\n' + mind + '}' :
                '{' + partial.join(',') + '}';
            gap = mind;
            return v;
        }
        throw "unexpected type " + typeof value; // To make suse strict happy
    }

// If the JSON object does not yet have a stringify method, give it one.

    if (typeof JSON.stringify !== 'function') {
#ifdef DEBUG
         Log.debug("Creating JSON.stringify");
#endif
        JSON.stringify = function (value, replacer, space) {

// The stringify method takes a value and an optional replacer, and an optional
// space parameter, and returns a JSON text. The replacer can be a function
// that can replace values, or an array of strings that will select the keys.
// A default replacer method can be provided. Use of the space parameter can
// produce text that is more easily readable.

            var i;
            gap = '';
            indent = '';

// If the space parameter is a number, make an indent string containing that
// many spaces.

            if (typeof space === 'number') {
                for (i = 0; i < space; i += 1) {
                    indent += ' ';
                }

// If the space parameter is a string, it will be used as the indent string.

            } else if (typeof space === 'string') {
                indent = space;
            }

// If there is a replacer, it must be a function or an array.
// Otherwise, throw an error.

            rep = replacer;
            if (replacer && typeof replacer !== 'function' &&
                    (typeof replacer !== 'object' ||
                    typeof replacer.length !== 'number')) {
                throw new Error('JSON.stringify');
            }

// Make a fake root object containing our value under the key of ''.
// Return the result of stringifying the value.

            return str('', {'': value});
        };
    }
#ifdef DEBUG
     else {     Log.debug("Using builtin JSON.stringify"); }
#endif


// If the JSON object does not yet have a parse method, give it one.

    if (typeof JSON.parse !== 'function') {
#ifdef DEBUG
         Log.debug("Creating JSON.parse");
#endif
        JSON.parse = json_parse;
   }
#ifdef DEBUG
     else {     Log.debug("Using builtin JSON.parse"); }
#endif


}());

//==== end of include ============================================

//================================================================
// include "FITSFileManager-parameters.jsh"
//================================================================
// FITSFileManager-parameters.js

// This file is part of FITSFileManager, see copyright in FITSFileManager.js

// -----------------------------------------------------------------------------------------------
// This script support the parameters loaded/saved by the PI 'Settings' mechanism
// and the configuration information currently only modifiable by editing this file.
// -----------------------------------------------------------------------------------------------

#include <pjsr/DataType.jsh>


// --- Configuration - Change the definitions as needed.
//                     Some definitions need to be consistent with others or with the code,
//                     but this should be fairly obvious.


// Define a version for the PARAMETERS ( a double, independent of the script version)
#define PARAMETERS_VERSION 1.0


// TODO Should not be a global or define, but scoped variables

// Default values of some parameters, used only on the first usage of the script (when the settings are not yet defined)

// Select the first sequence without -_. or the whole name in &1; (second group is non capturing)
#define FFM_DEFAULT_SOURCE_FILENAME_REGEXP /([^-_.]+)(?:[._-]|$)/
#define FFM_DEFAULT_TARGET_FILENAME_TEMPLATE "&1;_&binning;_&temp;C_&type;_&exposure;s_&filter;_&count;&extension;"
#define FFM_DEFAULT_GROUP_TEMPLATE "&targetDir;"







// Name of key in settings
#define FFM_SETTINGS_KEY_BASE  "FITSFileManager/"


// ====================================================================================================================
// Configuration module
// ====================================================================================================================


// ---------------------------------------------------------------------------------------------------------------------
// Definition of ConfigurationSet data
// ---------------------------------------------------------------------------------------------------------------------
// The configuration is a set of rules that define how to parse a FITS file to
// generate the synthetic variables.
// The configurations are named and grouped in a ConfigurationSet. Only one Configuration
// is active at a time.
// The coniguration object is a 'pure data' representation of the rules (so they can be serialized),
// it contains only strings, numbers, objects and arrays (regexp are represented as strings!).
// Some utility methods in this module support common functions, but it is expected that the
// methods directly operate on the configuration data.
// The ConfigurationSet is a singleton read/writen by the parameters module. A copy is manipulated by the
// Configuration Dialog and replace the singleton if the Dialog exit woth OK.
// One configuration is the 'current' configuration. It is a copy if the selected configuration
// and can be complemented by data used for processing (typcially the implementation of the resolvers).
// That copy is not modified by the ConfigurationDialog or saved as parameter. If a new current configuration
// is selected, a copy of the new configuration will replace the 'current configuration' and all derived
// data will be recalculated.





var ffM_Configuration = (function() {

   var i;

   // Default ConfigurationSet, serve also as example of the structure.

   // --------------------------------------------------------------------------------------------------
   // configurationSet data (ordered list of configurations, the order does not matter for the semantic)
   //    [cpnfiguration]

   // configuration data (the variables is an ordered list)
   //    {name: aString, description: aString, variableList: [variables]}

   // variable:
   //    {name: aString, description: aString, resolver: aName, parameters: {resolverName: {}}}
   // (the content of the 'parameters' object depends on the resolver, it has own key which is the
   //  name of the resolver, allowing to keep data for inactive resolvers)
   // ---------------------------------------------------------------------------------------------------------------------

   var defaultConfigurationSet =
   [
      { name: "Default",
       description: "Common FITS rules",
       variableList:
         [ { name: "type",
             description: "Type of image (flat, bias, ...)",
             show: true,
             resolver: "RegExpList",
             parameters:
               { RegExpList:
                   { key: "IMAGETYP",
                     reChecks:
                       [ { regexp: '/flat/i',
                           replacement: "flat"
                         },
                         { regexp: '/bias/i',
                           replacement: "bias"
                         },
                         { regexp: '/offset/i',
                           replacement: "bias"
                         },
                         { regexp: '/dark/i',
                           replacement: "dark"
                         },
                         { regexp: '/light/i',
                           replacement: "light"
                         },
                         { regexp: '/science/i',
                           replacement: "light"
                         },
                         { regexp: '/.*/',
                           replacement: "&0;"
                         },
                       ]
                   }
               }
           },
           { name: "filter",
             description: "Filter (clear, red, ...)",
             show: true,
             resolver: "RegExpList",
             parameters:
               { RegExpList:
                   { key: "FILTER",
                     reChecks:
                       [ { regexp: '/green/i',
                           replacement: "green"
                         },
                         { regexp: '/blue/i',
                           replacement: "blue"
                         },
                         { regexp: '/red/i',
                           replacement: "red"
                         },
                         { regexp: '/clear/i',
                           replacement: "clear"
                         },
                         { regexp: '/luminance/i',
                           replacement: "luminance"
                         },
                         { regexp: '/.*/',
                           replacement: "&0;"
                         }
                       ]
                       ,
                   }
               }
           },
           { name: "exposure",
             description: "Exposure in seconds",
             show: true,
             resolver: "Integer",
             parameters:
               { Integer:
                   { key: "EXPTIME", // also EXPOSURE
                     abs: true,
                     format: "%4.4d"
                   }
               }
           },
           { name: "temp",
             description: "Temperature in C",
             show: true,
             resolver: "Integer",
             parameters:
               { Integer:
                   { key: "SET-TEMP", // Also CCDTEMP and CCD-TEMP",
                     abs: true,
                     format: "%4.4d"
                   }
               }
           },
           { name: "binning",
             description: "Binning as 1x1, 2x2, ...",
             show: true,
             resolver: "IntegerPair",
             parameters:
               { IntegerPair:
                   { key1: "XBINNING",
                     key2: "YBINNING",
                     format: "%dx%d"
                   }
               }
           },
          //      "NightLongObs" : "LONG-OBS",
         // We should really use DATE-OBS and convert
        // "NightJD"      : "JD",

           { name: "night",
             description: "night (experimental)",
             show: true,
             resolver: "Night",
             parameters:
               { Night:
                   { keyLongObs: "LONG-OBS",
                     keyJD: "JD"
                   }
               }
           },
           { name: "filename",
             description: "Input file name",
             show: false,
             resolver: "FileName",
             parameters:
               { FileName:
                   {
                   }
               }
           },
           { name: "extension",
             description: "Input file extension",
             show: false,
             resolver: "FileExtension",
             parameters:
               { FileExtension:
                   {
                   }
               }
           }
         ],
        builtins: {
           rank: {format: "%4.4d"},
           count: {format: "%4.4d"},
        }
     }

   ];

   // Duplicate the default configuration ,
   // patch in the differences
   var defaultConf = defaultConfigurationSet[0];

   // CAHA version
   var userConf = deepCopyData(defaultConf);
   userConf.name = "User CAHA";
   userConf.description = "Test configuration based on CAHA image";
   // Just the differences from the base configuration - which is about everything
   userConf.variableList =
         [ { name: "type",
             description: "Type of image (flat, bias, ...)",
             show: true,
             resolver: "RegExpList",
             parameters:
               { RegExpList:
                   { key: "IMAGETYP",
                     reChecks:
                       [ { regexp: '/flat/i',
                           replacement: "flat"
                         },
                         { regexp: '/bias/i',
                           replacement: "bias"
                         },
                         { regexp: '/offset/i',
                           replacement: "bias"
                         },
                         { regexp: '/dark/i',
                           replacement: "dark"
                         },
                         { regexp: '/light/i',
                           replacement: "light"
                         },
                         { regexp: '/science/i',
                           replacement: "light"
                         },
                         { regexp: '/.*/',
                           replacement: "&0;"
                         },
                       ]
                   }
               }
           },
           { name: "filter",
             description: "Filter (clear, red, ...)",
             show: true,
             resolver: "RegExpList",
             parameters:
               { RegExpList:
                   { key: "INSFLNAM",
                     reChecks:
                       [ { regexp: '/.*/i',
                           replacement: "&0;"
                         }
                       ]
                   }

               }
           },
           { name: "exposure",
             description: "Exposure in seconds",
             show: true,
             resolver: "Integer",
             parameters:
               { Integer:
                   { key: "EXPTIME", // also EXPOSURE
                     abs: true,
                     format: "%4.4d",
                   }
               }
           },
           { name: "temp",
             description: "Temperature in C",
             show: true,
             resolver: "Integer",
             parameters:
               { Integer:
                   { key: "CCDTEMP", // Also CCDTEMP and CCD-TEMP",
                     abs: true,
                     format: "%4.4d"
                   }
               }
           },
           { name: "binning",
             description: "Binning as 1x1, 2x2, ...",
             show: true,
             resolver: "IntegerPair",
             parameters:
               { IntegerPair:
                   { key1: "CDELT1",
                     key2: "CDELT2",
                     format: "%dx%d"
                   }
               }
           },
           { name: "night",
             description: "night (experimental)",
             show: true,
             resolver: "Night",
             parameters:
               { Night:
                   { keyLongObs: "CAHA TEL GEOLON",
                   // We should really used DATE-OBS (if available) and convert
                    keyJD: "JUL-DATE"
                   }
               }
           },
           { name: "filename",
             description: "Input file name",
             show: false,
             resolver: "FileName",
             parameters:
               { FileName:
                   {
                   }
               }
           },
           { name: "extension",
             description: "Input file extension",
             show: false,
             resolver: "FileExtension",
             parameters:
               { FileExtension:
                   {
                   }
               }
           }
         ];

   defaultConfigurationSet.push(userConf);





   // --- singletons variables ---------------------------------------------------
   // The configuration set (a table of configurations)
   // Used only by the parameters and the DialogConfiguration handler
   var configurationTable = defaultConfigurationSet;

   // The active configuration name
   // The engine work on a COPY of the active configuration, this is why we keep the name only
   var activeConfigurationName = null;


   // Methods used to extract commonly needed information from the ConfigurationSet

   // Get the names of the all the configurations
   var getAllConfigurationNames = function(aConfigurationTable) {
      var names = [];
      for (var i=0; i<aConfigurationTable.length; i++) {
         names.push(aConfigurationTable[i].name);
      }
      return names;
   }

   // Get the rule by name
   var getConfigurationByName = function(aConfigurationTable, name) {
      for (var i=0; i<aConfigurationTable.length; i++) {
         if (aConfigurationTable[i].name === name) return aConfigurationTable[i];
      }
      return null;
   }

   // At least one configuration is required
   var removeConfigurationByName = function(aConfigurationTable, name) {
      if (aConfigurationTable.length>1) {
         for (var i=0; i<aConfigurationTable.length; i++) {
            if (aConfigurationTable[i].name === name) {
               aConfigurationTable.splice(i,1);
               break;
            }
         }
         // Return the name of the new first configuration
         return aConfigurationTable[0].name;
      }
      return null;
   }



   // --- List of all synthethic variables and their comments (2 parallel arrays)
   //     All synthethic variables are currently added to the columns of the file TreeBox
   // Currently just add all variables of the current resolver
   // TODO Handle non resolver variables, move somewhere else
   var syntheticVariableNames = [];
   var syntheticVariableComments = [];


   // The calling code must re-configure the GUI and the engine after calling this function
   var setActiveConfigurationName = function (nameOfNewActiveConfiguration) {
      activeConfigurationName = nameOfNewActiveConfiguration;
#ifdef DEBUG
      Log.debug("Configuration activated: ", activeConfigurationName);
#endif

      // TODO Should be moved
      // Rebuild the list of all sythetic variables that can be shown in an Input files window.
      var activeConfiguration = getConfigurationByName(configurationTable,nameOfNewActiveConfiguration);
      syntheticVariableNames = [];
      syntheticVariableComments = [];
      for (var i=0; i<activeConfiguration.variableList.length; i++) {
         var aVar = activeConfiguration.variableList[i];
         syntheticVariableNames.push(aVar.name);
         syntheticVariableComments.push(aVar.description);
      }
   }



   // -- Support for variable handling

   // Model of variable - define a new variable
   var defineVariable = function(name, description, resolver) {
      var initialValues = deepCopyData(ffM_Resolver.resolverByName(resolver).initial);
      var initialParameters = {};
      initialParameters[resolver] = initialValues;
      return {
         name: name,
         description: description,
         show: true,
         resolver: resolver,
         parameters: initialParameters,
      }
   }

   // Support for active configuration

   var getConfigurationTable = function() {
      return configurationTable;
   }
   var getActiveConfigurationName = function() {
      return activeConfigurationName;
   }
   // Return the reference to the active configuration (NOT TO A WORKIGN COPY)
   var getActiveConfigurationElement = function() {
      return getConfigurationByName(configurationTable,activeConfigurationName);
   }

   var replaceConfigurationTable = function(newConfigurationTable, nameOfNewActiveConfiguration) {
      configurationTable = newConfigurationTable;
      setActiveConfigurationName(nameOfNewActiveConfiguration);
   }

   var createWorkingConfiguration = function() {
      var configuration = getConfigurationByName(configurationTable,activeConfigurationName);
      if (configuration == null) {
         throw "FITSFileManager-parameters - Invalid configuration '"+ name + "'";
      }
      return deepCopyData(configuration);
   }



#ifdef NO
  // TODO See how this is defined in the new mecanism
   // List of FITS keywords shown by default (even if not present in any image) in the input files TreeBox
   var defaultShownKeywords_DEFAULT = [
      "IMAGETYP","FILTER","OBJECT"
      //"SET-TEMP","EXPOSURE","IMAGETYP","FILTER","XBINNING","YBINNING","OBJECT"
   ];
   var defaultShownKeywords_CAHA = [
      "IMAGETYP","INSFLNAM","OBJECT"
      //"SET-TEMP","EXPOSURE","IMAGETYP","FILTER","XBINNING","YBINNING","OBJECT"
   ];

#endif



   // Activate some default configuration
   setActiveConfigurationName(configurationTable[0].name);

   // --- public properties and methods ---------------------------------------
   return {

      // The public singletons
      getConfigurationTable: getConfigurationTable,
      getActiveConfigurationName: getActiveConfigurationName,
      getActiveConfigurationElement: getActiveConfigurationElement,

      // Manage active configurations
      createWorkingConfiguration: createWorkingConfiguration,
      setActiveConfigurationName: setActiveConfigurationName,

      // Methods on a configuration set
      replaceConfigurationTable: replaceConfigurationTable,
      getAllConfigurationNames: getAllConfigurationNames,
      getConfigurationByName: getConfigurationByName,
      removeConfigurationByName: removeConfigurationByName,

      // Support for variables
      defineVariable: defineVariable,

   // TODO Should probably be in the engine or variable handling,
      syntheticVariableNames: syntheticVariableNames,
      syntheticVariableComments: syntheticVariableComments,



   };
}) ();



// ====================================================================================================================
// User Interface Parameters
// ====================================================================================================================

// The object FFM_GUIParameters keeps track of the parameters that are saved between executions
// (or that should be saved)
function FFM_GUIParameters() {
   // Called at end of constructor
   this.initializeParametersToDefaults = function () {

      // Default temp
      this.targetFileNameTemplate = FFM_DEFAULT_TARGET_FILENAME_TEMPLATE;

      // Default regular expression to parse file name
      this.sourceFileNameRegExp = FFM_DEFAULT_SOURCE_FILENAME_REGEXP;

      this.orderBy = "&rank;" // UNUSED


      // Initialiy the first configuration is the default
      this.currentConfigurationIndex = 0;




      // Create templates (use defaults if not yet specified), precompile them
      this.groupByTemplate = FFM_DEFAULT_GROUP_TEMPLATE;
      var templateErrors = [];
      this.targetFileNameCompiledTemplate = ffM_template.analyzeTemplate(templateErrors, FFM_DEFAULT_TARGET_FILENAME_TEMPLATE);
      this.groupByCompiledTemplate = ffM_template.analyzeTemplate(templateErrors, FFM_DEFAULT_GROUP_TEMPLATE);
      if (templateErrors.length>0) {
         throw "PROGRAMMING ERROR - default built in templates invalid";
      }

      // Prepare list of regexp, groupBy template and target file template for use by the user interface.
      // The first element of the list is the last one selected by the user, the others are predefiend elements
      // (currently hardcoded here - could eventually be made editable)
      // There are two parallel arrays, one for the values, and one for a comment displayed in the selection box
      this.targetFileItemListText = [
            this.targetFileNameCompiledTemplate.templateString, // Must be adapted after parameter loading
            FFM_DEFAULT_TARGET_FILENAME_TEMPLATE,
            "&type;/&1;_&binning;_&temp;C_&exposure;s_&filter;_&count;&extension;",
            "&OBJECT;_&filter;_&count;&extension;",
            "&1;_&type?light;_&filter?clear;_&count;&extension;",
            ""
      ];
      this.targetFileItemListComment = [
            "last",
            "detailled, using part of file name",
            "directory by type",
            "Object and filter",
            "type and filter with defaults",
            "(clear)"
      ];

      this.regexpItemListText = [
         regExpToString(this.sourceFileNameRegExp), // Must be adapted after parameter loading
         FFM_DEFAULT_SOURCE_FILENAME_REGEXP,
         "/.*/"
      ];
      this.regexpItemListComment = [
         "last",
         "extract name",
         "(everything)"
      ];


      this.groupItemListText = [
            this.groupByCompiledTemplate.templateString, // Must be adapted after parameter loading
            FFM_DEFAULT_GROUP_TEMPLATE,
            "&filter;",
            "&type?;&filter?;",
            ""
      ];
      this.groupItemListComment = [
            "last",
            "count by directory",
            "count by filter",
            "count by type and filter if present",
            "none (count globally)"
      ];



   }


   // For debugging and logging - result MUST be escaped if written to the console
   this.toString = function() {
      var s = "GUIParameters:\n";
      s += "  targetFileNameTemplate:         " + this.targetFileNameCompiledTemplate.templateString + "\n";
      s += "  sourceFileNameRegExp:           " + regExpToString(this.sourceFileNameRegExp) + "\n";
      s += "  orderBy:                        " + this.orderBy + "\n";
      s += "  groupByTemplate:                " + this.groupByCompiledTemplate.templateString + "\n";
      s += "  currentConfigurationIndex:          " + this.currentConfigurationIndex + "\n";
      return s;
   }

   this.initializeParametersToDefaults();


}

FFM_GUIParameters.prototype.loadSettings = function() {


   function load( key, type )
   {
      var setting = Settings.read( FFM_SETTINGS_KEY_BASE + key, type );
#ifdef DEBUG
      var text = (setting===null ? 'null' : setting.toString());
      // To workaround slow console on 1.7
      if (text.length > 100) {
         text=text.substring(0,100) + "...";
      }
      debug("FFM_GUIParameters.load: "+ key+ ": "+ text + ", ok: " + Settings.lastReadOK);
#endif
      return setting;
   }

   function loadIndexed( key, index, type )
   {
      return load( key + '_' + index.toString(), type );
   }

   var o, t, parameterVersion, templateErrors, configurations, activeConfigurationName;
   if ( (parameterVersion = load( "version",    DataType_Double )) !== null ) {
      if (parameterVersion > PARAMETERS_VERSION) {
         Console.show();
         Console.writeln("Warning: Settings '", FFM_SETTINGS_KEY_BASE, "' have paramter version ", parameterVersion, " later than script parameter version ", PARAMETERS_VERSION, ", settings ignored");
         Console.flush();
      } else {
         if ( (o = load( "targetFileNameTemplate",    DataType_String )) !== null ) {
           templateErrors = [];
           t =   ffM_template.analyzeTemplate(templateErrors,o);
           if (templateErrors.length===0) {
               this.targetFileNameCompiledTemplate = t; // Template correct
           }

         };
         if ( (o = load( "sourceFileNameRegExp",    DataType_String )) !== null ) {
            try {
               this.sourceFileNameRegExp = RegExpFromString(o);
            } catch (err) {
               // Default in case of error in load
               this.sourceFileNameRegExp = FFM_DEFAULT_SOURCE_FILENAME_REGEXP;
#ifdef DEBUG
               debug("loadSettings: bad regexp - err: " + err);
#endif
            }
         };
         if ( (o = load( "orderBy",                  DataType_String )) !== null ) {
            this.orderBy = o;
         }
         if ( (o = load( "groupByTemplate",          DataType_String )) !== null ) {
            templateErrors = [];
            t = ffM_template.analyzeTemplate(templateErrors, o);
            if (templateErrors.length ===0) {
               this.groupByCompiledTemplate = t;
            }
         }

         // Restore the 'last' value in the list of predfined choices
         this.regexpItemListText[0] = regExpToString(this.sourceFileNameRegExp);
         this.groupItemListText[0] = this.groupByCompiledTemplate.templateString;
         this.targetFileItemListText[0] = this.targetFileNameCompiledTemplate.templateString;

         // After 0.8
   if (parameterVersion>0.8) {
            if ( (o = load( "configurations",          DataType_String )) !== null ) {
               configurations = JSON.parse(o);
            }
            if ( (o = load( "activeConfiguration",          DataType_String )) !== null ) {
               activeConfigurationName = o;
            }
            ffM_Configuration.replaceConfigurationTable(configurations, activeConfigurationName);
         }
  }

   } else {
      Console.show();
      Console.writeln("Warning: Settings '", FFM_SETTINGS_KEY_BASE, "' do not have a 'version' key, settings ignored");
      Console.flush();
   }

};


FFM_GUIParameters.prototype.saveSettings = function()
{
   function save( key, type, value ) {
#ifdef DEBUG
      var text = value.toString();
      // To workaround slow console on 1.7
      if (text.length > 100) {
         text=text.substring(0,100) + "...";
      }
      debug("saveSettings: key="+key+", type="+ type+ ", value=" +text);
#endif
      Settings.write( FFM_SETTINGS_KEY_BASE + key, type, value );
   }

   function saveIndexed( key, index, type, value ) {
#ifdef DEBUG
      debug("saveSettings: key="+key+", index="+ index+ ", type="+ type+ ", value=" +value.toString());
#endif
      save( key + '_' + index.toString(), type, value );
   }

   save( "version",                    DataType_Double, PARAMETERS_VERSION );
   save( "targetFileNameTemplate",     DataType_String, this.targetFileNameCompiledTemplate.templateString );
   save( "sourceFileNameRegExp",       DataType_String, regExpToString(this.sourceFileNameRegExp) );
   save( "orderBy",                    DataType_String, this.orderBy );
   save( "groupByTemplate",            DataType_String, this.groupByCompiledTemplate.templateString );
   save( "configurations",             DataType_String, JSON.stringify(ffM_Configuration.getConfigurationTable() ));
   save( "activeConfiguration",        DataType_String, ffM_Configuration.getActiveConfigurationName() );
}


FFM_GUIParameters.prototype.targetTemplateSelection =  [
   FFM_DEFAULT_TARGET_FILENAME_TEMPLATE
];
FFM_GUIParameters.prototype.groupTemplateSelection = [
   FFM_DEFAULT_GROUP_TEMPLATE
];
FFM_GUIParameters.prototype.regexpSelection = [
   FFM_DEFAULT_SOURCE_FILENAME_REGEXP.toString()
];


//==== end of include ============================================
//================================================================
// include "FITSFileManager-fits.jsh"
//================================================================
// FITSFileManager-fits.jsh

// This file is part of FITSFileManager, see copyright in FITSFileManager.js

#include <pjsr/DataType.jsh>


// ------------------------------------------------------------------------------------------------------------------------
// Read the FITS keywords of an image file, supports the HIERARCH convention
// ------------------------------------------------------------------------------------------------------------------------
// Input:  The full path of a file
// Return: An array FITSKeyword, identical to what would be returned by ImageWindow.open().keywords.
//         at least for correct keywords (errors and border line cases may be handled differently)
// Throws: Error if bad format
// The value of a FITSKeyword value is the empty string if the keyword had no value
// Code adapted from FitsKey and other scripts
// ------------------------------------------------------------------------------------------------------------------------
var ffM_loadFITSKeywordsList =  function loadFITSKeywordsList(fitsFilePath ) {

   function searchCommentSeparator( b ) {
      var inString = false;
      for ( var i = 10; i < 80; ++i )
         switch ( b.at( i ) )
         {
         case 39: // single quote
            inString ^= true;
            break;
         case 47: // slash
            if ( !inString )
               return i;
            break;
         }
      return -1;
   }

   // in HIERARCH the = sign is after the real keyword name
   function searchHierarchValueIndicator( b ) {
      for ( var i = 9; i < 80; ++i )
         switch ( b.at( i ) )
         {
         case 39: // single quote, = cannot be later
            return -1;
         case 47: // slash, cannot be later
            return -1;
         case 61: // =, may be value indicator after all
            return i;
         }
      return -1;
   }

#ifdef DEBUG_SHOW_FITS
      debug("ffM_loadFITSKeywordsList: loading '" + fitsFilePath + "'");
#endif

   var f = new File;
   f.openForReading( fitsFilePath );
   try {

   var keywords = [];
   for ( ;; ) {
      var rawData = f.read( DataType_ByteArray, 80 );
#ifdef DEBUG_FITS
      debug("ffM_loadFITSKeywordsList: line - '" + rawData.toString() + "'");
#endif

      var name = rawData.toString( 0, 8 );
      if ( name.toUpperCase() === "END     " ) { // end of HDU keyword list?
         break;
      }
      if ( f.isEOF ) {
         throw new Error( "Unexpected end of file reading FITS keywords, file: " + fitsFilePath );
      }

      var value = "";
      var comment = "";
      var hasValue = false;

      // value separator (an equal sign at byte 8) present?
      if ( rawData.at( 8 ) === 61 ) {
         // This is a valued keyword
         hasValue = true;
         // find comment separator slash
         var cmtPos = searchCommentSeparator( rawData );
         if ( cmtPos < 0) {
            // no comment separator
            cmtPos = 80;
         }
         // value substring
         value = rawData.toString( 9, cmtPos-9 );
         if ( cmtPos < 80 ) {
            // comment substring
            comment = rawData.toString( cmtPos+1, 80-cmtPos-1 );
         }
      } else if (name === 'HIERARCH') {
         var viPos = searchHierarchValueIndicator(rawData);
         if (viPos > 0) {
            hasValue = true;
            name = rawData.toString(9, viPos-10);
            // find comment separator slash
            var cmtPos = searchCommentSeparator( rawData );
            if ( cmtPos < 0 ) {
               // no comment separator
               cmtPos = 80;
            }
            // value substring
            value = rawData.toString( viPos+1, cmtPos-viPos-1 );
            if ( cmtPos < 80 ) {
               // comment substring
               comment = rawData.toString( cmtPos+1, 80-cmtPos-1 );
            }
         }
      }

      // If no value in this keyword
      if (! hasValue) {
         comment = rawData.toString( 8, 80-8 ).trim();
      }


      // Perform a naive sanity check: a valid FITS file must begin with a SIMPLE=T keyword.
      if ( keywords.length === 0 ) {
         if ( name !== "SIMPLE  " && value.trim() !== 'T' ) {
            throw new Error( "File does not seem to be a valid FITS file (SIMPLE T not found): " + fitsFilePath );
         }
      }

      // Add new keyword.
      var fitsKeyWord = new FITSKeyword( name, value, comment);
      fitsKeyWord.trim();
      keywords.push(fitsKeyWord);
#ifdef DEBUG_SHOW_FITS
      debug("ffM_loadFITSKeywordsList: - " + fitsKeyWord);
#endif

   }
   } finally {
   f.close();
   }
#ifdef DEBUG_SHOW_FITS
      debug("ffM_loadFITSKeywordsList: (end) ");
#endif

   return keywords;
};






// ====================================================================================================================
// FITS Keywords support module
// ====================================================================================================================

var ffM_FITS_Keywords = (function() {


   // --- private properties and methods ---------------------------------------

   // Unquote a string value
   var unquote = function unquote(s) {
      if (s===null) { return null;} // Could this happen ?
      if (s.length > 0 && s.charCodeAt(0) === 39 && s.charCodeAt(s.length-1) === 39) {
         // Unquoted string
         s = s.substring(1,s.length-1);
         // Replace double quotes inside by single quotes
         s = s.replace("''","'","g");
         if (s.length>0) {
            // Not the empty string,
            // trim trailing spaces only, but keep the first character if it is a space
            s = s.trimRight();
            if (s.length===0) {
               s = ' ';
            }
         }
      }
      return s;
   }



   // ------------------------------------------------------------------------------------------------------------------------
   // ImageKeywords support - An 'ImageKeywords' keeps track of the FITS keywords of a file, both as an array ordered
   // as in the file PDU and as a map of name to FITSKeyword for the value keywords (keywords that are not null)
   // ------------------------------------------------------------------------------------------------------------------------
   // Prototype for methods operating on ImageKeywords
   var imageKeywordsPrototype = {

      // -- Load the FITS keywords from the file, adding them to the value map too
      loadFitsKeywords:  function loadFitsKeywords(filePath) {
         var imageKeywords = this;
         var name, fitsKeyFromList, i;
         imageKeywords.fitsKeywordsList = ffM_loadFITSKeywordsList(filePath);
         // Make a map of all fits keywords with a value (this remove the comment keywords)
         imageKeywords.fitsKeywordsMap = {};
         for (i=0; i<imageKeywords.fitsKeywordsList.length; i++) {
            fitsKeyFromList = imageKeywords.fitsKeywordsList[i];
            if (!fitsKeyFromList.isNull) {
               name = fitsKeyFromList.name;
               // IMPORTANT: FitsKey is shared with the list
               // TODO Check for duplicates (not supported)
               imageKeywords.fitsKeywordsMap[name] = fitsKeyFromList;
            }
         }
      },

      // -- return the FITSKeyword by name (if keyword has a value), return null otherwise
      getValueKeyword: function getValueKeyword(name) {
         var imageKeywords = this;
         if (imageKeywords.fitsKeywordsMap.hasOwnProperty(name)) {
            return imageKeywords.fitsKeywordsMap[name];
         } else {
            return null;
         }
      },
      // -- Return the value (that is the raw String, as read) of a FITSKeyword
      getValue: function getValue(name) {
         var kw = this.getValueKeyword(name);
         if (kw === null) {
            return null;
         } else {
            return kw.value;
         }
      },
      // -- Return the value as a stripped from outside quotes and trimmed (the PI way)
      getStrippedValue: function getStrippedValue(name) {
         var kw = this.getValueKeyword(name);
         if (kw === null) {
            return null;
         } else {
            return kw.strippedValue;
         }
      },
      // -- Return the FITS keyword value as a string, unquoted if it was a string, following the FITS rules
      // (remove outside quote and inside double quote, trim trailing spaces but not leading spaces)
      // Note that it is not possible to distinguish a string value from a boolean or numeric value with
      // the same representation (both '123' and 123 will result in the same string)
      // The unquoted value is suitable for display
      // This assumes that the value was trimmed (the first and last characters must be the quotes if it is a string value)
      getUnquotedValue: function getUnquotedValue(keywordName) {
         var kw = this.getValueKeyword(keywordName);
         if (kw === null) {
            return null;
         } else {
            return unquote(kw.value);
         }
      },


      // -- return the name of all value key words
      getNamesOfValueKeywords: function getNamesOfValueKeywords() {
         var imageKeywords = this;
         return Object.keys(imageKeywords.fitsKeywordsMap);
      }

   };

   // Factory method for an empty ImageKeywords
   var makeImageKeywords = function makeNew() {
      var imageKeywords = Object.create(imageKeywordsPrototype);
      imageKeywords.fitsKeywordsMap = {};
      imageKeywords.fitsKeywordsList = [];
      return imageKeywords;
   }

   // Factory method of an imageKeywords
   var makeImageKeywordsfromFile = function makeImageKeywordsfromFile(filePath) {
      var imageKeywords = makeImageKeywords();
      imageKeywords.loadFitsKeywords(filePath);
      return imageKeywords;
   };



   // ------------------------------------------------------------------------------------------------------------------------
   // KeywordsSet support - Keeps track of all values keywords in a set of files, in a specific order
   // ------------------------------------------------------------------------------------------------------------------------
   // private prototype
   var keywordsSetPrototype = {
       putKeyword: function putKeyword(name) {
           var keywordsSet = this;
           if (!keywordsSet.allValueKeywordNames.hasOwnProperty(name)) {
               keywordsSet.allValueKeywordNames[name] = keywordsSet.allValueKeywordNameList.length;
               keywordsSet.allValueKeywordNameList.push(name);
           }
       },
       putAllImageKeywords: function putAllImageKeywords(imageKeywords) {
           var keywordsSet = this;
           var kwList = imageKeywords.fitsKeywordsList;
           for (var i=0; i<kwList.length; i++) {
               if (!kwList[i].isNull) {
                   keywordsSet.putKeyword(kwList[i].name);
               }
           }
       },
       // Use size on purpose, as this is a function (length of array is a property)
       size: function size() {
         return this.allValueKeywordNameList.length;
       }
   }

   // Factory method to create empty KeywordsSet
   var makeKeywordsSet = function makeKeywordsSet () {
      var keywordsSet = Object.create(keywordsSetPrototype);
      keywordsSet.allValueKeywordNameList = [];
      keywordsSet.allValueKeywordNames = {}; // Name to index
      return keywordsSet;
   }



   // --- public properties and methods ---------------------------------------

   // Return public methods of this module
   return {
      makeImageKeywordsfromFile: makeImageKeywordsfromFile,
      makeKeywordsSet: makeKeywordsSet,
      // Made public for short format of columns
      unquote: unquote,

      // For unit testing only
      UT: {
         unquote: unquote,
      }
   }


}) ();

//==== end of include ============================================
//================================================================
// include "FITSFileManager-helpers.jsh"
//================================================================
// FITSFileManager-helper.js

// This file is part of FITSFileManager, see copyright in FITSFileManager.js

#include <pjsr/DataType.jsh>




//=========================================================================================================================
// String and formatting utility functions
// ------------------------------------------------------------------------------------------------------------------------



// --- RegExp utility functions


// Return the RE as if it would be in source (with / and flags)
function regExpToString(re) {
   if (re === null) {
      return "";
   } else {
      return re.toSource();
   }
}


// Parse a regexp WITH the '/' and flags, throw exception in case of error
function regExpFromString(reString) {
   if (reString === null) {
      return null;
   }
   if (typeof reString !=='string') {
      throw "PROGRAMMING ERROR - Invalid regexp string, got " + typeof reString;
   }

   if (reString.length===0) {
      return null;
   } else if (reString.length<2) {
      throw "Invalid regular expression - too small"
   } else {
      var reChar = reString.charCodeAt(0);
      if (reChar !== 47) {
         throw "Invalid regular expression - dot not start with /"
      }
      var lastSlash = reString.lastIndexOf("/");
      if (lastSlash<=0 || lastSlash>reString.length) {
         throw "Invalid regular expression - no terminating /"
      }
      var rePart = reString.substring(1,lastSlash);
      var flagsPart = reString.substring(lastSlash+1);

      return new RegExp(rePart, flagsPart);
   }
}

// Parse a regular expression typed by the user. If it does not start wit /,
// assume that it is surrounded with / and has no flag.
function regExpFromUserString(reString) {
   if (reString === null) {
      return null;
   }
   if (typeof reString !=='string') {
      throw "PROGRAMMING ERROR - Invalid regexp string";
   }

   if (reString.length>=2 && reString.charCodeAt(0) !== 47) {
      reString = "/" + reString + "/";
   }
   return regExpFromString(reString);

}


// Create a unique name with the same prefix
function createUniqueName(baseName, existingNames) {
   var reSuffix = /^[^_]+_(\d+)/;
   var reNoSuffix = /_\d+$/;
   var largestSuffix = 0;
   var baseWithoutSuffix = baseName.replace(reNoSuffix,'');
   var reSuffix = /^[^_]+_(\d+)/;
   for (var i=0; i<existingNames.length;i++) {
      var exisitingName = existingNames[i];
      var existingNameWithoutSuffix = exisitingName.replace(reNoSuffix,'');
      if (existingNameWithoutSuffix === baseWithoutSuffix) {
         var suffixMatch = exisitingName.match(reSuffix);
         if (suffixMatch) {
            var n = parseInt(suffixMatch[1]);
            if (n>largestSuffix) {
               largestSuffix = n;
            }
         }
      }
   }
   return baseWithoutSuffix +"_" +(largestSuffix+1)
}


//=========================================================================================================================
// Object data support
// ------------------------------------------------------------------------------------------------------------------------
// Deep copy an object (with object, array, basic types and RegExp only, Date and RegExp are in general not used
function deepCopyData(object) {
   var i;
   var result;
   if (object === null) {
      result = null;
   } else if (Array.isArray(object)) {
      result = [];
      for (i=0; i<object.length;i++) {
         result.push(deepCopyData(object[i]));
      }
   } else if (typeof object !== "object") {
      result = object;
   } else if (object.constructor === Date) {
      result = new Date(object.getTime());
   } else if (object.constructor === String) {
      result = new String(object);
   } else if (object.constructor === Boolean) {
      result = new Boolean(object);
   } else if (object.constructor === Number) {
      result = new Number(object);
   } else if (object.constructor === RegExp) {
      result = regExpFromString(regExpToString(object));
   } else if (object.constructor === Function) {
      throw "Cannot deep copy function";
   } else {
      // Any other data object
      result = {};
      var ps = Object.getOwnPropertyNames(object);
      for ( i=0; i<ps.length;i++) {
         result[ps[i]] = deepCopyData(object[ps[i]]);
      }
   }

   return result;
}





//=========================================================================================================================
// File utility functions
// ------------------------------------------------------------------------------------------------------------------------

function copyFile( sourceFilePath, targetFilePath ) {
   var f = new File;

   f.openForReading( sourceFilePath );
   var buffer = f.read( DataType_ByteArray, f.size );
   f.close();

   f = new File();
   f.createForWriting( targetFilePath );
   f.write( buffer );
   //f.flush(); // optional; remove if immediate writing is not required
   f.close();
}


// Load a file and save it with the new name,
// The file name is added to the ORIGFILE if not already present
// (this is not a standard FITS keyword, another possibility is to use
// the FILENAME FILEXT keywords, but they are not standard either)
// An HISTORY record is also added
#define FFM_FITS_HISTORY_LEADER "PI FITSFileManager from "

function loadSaveFile( sourceFilePath, targetFilePath ) {

   // More than one image may be loaded - this is an issue
   var images = ImageWindow.open( sourceFilePath,"FITSFileManagerLoadSaveAs_Temp", true );
   if (images.length !== 1) {
      throw "File '" + sourceFilePath + "' contains " + images.length + " images, this is not supported.";
   }
   var image = images[0];
   var keywords = image.keywords;
   var firstORIGFILE = false;
   for (var i=0; i<keywords.length; i++) {
      if (keywords[i].name === "ORIGFILE" && keywords[i].value) {
         firstORIGFILE = keywords[i].value;
         break;
      }
   }
   if (firstORIGFILE) {
      Console.writeln("Kept ORIGFILE as: " + firstORIGFILE);
   } else {
      var kw = new FITSKeyword( "ORIGFILE",
               File.extractName(sourceFilePath) + File.extractExtension(sourceFilePath),
               "Original name (FITSFileManager)");
      Console.writeln("Adding " + kw.name + ": '" + kw.value + "'");
      keywords.push( kw );
   }
   var kw = new FITSKeyword( "HISTORY", "", "PI FitsFileManager renamed as " + File.extractName(targetFilePath) + File.extractExtension(targetFilePath));
   keywords.push( kw );
   image.keywords = keywords;

   image.saveAs(targetFilePath,  false, false, false, false);

   image.forceClose();

}





//=========================================================================================================================
// Conversion support
// ------------------------------------------------------------------------------------------------------------------------
var ffM_LookupConverter = (function() {

   var backReferenceRegExp = /&[0-9]+;/;
   var allBackReferenceNumberRegExp = /&([0-9])+;/g;

   var converterPrototype = {
      convert: function convert(unquotedName) {
         if (unquotedName === null) { return null }
         for (var i=0; i<this.compiledConversionTable.length; i++) {
            var compiledConversionEntry = this.compiledConversionTable[i];
            if (compiledConversionEntry[0].test(unquotedName)) {
               return compiledConversionEntry[1](compiledConversionEntry,unquotedName);
            }
         }
         // If not recognized, reject (use a . regexp to recognize anything)
         return null;
     }
   };

   // Create a lookup converter
   // Parameters:
   //      conversionTable: Array  of {regexp:, replacement:}, teh regexp is a string
   // Return: A converter object that convert a string according to the rules.
   return {
      makeLookupConverter: function makeLookupConverter (conversionTable) {
         var c = Object.create(converterPrototype);
         // The conversion table is 'compiled' in a form where the first element is
         // the regular expression (as received) and the second is the method that
         // does the replacement. Unless a template variable is used in the replacement
         // variable, the function just copy the string. More complex replacement
         // can be done, using the source value and matched parts, and possibly formatting
         // as to lower case.
         var compiledConversionTable = [];
         for (var i=0; i<conversionTable.length; i++) {
            var conversionEntry = conversionTable[i];
            var conversionRegExp = regExpFromString(conversionEntry.regexp);
            var conversionResultTemplate = conversionEntry.replacement;
            var conversionResultFunction;
            if (conversionResultTemplate==="&0;") {
               // Assumed frequent case of copying input
               conversionResultFunction = function(compiledEntry, unquotedName) {
                  // Cleanup from special characters
                  return ffM_variables.filterFITSValue(unquotedName);
               }
            } else if (backReferenceRegExp.test(conversionResultTemplate)) {
               // There are back refernce, using a replacing function
               conversionResultFunction = (function(conversionResultTemplate){
                  return function(compiledEntry, unquotedName) {
                     // Get the values of the subexpression (before we just tested the presence)
                     var matchedGroups = compiledEntry[0].exec(unquotedName);
                     var replaceHandler = function(fullString, p1, offset, string) {
                        var matchIndex = + p1; // Convert to index
                        if (matchIndex>= matchedGroups.length) {
                           // TODO Generate error in a more firendly way
                           return "BACKREFERENCETOOLARGE"; // Cannot replace, index too large
                        } else {
                           // Cleanup the returned value to avoid special characters
                           return ffM_variables.filterFITSValue(matchedGroups[matchIndex]);
                        }
                     };

                    return conversionResultTemplate.replace(allBackReferenceNumberRegExp,replaceHandler);
                  }
               })(conversionResultTemplate);

            } else {
               // Literal copy of template (no back reference), make sure we reference the value!
               conversionResultFunction = (function(conversionResultString) {
                  return function(ignored1, ignored2) {
                     return conversionResultString;
                  }
               })(conversionResultTemplate);
            }
            compiledConversionTable.push([conversionRegExp, conversionResultFunction]);
         }
         c.compiledConversionTable = compiledConversionTable;
         return c;
      }
   }
}) ();




// TODO Review location of module

// Global object containing the resolver definitions
// These definitions are complemented by the modules when they are loaded
var ffM_Resolver = (function(){

   var i;

   // Describe the resolver types
   // The 'initial' property value will be deep copied to the parameter property when a new definition is created
   // The 'control' property will be populatedby the GUI when they are created
   var resolvers = [
      {name: 'RegExpList', description: 'Type of image (flat, bias, ...)',
            initial:{key: '?', reChecks: [{regexp: /.*/, replacement: '?'}]},  control: null, parserFactory:null},
      {name: 'Text', description: 'Text of FITS keyword value',
            initial:{key: '?', format: '%ls', case: 'NONE'}, control: null, parserFactory:null},
      {name: 'Integer', description: 'Integer value',
            initial:{key: '?', abs: true, format:'%4.4d'}, control: null, parserFactory:null},
      {name: 'IntegerPair', description: 'Pair of integers (binning)',
            initial:{key1: '?', key2: '?', format:'%dx%d'}, control: null, parserFactory:null},
      {name: 'Constant', description: 'Constant value',
            initial:{value: ''}, control: null, parserFactory:null},
      {name: 'FileName', description: 'Source file name',
            initial:{}, control: null, parserFactory:null},
      {name: 'FileExtension', description: 'Source file extension',
            initial:{}, control: null, parserFactory:null},
      {name: 'Night', description: 'Night (experimental)',
            initial:{keyLongObs: 'LONG-OBS', keyJD: 'JD'}, control: null, parserFactory:null}
   ];

   var resolverByName = function(name) {
      for (var i=0; i<resolvers.length; i++) {
         if (resolvers[i].name === name) return resolvers[i];
      }
      return null;
   }
   var resolverNames = [];
   for ( i=0; i<resolvers.length; i++) {
      resolverNames.push(resolvers[i].name);
   }
   return {
      resolverNames: resolverNames,
      resolverByName: resolverByName,
      resolvers: resolvers,
   }
}) ();



// ====================================================================================================================
// Template parsing and execution module
// ====================================================================================================================

// Create template support in a 'module' like object
var ffM_template = (function() {

  // Something like &stuf;
  var templateRegExp = /&[^&;]+;/g;
  // Extract parts of &var:truepart?falsepart;
  var variableRegExp = /^([^:?]+)(?::([^:?]*))?(?:\?([^:?]*))?/

  var testInvalidLiteralRegExp = /[&\(\);<>=!%*]/;

  // Create a rule that return the parameter literal verbatim
  var makeLiteralRule = function(templateErrors,literal){
    // TODO Check that literal does not contains & ( ) ; < > = ! ( ) and % unless formatting is implemented)
    if (testInvalidLiteralRegExp.test(literal)) {
      templateErrors.push("Invalid characters in literal sequence " + literal);
    }
    var literalRule = function(errors) {
      return literal;
    }
    literalRule.toString = function() {return "literalRule('" + literal +"')"};
    return literalRule;
  }

  // Create a rule that interpolate a variable expression
  var makeLookupRule = function(templateErrors, expression)  {
    var variableName, onFoundAction, onMissingAction;
     // expression has & and ; already removed

     // Parse the expression of variable:present?missing parts, resulting in the corresponding elements in execResult
    var execResult = expression.match(variableRegExp);
    if (execResult === null) {
       templateErrors.push("Invalid variable expression '" + expression + "'");
       return null;
    } else {
        variableName = execResult[1];
    }


    // Create the handler for the case ':present'
    // execResult[2] is the text after the colon and before the end or question mark
    if (execResult[2]==='') {
      // Nothing, we copy the null string, so this is a noop
      onFoundAction = function(expandErrors, variableName, value){
        return ''
      }
      onFoundAction.toString = function(){return "copyLiteral('')"};
    } else if (execResult[2]) {
      // Something, the 'present' text is copied verbatim
      onFoundAction = function(expandErrors, variableName, value){
        return execResult[2];
      }
      onFoundAction.toString = function(){return "formatValueAs('"+execResult[2]+"')"};
    } else {
      // No ':present' part, we copy the source value
      onFoundAction = function(expandErrors, variableName, value){
        return value; // TODO SHOULD SUPPORT FORMATTING THE value
      }
      onFoundAction.toString = function(){return "copyValue()"};
    }

    // Create the handler for the case '?missing'
    // execResult[3] is the text after the question mark
    if (execResult[3]==='') {
      // Nothing, an optional value, we copy the null string
      onMissingAction = function(expandErrors){
        return '';
      }
      onMissingAction.toString = function(){return "copyLiteral('')"};
    } else if (execResult[3]) {
      // The 'missing' text is copied verbatim
      onMissingAction = function(expandErrors){
        return execResult[3]; // There should be no format
      }
      onMissingAction.toString = function(){return "copyLiteral('"+execResult[2]+"')"};
    } else {
      // No ?missing' part, we cannot generate the template in case of missing value
      onMissingAction = function(expandErrors, variableName){
         expandErrors.push("No value for the variable '" + variableName + "'");
         return '';
      }
      onMissingAction.toString = function(){return "reject()"};
    }

    // The lookup variable rule itself, that will use the handlers above
    var lookUpRule = function(expandErrors,variableResolver) {
      var value = variableResolver(variableName);
      // do not use 'undefined' to be 'use strict' friendly
      if (value !== null) {
        return onFoundAction(expandErrors, variableName, value);
      } else {
        return onMissingAction(expandErrors, variableName, value);
      }
    }
    lookUpRule.toString = function() { return "lookUpRule('" + variableName + "':[onFound:" + onFoundAction + "]"+ ":[onMissing:" + onMissingAction + "])"; }
    return lookUpRule;
  }


   // --- public properties and methods ---------------------------------------
  return {
    // Analyze a template string and return the compiled template (or null),
    // push the errors to templateErrors, which must be an array
    analyzeTemplate: function(templateErrors, template) {
#ifdef DEBUG_TEMPLATE
      debug("analyzeTemplate:'" + template + "'");
#endif
      // The replacing handler global variables
      var rules = []; // Invalid if error is not empty
      var iNext = 0; // next character that will be examined
      var replaceHandler = function(match, offset, string) {
        //print ("  rh: ", match, offset, string);
        if (offset>iNext) {
          rules.push(makeLiteralRule(templateErrors,string.substring(iNext, offset)));
        }
        rules.push(makeLookupRule(templateErrors,match.substring(1,match.length-1)));
        iNext = offset + match.length;
        return ''; // replace by nothing, ignored anyhow
      }
      // Each match will create the rule for the preceding literal text and the current match
      // Use 'replace' as it provides the need match information, if the replacement is not really used.
      template.replace(templateRegExp, replaceHandler);
      // If required add literal rule for trailing literal text
      if (template.length>iNext) {
          rules.push(makeLiteralRule(templateErrors,template.substring(iNext)));
      }

      // -- Defines the compiled Template
      // This objects is returned when a template has been analyzed (it is like a compiled regexp,
      // although it is for generating text rather than parsing it).
      var templateRuleSet = {
         // toString for debugging
         toString: function() {return rules.toString();},
         // Original string as a property
         templateString: template,
         requiredVariables: [], // TODO
         optionalVariables: [], // TODO

         // Method to expand the template using the variables returned by the variableResolver,
         // return the exanded string, null in case of error
         // The expandErrors must be an array, errors will be pushed to that array.
         // If any error is pushed, the returned value is meaningless
         expandTemplate: function(expandErrors, variableResolver) {
            // Execute the rules one by one, pushing the result
            var result = [];
            for (var i = 0; i<rules.length; i++) {
               result.push(rules[i](expandErrors, variableResolver));
            }
            return result.join('');
         }
      };

      return templateRuleSet;
    } // analyzeTemplate

  } // ffM_template object
})();



// ====================================================================================================================
// Variable definition and lookup module
// ====================================================================================================================


var ffM_variables = (function() {

   // -- Install all the parsers in the  configuration (must be called once only on a configuration copy)
   var installParsers = function(configuration) {
#ifdef DEBUG
      debug("ffM_variables.installParsers: for",configuration.name);
#endif
      var variableList = configuration.variableList;
      for (var i=0; i<variableList.length; i++) {
         var variableDefinition = variableList[i];
         var resolverImpl = ffM_Resolver.resolverByName(variableDefinition.resolver);
         var parameters = variableDefinition.parameters[variableDefinition.resolver];
         variableDefinition.parser = resolverImpl.parserFactory(configuration, parameters);
      }
#ifdef DEBUG
      debug("ffM_variables.installParsers: nmb parsers installed:",variableList.length);
#endif
   }

   // All parsing rules have the own configuration parameters object as argument
   // the fits values and the variables parsed to far

   ffM_Resolver.resolverByName('Integer').parserFactory = function(configuration, parameters){
#ifdef DEBUG
      debug("resolver factory Integer for :",Log.pp(parameters));
#endif
     return (
         function parseInteger(ruleParameters,imageKeywords,imageVariables,inputFile) {
            var valueString = imageKeywords.getValue(ruleParameters.key);
            // Accept float also
            var valueF =  parseFloat(valueString);
            if (isNaN(valueF)) {
               return null;
            } else {
               if (ruleParameters.abs) { valueF = Math.abs(valueF);}
               return format(ruleParameters.format, Math.round( valueF));
            }
         }
      )
   }

   ffM_Resolver.resolverByName('Text').parserFactory = function(configuration, parameters){
#ifdef DEBUG
      debug("resolver factory Text for :",Log.pp(parameters));
#endif
     return (
         function parseText(ruleParameters,imageKeywords,imageVariables,inputFile) {
            var valueString = imageKeywords.getValue(ruleParameters.key);
            if (valueString === null) {
               return null;
            } else {
               var cleanedValue = filterFITSValue(valueString);
               if (ruleParameters.case === 'UP') {
                  cleanedValue = cleanedValue.toUpperCase();
               } else if (ruleParameters.case === 'DOWN') {
                  cleanedValue = cleanedValue.toLowerCase();
               }
               return format(ruleParameters.format, cleanedValue);
            }
         }
      )
   }

   ffM_Resolver.resolverByName('IntegerPair').parserFactory = function(configuration, parameters){
#ifdef DEBUG
      debug("resolver factory IntegerPair for :",Log.pp(parameters));
#endif
     return (
         function parseIntegerPair(ruleParameters,imageKeywords,imageVariables,inputFile) {
            var valueString1 = imageKeywords.getValue(ruleParameters.key1);
            var valueString2 = imageKeywords.getValue(ruleParameters.key2);
            // Accept float also
            var valueF1 =  parseFloat(valueString1);
            var valueF2 =  parseFloat(valueString2);
            if (isNaN(valueF1) || isNaN(valueF2)) {
               return null;
            } else {
               return format(ruleParameters.format, Math.round( valueF1), Math.round( valueF2));
            }
         }
      )
   }

   ffM_Resolver.resolverByName('Constant').parserFactory = function(configuration, parameters){
#ifdef DEBUG
      debug("resolver factory Constant for :",Log.pp(parameters));
#endif
     return (
         function parseConstant(ruleParameters,imageKeywords,imageVariables,inputFile) {
            return ruleParameters.value;
         }
      )
   }

   ffM_Resolver.resolverByName('RegExpList').parserFactory = function(configuration, parameters) {
      // Prepare lookup converter
      var lookupConverter = ffM_LookupConverter.makeLookupConverter(parameters.reChecks);
#ifdef DEBUG
      debug("resolver factory RegExpList for :",Log.pp(parameters));
#endif
      return (
         function parseRegExpList(ruleParameters,imageKeywords,imageVariables,inputFile) {
            var value = imageKeywords.getUnquotedValue(ruleParameters.key);
            if (value === null) return null;

            return lookupConverter.convert(value);
         }
      )
   }

   ffM_Resolver.resolverByName('FileName').parserFactory = function(configuration, parameters){
#ifdef DEBUG
     debug("resolver factory FileName for :",Log.pp(parameters));
#endif
     return (
         function parseFileName(ruleParameters,imageKeywords,imageVariables,inputFile) {
            return  File.extractName(inputFile);
         }
      )
   }

   ffM_Resolver.resolverByName('FileExtension').parserFactory = function(configuration, parameters){
#ifdef DEBUG
      debug("resolver factory FileExtension for :",Log.pp(parameters));
#endif
     return (
         function parseFileExtension(ruleParameters,imageKeywords,imageVariables,inputFile) {
            return  File.extractExtension(inputFile);
         }
      )
   }

   ffM_Resolver.resolverByName('Night').parserFactory = function(configuration, parameters){
#ifdef DEBUG
      debug("resolver factory Night for :",Log.pp(parameters));
#endif
     return (
         function parseNight(ruleParameters,imageKeywords,imageVariables,inputFile) {
            var longObs = imageKeywords.getValue(ruleParameters.keyLongObs); // East in degree
            // longObs = -110;
            // TODO Support default longObs
            var jd = imageKeywords.getValue(ruleParameters.keyJD);
            if (longObs && jd) {
               var jdLocal = Number(jd) + (Number(longObs) / 360.0) ;
               var nightText = (Math.floor(jdLocal) % 1000).toString();
               return nightText;
            } else {
               return null;
            }
         }
      )
   }





   // Extract the variables to form group names and file names from the file name and the FITS keywords
   // They act as 'synthethic' keywords (the purpose is to normalize their representation for ease of use)
   // Parameters:
   //    inputFile: Full path of input file (to extract file anme etc...)
   //    imageKeywords: A FitsFileManager imageKeyword object (all FITS keywords of the image)
   //    variableList: The variable definitions of the current rule
    function makeSyntheticVariables(inputFile, imageKeywords, variableList) {

      var inputFileName =  File.extractName(inputFile);

      var variables = {};

      for (var i=0; i<variableList.length; i++) {
         var variableDefinition = variableList[i];
         var parameters = variableDefinition.parameters[variableDefinition.resolver];
         variables[variableDefinition.name] = variableDefinition.parser(parameters,imageKeywords, variables, inputFile);
      }

#ifdef DEBUG
      debug("makeSyntheticVariables: made " + Object.keys(variables).length + " synthetics keys for file " + inputFileName);
#endif

      return variables;

   }





   // Remove special characters from FITS key values to avoid bizare or illegal file names
   // Leading and trailing blank and invalid characters are removed
   // Embedded invalid characters are collapsed to one underline.
   // An all blank value will return null, considering the keyword as 'missing' when used in templates,
   // this helps supporting files created with a program like SIPS that write keywords as OBJECT even
   // when it is all blank.
   // Parameter
   //    value: an unquoted string (without the FITS quotes, embedded quote will be handled as special characters)
   // Return:
   //    null if it is an all space value, the cleaned up string
   function filterFITSValue(value) {
      if (value === null) {
         return null;
      }
      var name = value.trim();
      var result = '';
      var i = 0;
      var hadValidChar = false;
      var mustAddUnderline = false;
      while (i<value.length) {
      var c = name.charAt(i);
      // TODO Make the list of special characters configurable
      if ( ("0" <= c && c <= "9") || ("a" <= c && c <= "z") || ("A" <= c && c <= "Z") || (c === '-') || (c === '.') || (c === '_') ) {
         if (mustAddUnderline) {
            result = result + '_';
            mustAddUnderline = false;
         };
         result = result + c;
         hadValidChar = true;
      } else if (hadValidChar) {
         mustAddUnderline = true;
      }
      i++;
      }
      return (result.length>0) ? result : null;
   }


   // NOT YET USED
   // From mschuster
   // ...something strange like l`~!@#$%^&()_-+= {}[];',ïnput.fit maps to l___nput.
   function filterViewId(id) {
      var fId = "";
      if (id.length == 0) {
         return "_";
      }
      var c = id.charAt(0);
      if ("0" <= c && c <= "9") {
         fId = fId + "_";
      }
      for (var i = 0; i != id.length; ++i) {
         c = id.charAt(i);
         fId = fId + (
            (("0" <= c && c <= "9") || ("a" <= c && c <= "z") || ("A" <= c && c <= "Z")) ? c : "_"
         );
         if (fId.length > 3 && fId.substring(fId.length - 4, fId.length) == "____") {
            fId = fId.substring(0, fId.length - 1);
         }
      }
      return fId;
   }

   // --- public properties and methods ---------------------------------------
   return {
      makeSyntheticVariables: makeSyntheticVariables,
      filterFITSValue: filterFITSValue,
      filterViewId: filterViewId,
      installParsers: installParsers,
   }

}) ();








//==== end of include ============================================
//================================================================
// include "FITSFileManager-engine.jsh"
//================================================================
// FITSFileMannager-engine.js


// This file is part of FITSFileManager, see copyright in FITSFileManager.js


// ------------------------------------------------------------------------------------------------------------------------
// Engine
// Own and manage the file lists, their variables and the generation of the target file names.
// Own the current 'compiled' configuration, the rules currently active with prepared parsers and lookup methods
// ------------------------------------------------------------------------------------------------------------------------

function FFM_Engine(guiParameters) {

   // This is from the GUI
   this.outputDirectory = "";
#ifdef DEBUG
   this.outputDirectory = "C:/temp";
#endif

   // Configuration related variables
   this.currentConfiguration = null;
   this.shownSyntheticKeyNames = {}; // A synthetic variable is shown in the source file table if its name is a key of this object


   // Variables that can be reset (when doing clear all)
   this.reset = function() {
      // Cache of file information. 3 parallel array. The order of the elements is usually NOT the same as shown in the GUI
      this.inputFiles = []; // Array of the full path of the input files
      this.inputFITSKeywords = [];  // Array of 'imageKeywords' for the corresponding input file
      this.inputVariables = [];  // Array of Map of stable synthethic variables for the corresponding input file

      // Global FITS key information - (the index of the name in keywordsSet.allValueKeywordNameList give also the column offset in the GUI)
      this.keywordsSet = ffM_FITS_Keywords.makeKeywordsSet();   // array of names of all accumulated FITS keywords from all files
      this.shownFITSKeyNames = {}; // A FITSKeyWord is shown in the source file table if its name is a key of this object

      this.resetTarget();
    };

    this.resetTarget = function () {
      // Target files is a subset of the inputFiles, each file is defined by an index
      // in input file and the corresponding new file name. Unchecked files are not
      // present in the list (the values are undefined).
      // The targetFilesIndices is in the order of the TreeBox and contain the index  of the corresponding
      // target file in the inputFiles array (note that not all inputFiles have a corresponging targetFile)
      // The targetFiles and errorPerFile are in the order of the inputFiles.
      // The targetFiles contains the file name without the base directory (common to all files) or null in case of error.
      // The errorPerFile contains the errors String or null if no error,
      this.targetFilesIndices = [];
      this.targetFiles = [];
      this.errorPerFile = [];
      this.nmbFilesTransformed = 0;
      this.nmbFilesInError = 0;
      this.nmbFilesSkipped = 0;
    };



   // -- Set the rules to load variables from images and context information
   // Called at initialization and in case of change
    this.setConfiguration = function(configuration) {
#ifdef DEBUG
      debug("FFM_Engine.setConfiguration - " + configuration.name);
#endif
      this.currentConfiguration = configuration; // This is a working copy;

      ffM_variables.installParsers(this.currentConfiguration);

      // We add default keywords, we do not remove old ones
      // Set 'is visible' for the list of default keywords
      var defaultShownKeywords = ["IMAGETYP","FILTER","OBJECT"]; // TODO configuration.defaultShownKeywords;
      for (var i = 0; i < defaultShownKeywords.length; ++i) {
         var name = defaultShownKeywords[i];
         this.shownFITSKeyNames[name] = true;
      }
      // Use the 'shown' attribute in the configuration
      this.shownSyntheticKeyNames = {};
      var vl = this.currentConfiguration.variableList;
      for (var i = 0; i<vl.length;i++) {
         var name = vl[i].name;
         if (vl[i].show) {
            this.shownSyntheticKeyNames[name] = true;
         }
      }

#ifdef DEBUG
      debug("FFM_Engine.setConfiguration - shownFITSKeyNames size",Object.keys(this.shownFITSKeyNames).length,
        "shownSyntheticKeyNames size",Object.keys(this.shownSyntheticKeyNames).length);
      // debug("FFM_Engine.setConfiguration - " + Log.pp(this.currentConfiguration));
#endif

    }

   this.isVariableVisible = function(name) {
       var v = this.shownSyntheticKeyNames.hasOwnProperty(name);
#ifdef DEBUG
      //debug("FFM_Engine.isVariableVisible -",name,v);
#endif
       return v;
   }


   // Rebuild after change of configuration  (GUI ,ust also refresh itself))
   this.rebuildAll = function() {
      for (var i=0; i<this.inputFiles.length; i++) {
         var fileName = this.inputFiles[i];
         this.inputFITSKeywords[i] = ffM_FITS_Keywords.makeImageKeywordsfromFile(fileName);
         // Create the synthethic variables using the desired rules
         this.inputVariables[i] = ffM_variables.makeSyntheticVariables(this.inputFiles[i],
            this.inputFITSKeywords[i],
             this.currentConfiguration.variableList)
       }
   }




   // -- Add a list of files
   this.addFiles = function (fileNames) {

#ifdef DEBUG
      debug("addFiles: Adding "+fileNames.length + " files");
#endif

      var nmbFilesAdded = 0;
      var nmbFilesDuplicated = 0;
      for ( var i = 0; i<fileNames.length; i++ ) {
#ifdef DEBUG
         debug("addFiles: Adding file " + i + ": '" + fileNames[i] + "', first check for duplicate, if ok add at " + this.inputFiles.length);
#endif
         if (this.inputFiles.indexOf(fileNames[i]) < 0) // Skip files already in the list
         {
            var imageKeywords = ffM_FITS_Keywords.makeImageKeywordsfromFile(fileNames[i]);
            this.inputFiles.push(fileNames[i]);
            this.inputFITSKeywords.push(imageKeywords);
            // Create the synthethic variables using the desired rules
            var variables = ffM_variables.makeSyntheticVariables(fileNames[i], imageKeywords,
                this.currentConfiguration.variableList);

            this.inputVariables.push(variables);
            nmbFilesAdded++;
         } else {
            nmbFilesDuplicated ++;
         }
      }

      Console.writeln("" + nmbFilesAdded + " file" + (nmbFilesAdded >1?"s":"") + " added" +
         (nmbFilesDuplicated>0? (", " + nmbFilesDuplicated + " duplicated file(s) ignored.") : "."));

   }

   // -- Remove a file by name
   this.removeFiles = function (fileName) {
      var index = this.inputFiles.indexOf(fileName);
      if (index < 0) {
         throw ("SCRIPT ERROR : removeFiles: file " + fileName + " not in inputFiles");
      }
      this.inputFiles.splice(index,1);
      this.inputFITSKeywords.splice(index,1);
      this.inputVariables.splice(index,1);
   }


   // -----------------------------------------------------------------------------------------------------
   // ---  Build the list of target files for the checked input files, result stored in object variables
   this.buildTargetFiles = function(listOfFiles) {
   // -----------------------------------------------------------------------------------------------------

#ifdef DEBUG
      debug("buildTargetFiles: list of " + listOfFiles.length + " files");
      debug("buildTargetFiles: targetFileNameTemplate = '" + guiParameters.targetFileNameCompiledTemplate.templateString  + "'");
      debug("buildTargetFiles: sourceFileNameRegExp = '" + guiParameters.sourceFileNameRegExp + "'");
      debug("buildTargetFiles: groupByTemplate = '" + guiParameters.groupByCompiledTemplate.templateString  + "'");
#endif

      this.resetTarget();


      // Separate directory from file name part in target template
      var targetFileNameTemplateString = guiParameters.targetFileNameCompiledTemplate.templateString;
      var indexOfLastSlash = targetFileNameTemplateString.lastIndexOf('/');
      if (indexOfLastSlash>0) {
         var targetDirectoryTemplate= targetFileNameTemplateString.substring(0,indexOfLastSlash);
         var targetNameTemplate= targetFileNameTemplateString.substring(indexOfLastSlash+1);
      } else {
         var targetDirectoryTemplate = '';
         var targetNameTemplate= targetFileNameTemplateString;
      }
#ifdef DEBUG
      debug("buildTargetFiles: targetDirectoryTemplate = '" + targetDirectoryTemplate + "', targetNameTemplate = '" +  targetNameTemplate + "'");
#endif

      // Compile directory template (copy others)
      var templateErrors = [];
      var targetDirectoryCompiledTemplate = ffM_template.analyzeTemplate(templateErrors,targetDirectoryTemplate);
      var groupByCompiledTemplate = guiParameters.groupByCompiledTemplate;
      var targetFileNameCompiledTemplate = guiParameters.targetFileNameCompiledTemplate;

#ifdef DEBUG
      debug("buildTargetFiles: targetDirectoryCompiledTemplate = " + targetDirectoryCompiledTemplate);
      debug("buildTargetFiles: groupByCompiledTemplate = " + groupByCompiledTemplate);
      debug("buildTargetFiles: targetFileNameCompiledTemplate = " + targetFileNameCompiledTemplate);
#endif

      // A map of group count values
      var countingGroups = {};


      // Variables used in the loop
      var group;
      var count;
      var expansionErrors;
      // For use by variable resolver (they are updated at each loop iteration and accessed by the lexical scope)
      var variables;
      var rankString;
      var countString;
      var fitsKeywords;

      // Replace variables, rank and regexp results (accessed by lexical scope)
      var targetDirectoryVariableResolver = function(v) {
               if (variables.hasOwnProperty(v)) {
                  return variables[v];
               } else if (v === "rank") {
                  return rankString;
               } else if (regexpVariables.hasOwnProperty(v)) {
                  return regexpVariables[v];
               } else {
                  return ffM_variables.filterFITSValue(fitsKeywords.getUnquotedValue(v));
               }
      };


      // Replace variables, rank, regexp results and targetDir (accessed by lexical scope)
      var groupByVariableResolver = function(v) {
               if (variables.hasOwnProperty(v)) {
                  return variables[v];
               } else if (v === "rank") {
                  return rankString;
               } else if (v === "targetDir") {
                  return targetDirectory;
               } else if (regexpVariables.hasOwnProperty(v) ) {
                  return regexpVariables[v];
               } else {
                  return ffM_variables.filterFITSValue(fitsKeywords.getUnquotedValue(v));
               }
      };

      // Replace variables, rank, regexp results and count (accessed by lexical scope)
      var targetFileVariableResolver = function(v) {
               if (variables.hasOwnProperty(v)) {
                  return variables[v];
               } else if (v === "rank") {
                  return rankString;
               } else if (v === "count") {
                  return countString;
               } else if (regexpVariables.hasOwnProperty(v) ) {
                  return regexpVariables[v];
               } else {
                  return ffM_variables.filterFITSValue(fitsKeywords.getUnquotedValue(v));
               }
      };


      // Calculate transformation for each file, in order (order is important for the group functionality)

      for (var inputOrderIndex = 0; inputOrderIndex < listOfFiles.length; ++inputOrderIndex) {

            // Get all information required on input file, its variables, FITS keys ...

            var inputFile = listOfFiles[inputOrderIndex];

            var inputFileIndex = this.inputFiles.indexOf(inputFile);
            if (inputFileIndex < 0) {
               throw ("SCRIPT ERROR : buildTargetFiles: file not in inputFiles: " + inputFile + " (" + inputOrderIndex + ")");
            }
#ifdef DEBUG
            debug("buildTargetFiles: " + inputOrderIndex + ": processing inputFile[" + inputFileIndex + "] = " + inputFile);
#endif

            var inputFileName =  File.extractName(inputFile);

            // synthetic variables are preclaculated at when the file is added and explictely recalculated if
            // the configuration is changed, so we use the precalculated values.
            variables = this.inputVariables[inputFileIndex];
            // FITS keywords are also stable in a file
            fitsKeywords = this.inputFITSKeywords[inputFileIndex];

            // Some variables cannot be precalculated when the file is loaded, because
            // they depend on properties that vary in the main dialog, like the regular expression to
            // parse file names or the order of the files or other data that may change after file is loaded.

            // Variable to represent parts of the input file name, as parsed by a user specified regexp.
             //   &1; &2;, ... The corresponding match from the sourceFileNameRegExp
            var regexpVariables = [];
            if (guiParameters.sourceFileNameRegExp !== null) {
               var inputFileNameMatch = guiParameters.sourceFileNameRegExp.exec(inputFileName);
#ifdef DEBUG
               debug ("buildTargetFiles: inputFileNameMatch= " + inputFileNameMatch);
#endif
               if (inputFileNameMatch !== null) {
                  for (var j = 0; j<inputFileNameMatch.length; j++) {
                     regexpVariables[j.toString()] = inputFileNameMatch[j]
                  }
               }
            }

            //   &rank;      The rank in the list of files of the file being moved/copied
            rankString = format(this.currentConfiguration.builtins.rank.format,inputOrderIndex);


            // Do the template operations (keeping track of error)

            expansionErrors = [];
            var targetDirectory = targetDirectoryCompiledTemplate.expandTemplate(expansionErrors,targetDirectoryVariableResolver);
#ifdef DEBUG
            debug("buildTargetFiles: expanded targetDirectory = " + targetDirectory + ", errors = " + expansionErrors);
#endif

            if (expansionErrors.length ===0) {
            // Expand the groupByTemplate to form the id of the counter (targetDir may be used)
               group = groupByCompiledTemplate.expandTemplate(expansionErrors,groupByVariableResolver);
#ifdef DEBUG
               debug("buildTargetFiles: expanded group = " + group + ", errors: " + expansionErrors.join(","));
#endif
               if (expansionErrors.length === 0) {

                  // The count variable must be handeld especially
                  //   &count;    The number of file in the same group
                  count = 0;
                  if (countingGroups.hasOwnProperty(group)) {
                   count = countingGroups[group];
                  }
                  count ++;
                  countingGroups[group] = count;
                  countString = format(this.currentConfiguration.builtins.count.format,count);
#ifdef DEBUG
                  debug("buildTargetFiles: for group = " + group + ", count = " + countString);
#endif

                  var targetString = targetFileNameCompiledTemplate.expandTemplate(expansionErrors,targetFileVariableResolver);
#ifdef DEBUG
                  debug("buildTargetFiles: expanded targetString = " + targetString + ", errors: " + expansionErrors.join(","));
#endif
                  // This requires that the variable 'extension' be well defined
                  // Add a default extension
                  if (File.extractExtension(targetString).length === 0) {
                     targetString += variables['extension'];
                  }
               }
            }

            this.targetFilesIndices.push(inputFileIndex);
            if (expansionErrors.length>0) {
               this.targetFiles.push(null);
               this.errorPerFile.push(expansionErrors.join(", "));
               this.nmbFilesInError += 1;
            } else {
               this.targetFiles.push(targetString);
               this.errorPerFile.push(null);
               this.nmbFilesTransformed += 1;
            }
         }
#ifdef DEBUG
         debug("buildTargetFiles: Total files: " + this.targetFiles.length);
#endif

    }



    // --- Check that the operations can be executed for a list of files ------------------------------
    this.checkValidTargets = function(listOfFiles) {

      var errors = [];

      // Check if files are still in the same order, otherwise the &count; may have changed
      for (var i = 0; i < listOfFiles.length; ++i) {
         var inputFile = listOfFiles[i];
         var inputFileIndex = this.inputFiles.indexOf(inputFile);
         if (inputFileIndex < 0) {
            throw ("SCRIPT ERROR : check: file not in inputFiles: " + inputFile + " (" + i + ")");
         }
         if (this.targetFilesIndices.length<i ||
            this.targetFilesIndices[i] !== inputFileIndex) {
            // Sort order changed
            return ["The order of some column changed since last refresh, please refresh"];
         }
      }

      for (var i = 0; i < listOfFiles.length; ++i) {
         var inputFile = listOfFiles[i];
         if (! File.exists(inputFile)) {
            errors.push("File '"  + inputFile + "' is not present any more, please refresh'");
         }
      }

      // Check if any file is in error
      for (var i=0; i<this.errorPerFile.length; i++) {
         var index = this.targetFilesIndices[i];
         var inputFile = this.inputFiles[index];
         if (this.errorPerFile[i]) {
            errors.push("File '"+ inputFile + "' has variable expansion error");
         }
      }

      // Check duplicates target names
      var targetFileNameInputFile = {};
      for (var i=0; i<this.targetFiles.length; i++) {
         var index = this.targetFilesIndices[i];
         var targetString = this.targetFiles[i];
         // Null are skipped files or files in error
         if (targetString !== null) {
            var inputFile = this.inputFiles[index];
            if (targetFileNameInputFile.hasOwnProperty(targetString)) {
               errors.push("File '"  + inputFile + "' generates same file '" + targetString + "' as '" + targetFileNameInputFile[targetString] +"'");
            }
            targetFileNameInputFile[targetString] = inputFile;
         }
      }

      // Check bad names (empty, /, ...)

      // Check existing target files
      for (var i=0; i<this.targetFiles.length; i++) {
         var index = this.targetFilesIndices[i];
         var targetString = this.targetFiles[i];
         // Null are skipped files or files in error
         if (targetString !== null) {
            var inputFile = this.inputFiles[index];
          var targetFilePath = this.outputDirectory + "/" + targetString;
            if (File.exists(targetFilePath)) {
               errors.push("File '"  + inputFile + "' generates the already existing file '" + targetFilePath + "'");
            }
         }
      }

      return errors;
    }



    // -- Make List of text accumulating the transformation rules for display --------------
   this.makeListsOfTransforms = function() {
      var listsOfTransforms = {
         inputFileIndices: [],
         inputFiles: [],
         targetFiles: [],
         errorMessages: [],
      }
      for (var i = 0; i<this.targetFiles.length; i++) {
         var index = this.targetFilesIndices[i];
         var inputFile = this.inputFiles[index];
         var targetFile = this.targetFiles[i];
         var errorList = this.errorPerFile[i];
         listsOfTransforms.inputFileIndices.push(i); // We want the index in the inputBox
         listsOfTransforms.inputFiles.push(inputFile);
         if (targetFile) {
            listsOfTransforms.targetFiles.push(targetFile);
            listsOfTransforms.errorMessages.push(null);
         } else {
            listsOfTransforms.targetFiles.push(null);
            listsOfTransforms.errorMessages.push(errorList);
         }
      }
      return listsOfTransforms;
    }


   // -- Execute copy, move or loadSave operation ----------------------------------------------------
   // Return a text that may be show to the user
   this.executeFileOperations = function (engine_mode) {

      var count = 0;

      for (var i=0; i<this.targetFiles.length; i++) {

         var index = this.targetFilesIndices[i];
         var targetString = this.targetFiles[i];
         var inputFile = this.inputFiles[index];

         var targetFile = this.outputDirectory + "/" + targetString;

#ifdef DEBUG
         debug("executeFileOperations: targetFile = " + targetFile );
#endif
         var targetDirectory = File.extractDrive(targetFile) +  File.extractDirectory(targetFile);
#ifdef DEBUG
         debug("executeFileOperations: targetDirectory = " + targetDirectory );
#endif

         // Create target directory if required
         if (!File.directoryExists(targetDirectory)) {
            console.writeln("mkdir " + targetDirectory);
            if (EXECUTE_COMMANDS) {
                  File.createDirectory(targetDirectory, true);
            } else {
                  console.writeln("*** COMMAND NOT EXECUTED - EXECUTE_COMMANDS IS FALSE FOR DEBUGGING PURPOSE");
            }
         }

         // TO BE ON SAFE SIDE, was already checked
         if (File.exists(targetFile)) {
         for ( var u = 1; ; ++u )  {
            for( var n = u.toString(); n.length < 4 ; n = "0" + n);
            // TODO This does not take 'extension' into account
               var tryFilePath = File.appendToName( targetFile, '-' + n );
#ifdef DEBUG
               debug("executeFileOperations: tryFilePath= " + tryFilePath );
#endif
               if ( !File.exists( tryFilePath ) ) { targetFile = tryFilePath; break; }
            }
         }

         switch (engine_mode) {
            case 0:
            console.writeln("move " + inputFile +"\n  to "+ targetFile);
            if (EXECUTE_COMMANDS) File.move(inputFile,targetFile);
            break;
         case 1:
            console.writeln("copy " + inputFile+"\n  to "+ targetFile);
            if (EXECUTE_COMMANDS)  copyFile(inputFile,targetFile);
            break;
         case 2:
            console.writeln("load  " + inputFile+"\n write "+ targetFile);
            if (EXECUTE_COMMANDS)  loadSaveFile(inputFile,targetFile);
            break;

         }

         count ++;

         // May be this allows abort ?
         processEvents();
         // May be useful as we load /save a lot of data or images
         gc();

      }
      var action = ["moved","copied","load/saved"][engine_mode];
      var text = count.toString() + " checked file(s) where " + action + " out of " + this.inputFiles.length + " file(s) in input list";
      console.writeln(text);
      return text;
   };

#ifdef IMPLEMENTS_FITS_EXPORT
   // -- Export the keywords of a list of files
   this.exportFITSKeywords = function() {
      var tab = String.fromCharCode(9);
      var f = new File();
      var fileName = "FITS_keys";
      var fileDir = this.outputDirectory;
      var t = fileDir + "/" + fileName + ".txt";
      // Create numbered file nameto create new file
      if ( File.exists( t ) ) {
         for ( var u = 1; ; ++u ) {
            for( var n = u.toString(); n.length < 4 ; n = "0" + n);
            var tryFilePath = File.appendToName( t, '-' + n );
            if ( !File.exists( tryFilePath ) ) { t = tryFilePath; break; }
         }
      }
      f.create(t);

      // output header (tab separated selected fits keyword + 'Filename')
      var allFITSKeyNames = this.keywordsSet.allValueKeywordNameList;
      for ( var i =0; i<allFITSKeyNames.length; i++) {
         var name = allFITSKeyNames[i];
         if (this.shownFITSKeyNames.hasOwnProperty(name)) {
            f.outTextLn(key+tab);
         }
      }
      f.outTextLn("Filename"+String.fromCharCode(10,13)); // LF, CR

      // output FITS data
      for ( var j =0; j< this.targetFilesIndices.length; j++) {
         var inputIndex = this.targetFilesIndices[i];

         // LIKEY NOT CORRECT
         var key = this.inputFITSKeywords[inputIndex].fitsKeywordsList;
         for ( var i = 0; i< allFITSKeyNames.length; i++) {

            var name = allFITSKeyNames[i];
            if (!this.shownFITSKeyNames.hasOwnProperty(name)) continue;
            for (var k in key) {
               if (!(key[k].name === name)) continue;
               if (key[k].isNumeric) {
                  var value = parseFloat(key[k].value)
               } else {
                  var value = key[k].value;
                  value = value.replace( /'/g, "" );
                  value = value.replace( / /g, "" ); //delete left space
                  value = value.replace( /:/g, "." );
               }

               f.outText(value.toString());
               for (var w = value.toString().length; w < 8; w++) f.outText(" ");
               f.outText(tab);
               k=-1;
               break;
            }
            if (k > -1) f.outText("        "+tab);
         }
         f.outTextLn(this.inputFiles[j]+String.fromCharCode(10,13));
      }
      f.close();
      console.writeln("FITSKeywords saved to ",t);
   }
#endif

   // -- Return true if move or copy is possible
   this.canDoOperation = function() {
         return !((!this.inputFiles.length) || (!this.outputDirectory));
   }

   // ----------------------------------------------------------------------------------------------------
   this.reset();
   // ----------------------------------------------------------------------------------------------------

}


//==== end of include ============================================
//================================================================
// include "FITSFileManager-text.jsh"
//================================================================
// FITSFileManager-text.js

// This file is part of FITSFileManager, see copyright in FITSFileManager.js


// This module contains the fixed text for help, tooltips, GUI labels and buttons,
// it does NOT contain the text of templates or used as part of the
// algorithms or configuration/settings (as FITS key names, last template used, ...)

// Text formatted as html is in the 'H' variable,
// Plain text (that could contain & and other xml markup entities) are in the variable 'T'.

// For the format of help see http://pixinsight.com/forum/index.php?topic=4598.msg31979#msg31979
//  It should support html 4 and some css



var Text = (function() {

// --- Local text, used to assemble other texts  ------------------------------
// Some texts are used as tooltip and reused to form the help text

var BASE_HELP_TEXT = "\
<p>FITSFileManager allows you to copy or move FITS image files to new locations, generating the \
new location path from a template with the replacement of variables with values extracted from FITS keys \
and other information.\
<p/>You select the files to move/copy (files can be individually checked or un-checked) \
and select a predefined template or enter a new template to generate the target path using variables to substitute values \
based on the source image file name, FITS keywords or synthethic variables. \
Various other parameters can be adapted to fine tune the path generation. \
The list of transformation is updated as you type templates and other parameters. \
The synthetic variables are defined in a 'configuration' that you may edit. \
";

var VARIABLE_HELP_TEXT = "\
<p/>The variables have the general form '&amp;name:present?missing;', although most of the time \
they have simply the form '&amp;name;'. The 'name' identifies the variable, it may be a FITS key or a synthetic variable name. \
<ul><li>The optional 'present' part is the string that will be used as the replacement value if the variable \
has a value, Usually ':present' is not specified and the value of the variable is used as the replacement string. You can also \
have an empty 'present' value (as &amp;TELESCOP:;), in which case the variable is checked for presence (an error is \
generated if the variable is missing) but its value does not contribute to the target path.</li> \
<li>The optional '?missing' part is used if the variable is not present in the file (for example '&OBJECT?unknown;'). \
You can also have an empty 'missing' value (like '&amp;binning?;') in which case there is no error if the variable  \
has no value. </li>\
</ul><p>The synthetic variables are described in the section 'target template' below. They are built \
from the FITS keywords (following rules specified in a 'configuration'). \
from the number of the file being processed or from the a regular expression applied to the file name. \
The source file regular expression can be used, for example, to extract the part of the file name \
before the first dash and use it as a prefix for all files. \
<p/>The files are processed in the order they appear in the table, the variable '&amp;rank;' provides \
the image number to help generating a unique file file. \
In addition a 'group' string can be generated using from the 'group' template. A '&amp;count;' \
variable is increased for each different group (for example each target directory if the group template \
contained the target directory). \
The values are cleaned up of special characters, so that they form legal file names. \
";


var TARGET_TEMPLATE_TOOLTIP_A = "\
Define how the target file path will be generated. The text of this field is used \
as the output path, except that the variables are replaced by their value according to \
the current 'configuration'.<br/>\
";

// Part used only in tooltip
var TARGET_TEMPLATE_TOOLTIP_B = "\
Variables (like  &amp;name; or &amp;name:present?absent;) are replaced by values defined from the file \
information and FITS keywords. The details on variables, especially the use of 'present' and 'absent' \
is defined in the help available by the icon at bottom right.<br/>\
";


var TARGET_TEMPLATE_TOOLTIP_C = "\
The variables include the FITS keywords and the synthetic variables defined in the configuration. \
A typical configuration contains the following keywords (your actual configuration may be different):<\br/>\
<dl>\
   <dt>&amp;binning;</dt><dd>Binning from XBINNING and YBINNING as integers, like 2x2.</dd>\
   <dt>&amp;exposure;</dt><dd>The exposure from EXPOSURE, but as an integer (assume seconds).<\dd>\
   <dt>&amp;extension;</dt><dd>The extension of the source file (with the dot.), will use input extension if not specified<\dd>\
   <dt>&amp;filename;</dt><dd>The file name part of the source file.<\dd>\
   <dt>&amp;filter;</dt><dd>The filter name from FILTER as lower case trimmed normalized name.<\dd>\
   <dt>&amp;night;</dt><dd>An integer identifying the night, requires JD and LONG-OBS - EXPERIMENTAL.<\dd>\
   <dt>&amp;temp;</dt><dd>The SET-TEMP temperature in C as an integer.<\dd>\
   <dt>&amp;type;</dt><dd>The IMAGETYP normalized to 'flat', 'bias', 'dark', 'light'.<\dd>\
   <dt>&amp;0; &amp;1;, ... </dt><dd>The corresponding match from the source file name regular expression field.<\dd>\
</dl>\
<p>The following keywords are dynamic (their values depends on the file order) and are always present:\
<dl>\
   <dt>&amp;count;</dt><dd>The number of the file being moved/copied int the current group.<\dd>\
   <dt>&amp;rank;</dt><dd>The number of the file in the order of the input file list.<\dd>\
</dl>You can enter the template or select one of the predefined one and optionaly modify it.<br/>\
";

var SOURCE_FILENAME_REGEXP_TOOLTIP = "\
Defines a regular expression (without the surrounding slashes) that will be applied to all file names, \
without the extension. The 'match' array resulting from the regular expression matching can be used \
in the target file name template as numbered variables. '&0;' represent the whole matched expression, '&1' the first group, and so on \
In case of error the field turns red. \
You can enter the regexp or select one of the predefined one and optionally modify it. \
<br/>See https:\/\/developer.mozilla.org\/en-US\/docs\/JavaScript\/Guide\/Regular_Expressions for more informations on regular expresssions. \
<p>\
";



var GROUP_TEMPLATE_TOOLTIP = "\
Defines the template to generate a group name used by the synthetic variable '&count;'. \
Each FITS image generate a group name exactly as it generates a path, but using the group template. \
A count of image will be kept for each different group name and can be used as the variable '&count;' in the \
target path template. \
All variables can be used in the group template, except &count;. In addition you can use the following variable:\
<dl><dt>&targetDir;</dt><dd>The directory part of the target file name.</dd>\
</dl>Leave blank or use a fixed name to have a single global counter.\
<p>Example of group templates:<dl>\
<dt>'&targetDir;'</dt><dd>counts images in each target directory.</dd> \
<dt>'&filter;'</dt><dd>counts the images separetely for each filter (independently of the target directory).<dd/> \
</dl>You can enter the template or select one of the predefined one and optionaly modify it.\
</p>";

var HELP_CONFIGURATION = "\
The configuration defines how the synthetic variables are created, for example the FITS key to \
use, the format of the value, specific operations like changing case or renaming filter names. \
FITSFileManager keeps track of multiple configurations, only one at a time may be active. \
Currently the list of configuration is fixed (this will be made more dynamic in the future). \
It is expected that most users could use the default configuration, or may be the default with \
some simple adjustments. It is also possible to have a set of configuration for different \
image type, but currently you must still change the configuration or the template between \
image types.  Some cofiguration options may require more advanced understanding of regular \
expressions and the FITS key words. \
<br/>Click on the Manage configurations... button to change or edit the configuration. \
";

var HELP_OPERATIONS = "<p>The operations Copy/Move copy or move the files directly, without \
adding any FITS keywords.  The operation Load/SaveAs loads each image temporarily in the workspace \
and save it to the new location. An ORIGFILE keyword with the original file name is added if it is not already present. \
<br/>The operation buttons may be disabled if the operation is not possible (for example if the \
output directory is not specified).</p>\
";


// --- Exported texts, refer as Text.H.xxx  -----------------------------------

return {
H: {

  // Combine help for global help
  HELP_TEXT: (
      "<html>" +
      "<h1><font color=\"#06F\">FITSFileManager</font></h1>" + BASE_HELP_TEXT +
      "<h3><font color=\"#06F\">Variables</font></h3/>" + VARIABLE_HELP_TEXT +
      "<h3><font color=\"#06F\">Target template</font></h3/>" + TARGET_TEMPLATE_TOOLTIP_A + TARGET_TEMPLATE_TOOLTIP_C +\
      "Example of template:\<br/><tt>&nbsp;&nbsp;&nbsp;&amp;1;_&amp;binning;_&amp;temp;C_&amp;type;_&amp;exposure;s_&amp;filter;_&amp;count;&amp;extension;</tt>"+
      "<h3><font color=\"#06F\">Source filename reg exp</font></h3>" + SOURCE_FILENAME_REGEXP_TOOLTIP +
      "Example of regular expression:<br/><tt>&nbsp;&nbsp;&nbsp;([^-_.]+)(?:[._-]|$)</tt><p>" +
      "<h3><font color=\"#06F\">Group template</font></h3>" +  GROUP_TEMPLATE_TOOLTIP +
      "Example of group definition:<br/><tt>&nbsp;&nbsp;&nbsp;&amp;targetdir;</tt><p> " +
      "<h3><font color=\"#06F\">Configuration of mappings</font></h3>" +  HELP_CONFIGURATION +
      "<h3><font color=\"#06F\">Operations</font></h3>" + HELP_OPERATIONS +
      "</html>"),

   TARGET_FILE_TEMPLATE_TOOLTIP: (
      TARGET_TEMPLATE_TOOLTIP_A+TARGET_TEMPLATE_TOOLTIP_B+TARGET_TEMPLATE_TOOLTIP_C),

   SOURCE_FILENAME_REGEXP_TOOLTIP:
      SOURCE_FILENAME_REGEXP_TOOLTIP,

   GROUP_TEMPLATE_TOOLTIP :
      GROUP_TEMPLATE_TOOLTIP,

   HELP_LABEL: ("<b>" + TITLE + " v" + VERSION + "</b> &mdash; Copy or move FITS image " +
           "files using values derived from FITS keywords and from original file name, using a template " +
           "to create the target directory/file name. See the help for more details."),

   // Tools tips CAN be html

   FILES_TREEBOX_TOOLTIP: ( "List of input files - you can add and remove files with the buttons below.\n" +
                               "Select the files you want to operate on with the check box (all by default),\n"+
                               "The columns shown the file name (always) and selected synthethic and loaded FITS keywords,\n" +
                               "select the columns you want to see with the 'text' icon button below at left.\n"+
                               "The FITS keywords and the variables can be used in a template even if they are not in a visible column.\n" +
                               "You can sort the files by clicking on a column header. Then click the 'refresh' button.\n"+
                               "Beware - sort is in alphabetical order even for numbers"),

   KEY_BUTTON_TOOLTIP: "Show all keywords and variables of the selected file,\nalso select the column to shown in the Input file list",

   DIRADD_BUTTON_TOOLTIP: "Add folder including subfolders",

   CHECK_SELECTED_BUTTON_TOOLTIP: "Check selected images",

   CHECK_UNSELECTED_BUTTON_TOOLTIP: "Uncheck selected images",

   REMOVE_FILES_BUTTON_TOOLTIP:  "Remove selected images from the list",

   REMOVE_ALL_FILES_BUTTON_TOOLTIP: "Remove all images from the list",



   KEYWORDNAMES_GROUPBOX_TOOLTIP: "The left side are the keywords used by default,\nThe right side are the keywords used in the current configuration. ",

   TYPECONVERSION_GROUPBOX_TOOLTIP: ("The value of the IMAGETYP keywords are tested with each regular expression in turn (left column),\n" +
           "at the first match, the corresponding value (right column) is returned as the synthetic variable &type;.\n" +
           "The variables &0;, &1; ... may be used to insert the matching groups of the regular expression.\n"+
           "The replaced values are 'cleaned' of special characters."),

   FILTERCONVERSION_GROUPBOX_TOOLTIP: ("The value of the FILTER keywords are tested with each regular expression in turn (left column),\n" +
           "at the first match, the corresponding value (right column) is returned as the synthetic variable &type;.\n" +
           "The variables &0;, &1; ... may be used to insert the matching groups of the regular expression.\n"+
           "The replaced values are 'cleaned' of special characters."),

   OUTPUTDIR_TOOLTIP: "Select the base output directory.\nAny directory created by your template will be below this directory.",

   OUTPUTDIR_SELECT_TOOLTIP: "Select the base output directory.",

   TRANSFORM_TREEBOX_TOOLTIP: ("List of selected files and how they will be converted.\n" +
                               "If there is any error, the corresponding file will be in red.\n"+
                               "You can correct the error or uncheck the corresponding input file\n"+
                               "The files are in the order of the input (do a Refresh if you sorted the input),\n"+
                               "This matters for the &count; and &rank; keywords."),

   CHECK_BUTTON_TOOLTIP: "Check that the target files are valid\nthis is automatically done before any other operation",

   REFRESH_BUTTON_TOOLTIP:  "Refresh the list of operations\nrequired after a sort on an header (there is on onSort event)",

   MOVE_BUTTON_TOOLTIP:  "Move the checked files to the output directory.\nNo HISTORY or ORIGFILE keyword added",

   COPY_BUTTON_TOOLTIP: "Copy the checked files in the output directory.\nNo HISTORY or ORIGFILE keyword added",

   LOADSAVE_BUTTON_TOOLTIP:  ("Load the checked files and save them in the output directory.\n" +
         "BEWARE: Not supported for files containing multiple HDU (multiple images).\n"+
         "Add ORIGFILE keyword with original file name if not already present.\n" +
         "Add HISTORY keyword with new file name.\n"),

   CONFIGURE_BUTTON_TOOLTIP:  ("Select a new current configuration or modify it.\n"+
                               "A configuration defines the synthetic variables and their format.\n"),
   SELECT_CONFIGURATION_BUTTON_TOOLTIP: ("Select the configuration to use as current configuration"),

   HELP_BUTTON_TOOLTIP:  "Browse Documentation",

   COMPLETION_CONTINUE_BUTTON_TOOLTIP : "Continue working in FITSFileManager, moved files have been removed from input list",
   COMPLETION_KEEP_BUTTON_TOOLTIP : "Keep checked files in input list",
   COMPLETION_REMOVE_BUTTON_TOOLTIP : "Remove checked files from input list",
   COMPLETION_LEAVE_BUTTON_TOOLTIP : "Exit FITS file manager",


   VARIABLE_SELECTION_TOOLTIP: ("Synthetic variables available in this configuration.<br/>" +
      "You may add, remove or move variable using the buttons below the list.<br/>" +
      "Select a variable to view or edit its parameters."
      ),

   VARIABLE_RESOLVER_TOOLTIP: ("Select one of the variable resolution algorithm:<dl>"+
      "<dt>Text</dt><dd>The text of the FITS key, filtered from special characters.</dd>" +
      "<dt>Integer</dt><dd>A FITSS keyword value parsed as an integer value</dd>"+
      "<dt>IntegerPair</dt><dd>Two FITS keyword values parsed as two integers value</dd>" +
      "<dt>RegExpList</dt><dd>A FITS keyword and a list of regular expression and replacements</dd>" +
      "<dt>Constant</dt><dd>A fixed value (for test or parameterization)</dd>" +
      "<dt>Night</dt><dd>An algorithm to help classify by night (used mostly for sorting)</dd>" +
      "<dt>FileName</dt><dd>The source file name</dd>" +
      "<dt>FileExtension</dt><dd>The source file extension</dd>" +
      "</dl>"
      ),
   },

   // T: raw text
   T: {
      KEYWORDNAMES_GROUPBOX_TITLE: "Keyword remapping ",
      TYPECONVERSION_GROUPBOX_TITLE: "Remapping of IMAGETYP ",
      FILTERCONVERSION_GROUPBOX_TITLE: "Remapping of FILTER ",
      REMAPPING_SECTION_PART_TEXT: "Remapping of keywords and values",
      OUPUT_SECTION_TEXT_PART:  "Output base directory",
      GET_DIRECTORY_DIALOG_CAPTION: "Select Output Directory",
      RESULT_SECTION_PART_TEXT:  "Resulting operations" ,

      CHECK_BUTTON_TEXT: "Check validity",
      REFRESH_BUTTON_TEXT: "Refresh list",
      MOVE_BUTTON_TEXT: "Move files",
      COPY_BUTTON_TEXT: "Copy files",
      LOADSAVE_BUTTON_TEXT:  "Load / SaveAs files",
      CONFIGURE_BUTTON_TEXT:  "Manage configurations...",

      COMPLETION_TITLE: "FITSFileManager operation result",
      COMPLETION_CONTINUE_BUTTON_TEXT : "Continue in FITSFileManager",
      COMPLETION_KEEP_BUTTON_TEXT : "Continue in FITSFileManager\nKeep checked files",
      COMPLETION_REMOVE_BUTTON_TEXT : "Continue in FITSFileManager\nRemove checked files",
      COMPLETION_LEAVE_BUTTON_TEXT : "Leave FITSFileManager",

   }

   }; // return
}) ();


//==== end of include ============================================

#include <pjsr/Sizer.jsh>
//#include <pjsr/FrameStyle.jsh>
#include <pjsr/TextAlign.jsh>
#include <pjsr/StdIcon.jsh>
#include <pjsr/StdCursor.jsh>
#include <pjsr/StdButton.jsh>
#include <pjsr/FrameStyle.jsh>
#include <pjsr/Color.jsh>

#include <pjsr/ButtonCodes.jsh>
#include <pjsr/FocusStyle.jsh>


//================================================================
// include "FITSFileManager-config-gui.jsh"
//================================================================
// FITSFileManager-config-gui.js

// This file is part of FITSFileManager, see copyright in FITSFileManager.js


// This file contains tww modules:
//    ffM_GUI_support - General utility controls (currently used in this file only)
//    ffM_GUI_config  - Dialog to manage configuration sets


#define GROUP_BOX_MIN_SIZE 22

// ========================================================================================================================
// ffM_GUI_support - GUI support methods and controls
//   This module contains controls independent of the FITSFileManager data model
//   (but specialized in term of layout for use in the FITSFileManager gui)
// ========================================================================================================================
var ffM_GUI_support = (function (){

   // ---------------------------------------------------------------------------------------------------------
   // IconButtonBar - an horizontal bar of buttons defined by icons
   // parameters:
   //     buttons:  an array of objects  {icon:aBitmapName, toolTip: aText, action: aFunction()}
   // ---------------------------------------------------------------------------------------------------------
   var IconButtonBar = function(parent, buttons) {
      this.__base__ = Control;
      this.__base__(parent);

      var i, button, toolButton;

      this.sizer =  new HorizontalSizer;
      this.sizer.margin = 6;
      this.sizer.spacing = 4;

      for (var i=0; i<buttons.length;i ++) {
         var button = buttons[i];
#ifdef DEBUG
         Log.debug(i,button.icon, button.toolTip);
#endif
         var toolButton = new ToolButton( parent );
         this.sizer.add(toolButton);
         toolButton.icon = new Bitmap( button.icon );
         toolButton.toolTip = button.toolTip;
         toolButton.onClick = button.action;
      }
      this.sizer.addStretch();
   }
   IconButtonBar.prototype = new Control;




   // ---------------------------------------------------------------------------------------------------------
   // ManagedList_Box: A list selection box with the possibility to add, remove and move elements,
   //                  the list elements amy have multiple properties represented in multiple columns.
   // parameters:
   //    parent: UI Control
   //    modeldescription: An array of element model descriptions (one description for each column),
   //         each description is an object with the property (more could be added later):
   //            propertyName: name of the propery to use in the element model for the corresponding column
   //   initialListModel: A (usually empty) initial list model (an array of object with the properties described by modelDescription)
   //                    Usually the model is specified by the modelListChange call (including for initial populate)
   //   elementFactory: A function() that return a new element model (called when new is created)
   //   toolTip: The tool tip text
   //   selectionCallBack: a function(elementModel) that return the element model currently selected. an be null if
   //                      no element is selected (for example last one is deleted).
   //   sorted: if true the elements are sorted on the first column
   // ---------------------------------------------------------------------------------------------------------
   function ManagedList_Box(parent, modelDescription, initialListModel, elementFactory, toolTip, selectionCallback, sorted) {
      this.__base__ = Control;
      this.__base__(parent);

      var i;
      (typeof sorted === 'undefined') && (sorted = false);

      // -- Model (the model will be updated in place)
      var listModel = [];


      // -- private methods
      // Create a node based on the model described in modelDescription
      var makeNode = function(treeBox, nodeModel, index) {
         //Log.debug('ManagedList_Box: makeNode -',Log.pp(nodeModel),Log.pp(modelDescription));
         var node = new TreeBoxNode( treeBox, index);
         for (var i=0; i<modelDescription.length;i++){
            node.setText(i, nodeModel[modelDescription[i].propertyName].toString());
         }
         return node;
      }



      // -- UI

      this.sizer = new VerticalSizer;

      var treeBox = new TreeBox(this);
      this.sizer.add(treeBox);

      treeBox.rootDecoration = false;
      treeBox.numberOfColumns = 2;
      treeBox.multipleSelection = false;
      treeBox.headerVisible = false;
      treeBox.headerSorting = false;

      treeBox.sort(0,sorted); // DO NOT SEEMS TO WORK

      treeBox.style = Frame.FrameStyleSunken;

      treeBox.toolTip = toolTip;


      // -- Model update methods
      // The list model is changed (the model is an array that is updated in place)
      this.modelListChanged = function(newModelList) {

         // Clear current list display
         var i;
         var nmbNodes = treeBox.numberOfChildren;
         for (i=nmbNodes; i>0; i--) {
            treeBox.remove(i-1);
         }
         // Update variable tracking current model
         listModel = newModelList;

         // Add new nodes
         for (i=0; i<listModel.length; i++) {
            // Just making the nodes add them to the treeBox
            makeNode(treeBox, listModel[i], i);
         }
      }
      // Create initial model
      this.modelListChanged(initialListModel);

     // The current values of the current row model changed
     this.currentModelElementChanged = function() {
        if (treeBox.selectedNodes.length>0) {
           var selectedNode = treeBox.selectedNodes[0];
           var selectedIndex = treeBox.childIndex(selectedNode);
           if (selectedIndex>=0 && selectedIndex<listModel.length) {
              for (var i=0; i<modelDescription.length;i++){
                 selectedNode.setText(i, listModel[selectedIndex][modelDescription[i].propertyName].toString());
              }
            }
        }
      }


     // -- Internal actions
     var upAction = function() {
        var nodeToMove,  nodeIndex, element1, element2;
        if (treeBox.selectedNodes.length>0) {
           nodeToMove = treeBox.selectedNodes[0];
           nodeIndex = treeBox.childIndex(nodeToMove);
           if (nodeIndex>0) {
              // Update model
              element1 = listModel[nodeIndex-1];
              element2 = listModel[nodeIndex];
              listModel.splice(nodeIndex-1, 2, element2, element1);
              // update UI
              treeBox.remove(nodeIndex);
              treeBox.insert(nodeIndex-1,nodeToMove);
              treeBox.currentNode = nodeToMove;
           }
        }
     }
     var downAction = function() {
        var nodeToMove,  nodeIndex, element1, element2;
        if (treeBox.selectedNodes.length>0) {
           nodeToMove = treeBox.selectedNodes[0];
           nodeIndex = treeBox.childIndex(nodeToMove);
           if (nodeIndex<treeBox.numberOfChildren-1) {
              // Update model
              element1 = listModel[nodeIndex];
              element2 = listModel[nodeIndex+1];
              listModel.splice(nodeIndex, 2, element2, element1);
              // update UI
              treeBox.remove(nodeIndex);
              treeBox.insert(nodeIndex+1,nodeToMove);
              treeBox.currentNode = nodeToMove;
           }
        }
     }
     var addAction = function() {
        var nodeBeforeNew, newNode;
        var nodeIndex = treeBox.numberOfChildren;
        if (treeBox.selectedNodes.length>0) {
           nodeBeforeNew = treeBox.selectedNodes[0];
           nodeIndex = treeBox.childIndex(nodeBeforeNew) + 1;
        }
        // Create node
        var element = elementFactory();
        if (element !== null) {
           // insert node in model then ui
           listModel.splice(nodeIndex, 0, element);
           newNode = makeNode(treeBox, element, nodeIndex);
           treeBox.currentNode = newNode;
           treeBox.onNodeSelectionUpdated();
        }
     }

     var deleteAction = function() {
        var nodeToDelete,  nodeIndex;
        if (treeBox.selectedNodes.length>0) {
           nodeToDelete = treeBox.selectedNodes[0];
           nodeIndex = treeBox.childIndex(nodeToDelete);
           listModel.splice(nodeIndex,1);
           treeBox.remove(nodeIndex);
        }
     }

     treeBox.onNodeSelectionUpdated  = function() {
        // There could be an empty selection (last element removed)
        if (treeBox.selectedNodes.length>0) {
           var selectedNode = treeBox.selectedNodes[0];
           var selectedIndex = treeBox.childIndex(selectedNode);
           if (selectedIndex>=0 && selectedIndex<listModel.length) {
              selectionCallback(listModel[selectedIndex]);
            }
         }
     }


     var buttons = [
        {
           icon: ":/images/interface/prevButton.png",
           toolTip: "Move item up",
           action: upAction
           },
        {
           icon: ":/images/interface/nextButton.png",
           toolTip:  "Move item down",
           action: downAction
        },
        {
           icon: ":/images/icons/add.png",
           toolTip:  "Add new item",
           action: addAction
        },
        {
           icon: ":/images/icons/cancel.png",
           toolTip:  "Delete item",
           action: deleteAction
        },
     ];


     var iconButtonBar = new IconButtonBar(this, buttons);
     this.sizer.add(iconButtonBar);

     //this.adjustToContents();


  }
  ManagedList_Box.prototype = new Control;


  return {
     IconButtonBar: IconButtonBar,
     ManagedList_Box: ManagedList_Box,
  }

}) ();





// ========================================================================================================================
// ffM_GUI_config - Configuration dialog
//    The Dialog used to update a ConfigurationSet
// Usage:
//    A factory method is exposed to create the Dialog
//    Before being executed it must be configured with the ConfigurationSet and the name of the configuration
//       to select at start.
//    The dialog will make a local copy of the Configuration object and update it in place. It will leave
//    data for all resolvers selected for a variable, so in case a resolver is changed and then changed back to
//    the initial value, the original data is recovered.  It is exepected that this redundant data will be
//    removed before export.
//    The dialog expose properties to get the updated ConfigurationSet and currently edited set,
//    the caller should use them in case of successful return (and ignored them in case of cancel).
// ========================================================================================================================

var ffM_GUI_config = (function (){

   // -- Private methods
   function makeOKCancel(parentDialog) {
      var cancel_Button, ok_Button;

      var c = new Control(parentDialog);
      c.sizer = new HorizontalSizer;
      c.sizer.margin = 6;
      c.sizer.spacing = 4;

      cancel_Button = new PushButton( c );
      cancel_Button.text = "Cancel";
      cancel_Button.enabled = true;
      cancel_Button.onClick = function() {
         parentDialog.cancel();
      }
      ok_Button = new PushButton( c );
      ok_Button.text = "OK";
      ok_Button.enabled = true;
      ok_Button.onClick = function() {
         parentDialog.ok();
      }

      c.sizer.addStretch();
      c.sizer.add(cancel_Button);
      c.sizer.add(ok_Button);

      return c;

  }

  // ---------------------------------------------------------------------------------------------------------

  // Top pane - Selection of configuration
  function ConfigurationSelection_ComboBox (parent, initialNames, configurationSelectedCallback) {
      this.__base__ = ComboBox;
      this.__base__(parent);

      this.configurationNames = initialNames;

      var i;

      // -- UI
      this.toolTip = Text.H.SELECT_CONFIGURATION_BUTTON_TOOLTIP;
      this.enabled = true;
      this.editEnabled = false;
      for (i=0; i<initialNames.length;i++) {
         this.addItem(initialNames[i]);
      }
      if (this.configurationNames.length>0) {
         this.currentItem = 0;
      }

      // -- callback
      this.onItemSelected = function() {
         if (this.currentItem>=0 && this.currentItem<this.configurationNames.length) {
            configurationSelectedCallback(this.configurationNames[this.currentItem]);
         }
      }

      // -- Model update method, provide now list of names and new current name
      this.configure = function(names, selectedName) {
         this.configurationNames = names;
         this.clear();
         for (i=0; i<names.length;i++) {
            this.addItem(names[i]);
            if (names[i] === selectedName) {
               this.currentItem = i;
            }
         }
      }
  }
  ConfigurationSelection_ComboBox.prototype = new ComboBox;



  // Helper to validate and normalize input text,
  // There is no real conversion, as we keep all information in text format
  // the toString() is just to avoid crash and help debug in case a non string object is received
  var testInvalidVariableNameRegExp = /[&\(\);<>=!%*]/;
  var removeVariableReferencesRE = /&[0-9]+;/g
  var propertyTypes = {
     FREE_TEXT: {
        name: "FREE_TEXT",
        propertyToText: function(value) {return value.toString()},
        textToProperty: function(text) {return text},
     },
     REG_EXP: {
        // Check that this is a valid regular expression
        name: "REG_EXP",
        propertyToText: function(value) {return value.toString()},
        textToProperty: function(text) {return regExpToString(regExpFromUserString(text))},
     },
     REG_EXP_REPLACMENT: {
        // Check that this is a valid replacement for a regular expression,
        // this ensures that the only &<number>; are used (no &variable; or dangling &)
        name: "REG_EXP_REPLACMENT",
        propertyToText: function(value) {return value.toString()},
        textToProperty: function(text) {
           var withoutRef = text.replace(removeVariableReferencesRE,'');
           if (withoutRef.indexOf("&")>=0) {
               throw "Invalid replacement string"}
            else {return text}
        },
     },
     // Check that the characters are valid for a variable name
     VAR_NAME: {
        name: "VAR_NAME",
        propertyToText: function(value) {return value.toString()},
        textToProperty: function(text) {
            var t = text.trim();
            if (testInvalidVariableNameRegExp.test(t)) {
               throw "Invalid character in variable name"}
            else {return t}
         },
     },

  }



  // Utility pane - A Label - Labelled text field of a property of an object
  //  parent: UI parent control
  //  style: Style related properties (minLabelWidth, minDataWidth)
  //  name: Text of the label of this text field
  //  toolTip: the tooltip to present to the user
  //  property: Name of property of the target that will be used as source of text and destination of text
  //  propertyType: An object with two functions:
  //     propertyToText: that format the property as text
  //     textToProperty: that parse the text and return the property value, throw an exception in case of error
  //  valueChangedCallback(): function that will be called if the text is successfuly updated - if null there is no callback
  // The target (the object containing the property to edit) is specified dynamically (including at initialization)
  function TextEntryRow(parent, style, name, toolTip, property, propertyType, valueChangedCallback) {
     this.__base__ = Control;
     this.__base__(parent);
     var that = this;

     this.sizer = new HorizontalSizer;
     this.sizer.margin = 2;
     this.sizer.spacing = 2;

     this.property = property;
     this.target = null;

     var the_Label = new Label( this );
     this.sizer.add(the_Label);
     the_Label.textAlignment = TextAlign_Right|TextAlign_VertCenter;
     the_Label.minWidth = style.minLabelWidth;
     the_Label.text = name + ": ";
     the_Label.toolTip = toolTip;

     var text_Edit = new Edit(this);
     text_Edit.minWidth = style.minDataWidth;
     this.sizer.add(text_Edit)
     text_Edit.toolTip = toolTip;

     text_Edit.onTextUpdated = function() {
        var value;
        if (that.target !== null) {
           //Log.debug("TextEntryRow: onTextUpdated:",property,this.text);
           try {
              value =  propertyType.textToProperty(this.text);
              // Next lines will only execute if value was correctly parsed
              text_Edit.textColor = 0x000000;
              that.target[property] = value;
              if (valueChangedCallback != null) {
                 valueChangedCallback();
              }
           } catch (error) {
              text_Edit.textColor = 0xFF0000;
           }
        }
     }
     // Define the target object (that must have the property defined originally), null disables input
     this.updateTarget = function(target) {
        that.target = target;
        if (target === null) {
           text_Edit.text = '';
           text_Edit.enabled = false;
        } else {
           if (! target.hasOwnProperty(property)) {
              throw "Entry '" + name + "' does not have property '" + property + "': " + Log.pp(target);
           }
           text_Edit.text = propertyType.propertyToText(target[property]);
           text_Edit.enabled = true;
        }
     }

     this.updateTarget(null);
  }
  TextEntryRow.prototype = new Control;

   // TODO Refactor with Text Entry Row
  function BooleanEntryRow(parent, style, name, toolTip, property, valueChangedCallback) {
     this.__base__ = Control;
     this.__base__(parent);
     var that = this;

     this.sizer = new HorizontalSizer;
     this.sizer.margin = 2;
     this.sizer.spacing = 2;

     this.property = property;
     this.target = null;

     var the_Label = new Label( this );
     this.sizer.add(the_Label);
     the_Label.textAlignment = TextAlign_Right|TextAlign_VertCenter;
     the_Label.minWidth = style.minLabelWidth;
     the_Label.text = name + ": ";
     the_Label.toolTip = toolTip;

     var bool_CheckBox = new CheckBox(this);
     bool_CheckBox.minWidth = style.minDataWidth;
     this.sizer.add(bool_CheckBox)
     bool_CheckBox.toolTip = toolTip;

     bool_CheckBox.onCheck = function() {
        var value;
        if (that.target !== null) {
             value =  this.checked;
            bool_CheckBox.textColor = 0x000000;
            that.target[property] = value;
            if (valueChangedCallback != null) {
               valueChangedCallback();
            }
        }
     }
     // Define the target object (that must have the property defined originally), null disables input
     this.updateTarget = function(target) {
        that.target = target;
        if (target === null) {
           bool_CheckBox.checked = false;
           bool_CheckBox.enabled = false;
        } else {
           if (! target.hasOwnProperty(property)) {
              throw "Entry '" + name + "' does not have property '" + property + "': " + Log.pp(target);
           }
           bool_CheckBox.checked = target[property];
           bool_CheckBox.enabled = true;
        }
     }

     this.updateTarget(null);
  }
  BooleanEntryRow.prototype = new Control;


   // TODO Refactor with Text Entry Row
  function CheckListEntryRow(parent, style, name, toolTip, property, checkNames, checkValues, valueChangedCallback) {
     this.__base__ = Control;
     this.__base__(parent);
     var that = this;

     var i;

     this.sizer = new HorizontalSizer;
     this.sizer.margin = 2;
     this.sizer.spacing = 2;

     this.property = property;
     this.target = null;

     var the_Label = new Label( this );
     this.sizer.add(the_Label);
     the_Label.textAlignment = TextAlign_Right|TextAlign_VertCenter;
     the_Label.minWidth = style.minLabelWidth;
     the_Label.text = name + ": ";
     the_Label.toolTip = toolTip;

     var checkBoxContainer = new Control(this);
     this.sizer.add(checkBoxContainer);
     checkBoxContainer.sizer = new HorizontalSizer;
     var checkBoxes = [];

      for (var i=0; i<checkNames.length; i++) {

         var check_Label = new Label( this );
         checkBoxContainer.sizer.add(check_Label);
         check_Label.textAlignment = TextAlign_Right|TextAlign_VertCenter;
         check_Label.text = checkNames[i];
         checkBoxContainer.sizer.addSpacing(4);

         var bool_CheckBox = new RadioButton(checkBoxContainer);
         checkBoxContainer.sizer.add(bool_CheckBox);
         checkBoxes.push(bool_CheckBox);
         bool_CheckBox.toolTip = toolTip;
         bool_CheckBox.code = checkValues[i];
         checkBoxContainer.sizer.addSpacing(4);

         bool_CheckBox.onCheck = function() {
            var value;
            if (that.target !== null) {
               value =  this.code;
               bool_CheckBox.textColor = 0x000000;
               that.target[property] = value;
               if (valueChangedCallback != null) {
                  valueChangedCallback();
               }
            }
         }
     }
     checkBoxContainer.sizer.addStretch();

     // Define the target object (that must have the property defined originally), null disables input
     this.updateTarget = function(target) {
        that.target = target;
        if (target === null) {
            for (var i=0; i<checkBoxes.length;i ++) {
               checkBoxes[i].checked = false;
               checkBoxes[i].enabled = false;
            }
        } else {
            if (! target.hasOwnProperty(property)) {
               throw "Entry '" + name + "' does not have property '" + property + "': " + Log.pp(target);
            }
            var code = target[property];
            for (var i=0; i<checkBoxes.length;i ++) {
               checkBoxes[i].checked = (code === checkValues[i]);
               checkBoxes[i].enabled = true;
            }
        }
     }

     this.updateTarget(null);
  }
  CheckListEntryRow.prototype = new Control;


  // -- Middle right sub-panes (components to edit variables)
  function ResolverSelection_ComboBox(parent, mappingNames, mappingSelectionCallback) {
     this.__base__ = ComboBox;
     this.__base__(parent);

     var i;

     // -- UI
     this.toolTip = Text.H.VARIABLE_RESOLVER_TOOLTIP;
     this.enabled = true;
     this.editEnabled = false;
     for (i=0; i<mappingNames.length;i++) {
        this.addItem(mappingNames[i]);
     }
     this.currentItem = 0;
     this.enabled = false;

     // -- Call backs
     this.onItemSelected = function() {
        if (this.currentItem>=0) {
           mappingSelectionCallback(mappingNames[this.currentItem]);
        }
     }

     // -- Update model
     this.selectResolver = function(name) {
        if (name == null) {
              this.enabled = false;
        } else {
           for (i=0; i<mappingNames.length;i++) {
              if (name === mappingNames[i]) {
                 this.currentItem = i;
                 this.enabled = true;
                 break;
              }
           }
        }
     }
  }
  ResolverSelection_ComboBox.prototype = new ComboBox

  function ResolverSelectionRow(parent, rowStyle, name, mappingNames, mappingSelectionCallback) {
      this.__base__ = Control;
      this.__base__(parent);

      // -- UI
      this.sizer = new HorizontalSizer;
      this.sizer.margin = 2;
      this.sizer.spacing = 2;

      var the_Label = new Label( this );
      this.sizer.add(the_Label);
      the_Label.textAlignment = TextAlign_Right|TextAlign_VertCenter;
      the_Label.minWidth = rowStyle.minLabelWidth;
      the_Label.text = name + ": ";

      var resolver_ComboBox = new ResolverSelection_ComboBox (parent, mappingNames, mappingSelectionCallback);
      resolver_ComboBox.minWidth = rowStyle.minDataWidth;
      this.sizer.add(resolver_ComboBox);

      // -- Update model
      this.selectResolver = function(name) {
         resolver_ComboBox.selectResolver(name);
      }
  }
  ResolverSelectionRow.prototype = new Control;


  // ..............................................................................................
  // -- Controls for resolver
  // ..............................................................................................

  function MapFirstRegExpControl(parent, resolverName, rowStyle) {
     this.__base__ = Control;
     this.__base__(parent);
     var that = this;

     this.resolverName = resolverName;

     this.sizer = new VerticalSizer;

     // FITS Key
     var keyRow = new TextEntryRow(this, rowStyle, "FITS key", "Enter the name of a FITS key that will provide the value",
               "key", propertyTypes.FREE_TEXT, null);
     this.sizer.add(keyRow);


     var transformationDefinitionFactory = function() {
        return {regexp: /.*/, replacement: '?'}
     }

     var transformationSelectionCallback = function(transformationModel) {
        that.selectTransformationToEdit(transformationModel);
     }

     var regExpListSelection_GroupBox = new GroupBox(this);
     this.sizer.add(regExpListSelection_GroupBox);
     regExpListSelection_GroupBox.title ="List of RegExp->value";
     regExpListSelection_GroupBox.sizer = new VerticalSizer;

     var selectionLayoutControl = new Control;
     regExpListSelection_GroupBox.sizer.add(selectionLayoutControl);
     selectionLayoutControl.sizer = new HorizontalSizer;
     selectionLayoutControl.sizer.addStretch();

     var regExpListSelection_Box = new ffM_GUI_support.ManagedList_Box(
           selectionLayoutControl,
           [{propertyName: 'regexp'},{propertyName: 'replacement'}],
           [], // Its model will be initialized dynamically
           transformationDefinitionFactory,
           "Regular expression to text mapping.\nSelect a row to edit the details.\n"+
           "The first matching expression will define the value of the variable.\n"+
           "If none match, the variable is undefined. Use '/.*/' as the last expression to have a default value\n"+
           "&0; etc may be used to refer to the result of the regular expression",
           transformationSelectionCallback,
           false // Keep in order
     );
     selectionLayoutControl.sizer.add(regExpListSelection_Box);


     this.currentRegExpRow = new TextEntryRow(regExpListSelection_GroupBox, rowStyle, "Regexp",
        "A regular expression that will be tested against the key value", "regexp",
        propertyTypes.REG_EXP,
        function() {regExpListSelection_Box.currentModelElementChanged()});
     regExpListSelection_GroupBox.sizer.add(this.currentRegExpRow);

     this.currentReplacementRow = new TextEntryRow(regExpListSelection_GroupBox,rowStyle, "Replacement",
        "The replacement text to use if the regular expression matched.\n" +
        "&0; may be used to refer to the original text, &1;, &2; refers to parenthesized groups in the regular expression.", "replacement",
        propertyTypes.REG_EXP_REPLACMENT,
        function() {regExpListSelection_Box.currentModelElementChanged()});
     regExpListSelection_GroupBox.sizer.add(this.currentReplacementRow);


     this.selectTransformationToEdit = function(transformationModel) {
        this.currentRegExpRow.updateTarget(transformationModel);
        this.currentReplacementRow.updateTarget(transformationModel);
     }

     this.initialize = function(variableDefinition) {
        if (!variableDefinition.parameters.hasOwnProperty(resolverName)) {
           variableDefinition.parameters[resolverName] = deepCopyData(ffM_Resolver.resolverByName(resolverName).initial);
        }
     }

     this.populate = function(variableDefinition) {
        // initialize should probably be somewhere else
        this.initialize(variableDefinition);
        keyRow.updateTarget(variableDefinition.parameters[resolverName]);
        regExpListSelection_Box.modelListChanged(variableDefinition.parameters[resolverName].reChecks);
        // Workaround:
        // The modelListChanged above seem to generate callbacks events that update the regexp
        // and replacement row, but we want them to be disabled until the user explicitely select one.
        this.currentRegExpRow.updateTarget(null);
        this.currentReplacementRow.updateTarget(null);
     }

     this.leave = function() {
        this.currentRegExpRow.updateTarget(null);
        this.currentReplacementRow.updateTarget(null);
     }
  }
  MapFirstRegExpControl.prototype = new Control;

  // ..............................................................................................

  function ConstantValueResolverControl(parent, resolverName, rowStyle) {
     this.__base__ = Control;
     this.__base__(parent);

     this.resolverName = resolverName;

     this.sizer = new VerticalSizer;
     var constantValueRow = new TextEntryRow(this, rowStyle, "Value",
     "The fixed value for this variable",
     "value", propertyTypes.FREE_TEXT, null);
     this.sizer.add(constantValueRow);

     this.initialize = function(variableDefinition) {
        if (!variableDefinition.parameters.hasOwnProperty(resolverName)) {
           variableDefinition.parameters[resolverName] = deepCopyData(ffM_Resolver.resolverByName(resolverName).initial);
        }
     }

     this.populate = function(variableDefinition) {
        // Should probably be somewhere else
        this.initialize(variableDefinition);
        constantValueRow.updateTarget(variableDefinition.parameters[resolverName]);
     }
     this.leave = function() {
        // Nothing the clean
     }
  }
  ConstantValueResolverControl.prototype = new Control;

  // ..............................................................................................

  function TextValueResolverControl(parent, resolverName, rowStyle) {
     this.__base__ = Control;
     this.__base__(parent);

     this.resolverName = resolverName;

     this.sizer = new VerticalSizer;

     // FITS Key
     var keyRow = new TextEntryRow(this, rowStyle, "FITS key",
      "The name of a FITS key that will provide the value of this variable\n" +
      "(the FITS key value will be cleaned of special characters).",
      "key", propertyTypes.FREE_TEXT, null);
     this.sizer.add(keyRow);

     var formatRow = new TextEntryRow(this, rowStyle, "Format",
     "A valid C format string to display the value, for example '-%ls' to preceed the string with a dash\n"+
     "IMPORTANT - You must use '%ls', not '%s' to indicate the location of the string in the format",
     "format", propertyTypes.FREE_TEXT, null);
     this.sizer.add(formatRow);

     var caseRuleRow = new CheckListEntryRow(this, rowStyle, "Case conversion",
     "Case conversion",
     "case",['up','down','none'], ['UP','DOWN','NONE'], null);
     this.sizer.add(caseRuleRow);


     this.initialize = function(variableDefinition) {
        if (!variableDefinition.parameters.hasOwnProperty(resolverName)) {
           variableDefinition.parameters[resolverName] = deepCopyData(ffM_Resolver.resolverByName(resolverName).initial);
        }
     }

     this.populate = function(variableDefinition) {
        // Should probably be somewhere else
        this.initialize(variableDefinition);
        keyRow.updateTarget(variableDefinition.parameters[resolverName]);
        formatRow.updateTarget(variableDefinition.parameters[resolverName]);
        caseRuleRow.updateTarget(variableDefinition.parameters[resolverName]);
     }
     this.leave = function() {
        // Nothing the clean
     }
  }
  TextValueResolverControl.prototype = new Control;

  // ..............................................................................................

  function IntegerValueResolverControl(parent, resolverName, rowStyle) {
     this.__base__ = Control;
     this.__base__(parent);

     this.resolverName = resolverName;

     this.sizer = new VerticalSizer;

     // FITS Key
     var keyRow = new TextEntryRow(this, rowStyle, "FITS key",
      "The name of a FITS key that will provide the value",
      "key", propertyTypes.FREE_TEXT, null);
     this.sizer.add(keyRow);

     var useAbsoluteRow = new BooleanEntryRow(this, rowStyle, "Use absolute value",
        "If true, the absolute value of the corresponding value is taken (the sign is discarded)",
        "abs", null);
     this.sizer.add(useAbsoluteRow);

     var formatRow = new TextEntryRow(this, rowStyle, "Format",
     "A valid C format string to display the value, like '%4.4d', possibly with additional text like 'TEMP-%3d'",
     "format", propertyTypes.FREE_TEXT, null);
     this.sizer.add(formatRow);

     this.initialize = function(variableDefinition) {
        if (!variableDefinition.parameters.hasOwnProperty(resolverName)) {
           variableDefinition.parameters[resolverName] = deepCopyData(ffM_Resolver.resolverByName(resolverName).initial);
        }
     }

     this.populate = function(variableDefinition) {
        // Should probably be somewhere else
        this.initialize(variableDefinition);
        keyRow.updateTarget(variableDefinition.parameters[resolverName]);
        useAbsoluteRow.updateTarget(variableDefinition.parameters[resolverName]);
        formatRow.updateTarget(variableDefinition.parameters[resolverName]);
     }
     this.leave = function() {
        // Nothing the clean
     }
  }
  IntegerValueResolverControl.prototype = new Control;


  // ..............................................................................................

  function IntegerPairValueResolverControl(parent, resolverName, rowStyle) {
     this.__base__ = Control;
     this.__base__(parent);

     this.resolverName = resolverName;

     this.sizer = new VerticalSizer;

     // FITS Key 1
     var key1Row = new TextEntryRow(this, rowStyle, "FITS key 1",
      "The name of a FITS key that will provide the first value",
      "key1", propertyTypes.FREE_TEXT, null);
     this.sizer.add(key1Row);

     var key2Row = new TextEntryRow(this, rowStyle, "FITS key 2",
      "The name of a FITS key that will provide the second value",
      "key2", propertyTypes.FREE_TEXT, null);
     this.sizer.add(key2Row);

     var formatRow = new TextEntryRow(this, rowStyle, "Format",
     "A valid C format string to display the 2 values, for example '%dx%d'",
     "format", propertyTypes.FREE_TEXT, null);
     this.sizer.add(formatRow);

     this.initialize = function(variableDefinition) {
        if (!variableDefinition.parameters.hasOwnProperty(resolverName)) {
           variableDefinition.parameters[resolverName] = deepCopyData(ffM_Resolver.resolverByName(resolverName).initial);
        }
     }

     this.populate = function(variableDefinition) {
        // Should probably be somewhere else
        this.initialize(variableDefinition);
        key1Row.updateTarget(variableDefinition.parameters[resolverName]);
        key2Row.updateTarget(variableDefinition.parameters[resolverName]);
        formatRow.updateTarget(variableDefinition.parameters[resolverName]);
     }
     this.leave = function() {
        // Nothing the clean
     }
  }
  IntegerPairValueResolverControl.prototype = new Control;

  // ..............................................................................................

  function NightResolverControl(parent, resolverName, rowStyle) {
     this.__base__ = Control;
     this.__base__(parent);

     this.resolverName = resolverName;

     this.sizer = new VerticalSizer;

     var key1Row = new TextEntryRow(this, rowStyle, "LONG_OBS key",
     "The name of a FITS key that provide the longitude of the observatory",
     "keyLongObs", propertyTypes.FREE_TEXT, null);
     this.sizer.add(key1Row);

     var key2Row = new TextEntryRow(this, rowStyle, "JD key",
     "The name of a FITS key that contains the julian date of the observation",
     "keyJD", propertyTypes.FREE_TEXT, null);
     this.sizer.add(key2Row);


     this.initialize = function(variableDefinition) {
        if (!variableDefinition.parameters.hasOwnProperty(resolverName)) {
           variableDefinition.parameters[resolverName] = deepCopyData(ffM_Resolver.resolverByName(resolverName).initial);
        }
     }

     this.populate = function(variableDefinition) {
        // Should probably be somewhere else
        this.initialize(variableDefinition);
        key1Row.updateTarget(variableDefinition.parameters[resolverName]);
        key2Row.updateTarget(variableDefinition.parameters[resolverName]);
     }
     this.leave = function() {
        // Nothing the clean
     }
  }
  NightResolverControl.prototype = new Control;




  // ..............................................................................................


  // -- Middle pane - Variable definitions (two panes: add/remove variables and edit selected variables)
  function VariableUIControl(parent, variableDefinitionFactory ) {
     this.__base__ = Control;
     this.__base__(parent);
     var that = this;

     // -- Model data - set by selectVariable()
     this.currentVariableDefinition = null;

     // -- UI data
     // this.resolverSelection_GroupBox = null; // defined later
     this.currentResolver = null;

     this.sizer = new HorizontalSizer;
     this.sizer.margin = 6;
     this.sizer.spacing = 4;

     var labelWidth = this.font.width( "MMMMMMMMMMMM: " );
     var dataWidth = this.font.width( "MMMMMMMMMMMMMMMMMMMMMMM" );
     var rowStyle = {
        minLabelWidth: labelWidth,
        minDataWidth: dataWidth,
     }


     // -- GUI action callbacks
     // Variable selected in current configuration, forward to the model handling later in this object
     var variableSelectionCallback = function(variableDefinition) {
        //Log.debug("VariableUIControl: variableSelectionCallback - Variable selected: " +variableDefinition.name);
        that.selectVariable(variableDefinition);
     }




     // -- Left side - select variable being edited, add/remove variables

     var variableListSelection_GroupBox = new GroupBox(this);
     this.sizer.add(variableListSelection_GroupBox);
     variableListSelection_GroupBox.title = "Select variable";
     variableListSelection_GroupBox.sizer = new VerticalSizer; // Any sizer

     var variableListSelection_Box = new ffM_GUI_support.ManagedList_Box(
           variableListSelection_GroupBox,
          [{propertyName: 'name'},{propertyName: 'description'}],
          [], // Its model will be initialized dynamically
           variableDefinitionFactory,
           Text.H.VARIABLE_SELECTION_TOOLTIP,
           variableSelectionCallback,
           true // Sort by variable name
     );
     variableListSelection_GroupBox.sizer.add(variableListSelection_Box);



     //--  Right side - Enter parameters corresponding to selected variables
     var resolverSelectionCallback = function(resolverName) {
#ifdef DEBUG
        debug("VariableUIControl: resolverSelectionCallback - selected:",resolverName);
#endif
        that.updateResolver(resolverName);
     }

     var variableDetails_GroupBox = new GroupBox(this);
     variableDetails_GroupBox.title = "Parameters of variable";
     this.sizer.add(variableDetails_GroupBox);

     variableDetails_GroupBox.sizer = new VerticalSizer;



     this.resolverSelectionRow =  new ResolverSelectionRow(this, rowStyle, "Resolver", ffM_Resolver.resolverNames, resolverSelectionCallback);
     variableDetails_GroupBox.sizer.add(this.resolverSelectionRow);

     this.variableNameRow = new TextEntryRow(this, rowStyle, "Variable name",
        "Enter the name of the variable, it will be used as '&amp;name;' in a template",
        "name",
        propertyTypes.VAR_NAME,
        function() {variableListSelection_Box.currentModelElementChanged()});
     variableDetails_GroupBox.sizer.add(this.variableNameRow);

     this.descriptionRow = new TextEntryRow(this, rowStyle, "Description",
        "Enter a short description of the variable",
        "description",
        propertyTypes.FREE_TEXT,
        function() {variableListSelection_Box.currentModelElementChanged()});
     variableDetails_GroupBox.sizer.add(this.descriptionRow);

     this.variableShownRow = new BooleanEntryRow(this, rowStyle, "Show column by default",
        "If true, the corresponding column is shown in the Input file list by default",
        "show",
        function() {variableListSelection_Box.currentModelElementChanged()});
     variableDetails_GroupBox.sizer.add(this.variableShownRow);

     var addNewResolverControl = function(resolverControl){
        variableDetails_GroupBox.sizer.add(resolverControl);
        ffM_Resolver.resolverByName(resolverControl.resolverName).control = resolverControl;
        resolverControl.hide();
     }

     // Make all resolver controls, only the selected one will be shown
     addNewResolverControl(new ConstantValueResolverControl(variableDetails_GroupBox, 'Constant', rowStyle));

     addNewResolverControl(new TextValueResolverControl(variableDetails_GroupBox, 'Text', rowStyle));

     addNewResolverControl(new IntegerValueResolverControl(variableDetails_GroupBox, 'Integer', rowStyle));

     addNewResolverControl(new IntegerPairValueResolverControl(variableDetails_GroupBox, 'IntegerPair', rowStyle));

     addNewResolverControl(new MapFirstRegExpControl(variableDetails_GroupBox,'RegExpList', rowStyle));

     addNewResolverControl(new NightResolverControl(variableDetails_GroupBox,'Night', rowStyle));

     // FileName and FileExtension have no parameter
     ffM_Resolver.resolverByName('FileName').control = null
     ffM_Resolver.resolverByName('FileExtension').control = null

     variableDetails_GroupBox.sizer.addStretch();

     // TRICK - void growing when more proeprties are added to the right pane, the dialg can grow,
     // so it is possible to avoid this minimum size, but the effect is likely uggly.
     variableDetails_GroupBox.setMinHeight(GROUP_BOX_MIN_SIZE * this.font.lineSpacing + 2*this.sizer.margin);



     // -- Update the model
     // The resolver name was updated (by select box, by changing variable or initially)
     // null if no resolver (like an empty variable list)
     this.updateResolver = function(resolverName) {
        //Log.debug(" updateResolver",resolverName,that.currentResolver);
        // There could be no controller if there is no parameter
        if (that.currentResolver != null && that.currentResolver.control !== null) {
           that.currentResolver.control.leave();
           that.currentResolver.control.hide();
        }
        that.currentResolver =null;
        if (resolverName != null) {
           var resolver = ffM_Resolver.resolverByName(resolverName);
           if (resolver === null) {
              throw "Invalid resolver '" + resolverName + "' for variable '"+ variableDefinition.name+"'";
           }
           if (that.currentVariableDefinition !== null) {
              // record the new resolver
              that.currentVariableDefinition.resolver = resolverName;
              // Populate and show it
              that.currentResolver = resolver;
              if (resolver.control !== null) {
                 resolver.control.populate(that.currentVariableDefinition );
                 resolver.control.show();
              }
           }
        }
     }

     // The variable to edit was selected
     this.selectVariable = function(variableDefinition) {
#ifdef DEBUG
        debug("VariableUIControl: selectVariable - Variable selected ",variableDefinition.name );
#endif
         var resolverName;

        that.currentVariableDefinition = variableDefinition;

        // populate the common fields
        this.variableNameRow.updateTarget(variableDefinition);
        this.descriptionRow.updateTarget(variableDefinition);
        this.variableShownRow.updateTarget(variableDefinition);

        // Find new resolver, populate and show it
        resolverName = that.currentVariableDefinition.resolver;
        that.updateResolver(resolverName);

        // Update the UI to have the resolver type of the current variable
        this.resolverSelectionRow.selectResolver(resolverName);
      }

     // The list of variables was changed externally (initial or change)
     this.updateVariableList = function(newVariableList) {
        // Update UI
        variableListSelection_Box.modelListChanged(newVariableList);
     }

  }

  VariableUIControl.prototype = new Control;


   function ConfigurationGroupBox(parent, configurationSelectedCallback, configurationDuplicateCallback, configurationDeleteCallback ) {
      this.__base__ = GroupBox;
      this.__base__(parent);
      var that = this;

      // Initialized by configure()
      this.editedConfigurationSet = null;

      var configurationNameMinWidth = this.font.width( "MMMMMMMMMMMMMMMMM: " );
      this.title = "Configuration selection";

      this.sizer = new VerticalSizer;
      this.sizer.margin = 5;
      this.sizer.spacing = 5;

      var nameRow = new Control(this);
      this.sizer.add(nameRow);

      nameRow.sizer = new HorizontalSizer;

      nameRow.sizer.margin = 1;
      nameRow.sizer.spacing = 2;
      this.configurationSelection_ComboBox = new ConfigurationSelection_ComboBox(nameRow, [], configurationSelectedCallback);
      nameRow.sizer.add(this.configurationSelection_ComboBox);
      this.configurationSelection_ComboBox.minWidth = configurationNameMinWidth;


      nameRow.sizer.addStretch();

      var name_Label = new Label( nameRow );
      nameRow.sizer.add(name_Label);
      name_Label.textAlignment = TextAlign_Right|TextAlign_VertCenter;
      name_Label.text = " Name: ";

      this.configurationName_Edit = new Edit(nameRow);
      nameRow.sizer.add(this.configurationName_Edit);
      this.configurationName_Edit.minWidth = configurationNameMinWidth;
      this.configurationName_Edit.toolTip = "Name of configuration\nIllegal special characters makes the field red.";
      this.configurationName_Edit.onTextUpdated = function() {
         // We should always have a selected configuration at this point, but play it safe
         if (that.selectedConfiguration !== null) {
            var t = this.text.trim();
            var configurationNames = ffM_Configuration.getAllConfigurationNames(that.editedConfigurationSet);
            // Avoid duplicate or illegal names (the check is a hack, but this is java script after all)
            // Characters are limited to avoid problem in case the name is used in some expression later
            if (t.length >0 && !(/[()\[\]{}&$;!?'".,]/.test(t)) && configurationNames.indexOf(t)===-1) {
               that.updateName(t);
            } else if (that.currentConfigurationName !== t) {
               // Red only if the trimmed value is not the current one
               this.textColor=0x00FF0000;
            } else {
               // Back to the current one, assumne legal
               this.textColor=0x00000000;
            }
         }
      }

      var addConfigurationButton = new ToolButton( nameRow );
      nameRow.sizer.add(addConfigurationButton);
      addConfigurationButton.icon = new Bitmap( ":/images/icons/copy.png" );
      addConfigurationButton.toolTip = "Add a configuration (duplicate current one)";
      addConfigurationButton.onClick = function() {
         configurationDuplicateCallback(that.currentConfigurationName);
      }

      var removeConfigurationButton = new ToolButton( nameRow );
      nameRow.sizer.add(removeConfigurationButton);
      removeConfigurationButton.icon = new Bitmap( ":/images/icons/cancel.png" );
      removeConfigurationButton.toolTip = "Delete the current configuration";
      removeConfigurationButton.onClick = function() {
         var msg = new MessageBox( "Do you want to delete the configuration '" +that.currentConfigurationName + "' ?",
                          "Are you sure?", StdIcon_Question, StdButton_Yes, StdButton_No );
         if (msg.execute() == StdButton_Yes) {
          configurationDeleteCallback(that.currentConfigurationName);
         }
      }

      var descriptionRow = new Control(this);
      this.sizer.add(descriptionRow);
      descriptionRow.sizer = new HorizontalSizer;

      var description_Label = new Label( descriptionRow );
      descriptionRow.sizer.add(description_Label);
      description_Label.textAlignment = TextAlign_Right|TextAlign_VertCenter;
      description_Label.text = " Description: ";

      descriptionRow.sizer.margin = 1;
      descriptionRow.sizer.spacing = 2;
      this.configurationComment_Edit = new Edit(descriptionRow);
      descriptionRow.sizer.add(this.configurationComment_Edit);
      this.configurationComment_Edit.toolTip = "Short description of the configuration";

      descriptionRow.sizer.addSpacing(50); // Pretty approximate

      // To track edited comment
      this.selectedConfiguration = null;


      this.configure = function(editedConfigurationSet, currentConfigurationName) {
         this.editedConfigurationSet = editedConfigurationSet;
         this.currentConfigurationName = currentConfigurationName;
         var configurationNames = ffM_Configuration.getAllConfigurationNames(editedConfigurationSet);
         this.configurationSelection_ComboBox.configure(configurationNames, currentConfigurationName);
         this.selectedConfiguration = ffM_Configuration.getConfigurationByName(editedConfigurationSet, currentConfigurationName);
         if (this.selectedConfiguration === null) {
            throw "Internal Error - configuration name '" + currentConfigurationName + "' not in list of configurations";
         }
         this.configurationComment_Edit.text = this.selectedConfiguration.description;
         this.configurationName_Edit.text = currentConfigurationName;
         this.configurationName_Edit.textColor =0x00000000;
      }
      this.updateName = function(newConfigurationName) {
         // Assume the name is allowed and not duplicated (checked by caller),
         // set the new current name in our working copy (our selected configuration)
         this.selectedConfiguration.name = newConfigurationName;
         this.currentConfigurationName = newConfigurationName;
         // Update the whole configuration selection box with the new names
         var configurationNames = ffM_Configuration.getAllConfigurationNames(this.editedConfigurationSet);
         if (configurationNames.indexOf(newConfigurationName)===-1) {
            throw "Internal Error - new configuration name '" + newConfigurationName + "' not in list of configurations";
         }
         this.configurationSelection_ComboBox.configure(configurationNames, newConfigurationName);
         this.configurationName_Edit.textColor=0x00000000;
      }
      this.configurationComment_Edit.onTextUpdated = function() {
         if (that.selectedConfiguration !== null) {
            that.selectedConfiguration.description = this.text;
         }
      }
   }
   ConfigurationGroupBox.prototype = new GroupBox;

  // ---------------------------------------------------------------------------------------------------------
   function BuiltinVariableGroup(parent) {
      this.__base__ = GroupBox;
      this.__base__(parent);
      var that = this;

      this.title = "Built in variable parameters";
      this.sizer = new HorizontalSizer;

      var labelWidth = this.font.width( "MMMMMMMMMMMM: " );
      var dataWidth = this.font.width( "MMMMMMMMMMMMMMMMMMMMMMM" );
      var rowStyle = {
        minLabelWidth: labelWidth,
        minDataWidth: dataWidth,
      }

      //this.sizer.addStretch();

      // Left columns
      var colLayout1_Control = new Control;
      this.sizer.add(colLayout1_Control);
      colLayout1_Control.sizer = new VerticalSizer;

      var rankFormat = new TextEntryRow(colLayout1_Control, rowStyle, "&rank; format",
      "Enter a valid C format string for the &rank; value, like '%3.3d'\nYou can also add text around like 'N%3.3d'",
      "format", propertyTypes.FREE_TEXT, null);
      colLayout1_Control.sizer.add(rankFormat);

      // Right column
      var colLayout2_Control = new Control;
      this.sizer.add(colLayout2_Control);
      colLayout2_Control.sizer = new VerticalSizer;

      var countFormat = new TextEntryRow(colLayout2_Control, rowStyle, "&count format",
      "Enter a valid C format string for the &count; value, like '%3.3d'\nYou can also add text around like 'group-%d'",
      "format", propertyTypes.FREE_TEXT, null);
      colLayout2_Control.sizer.add(countFormat);


      this.configure = function(editedConfigurationSet, currentConfigurationName) {
         var configurationNames = ffM_Configuration.getAllConfigurationNames(editedConfigurationSet);
         this.selectedConfiguration = ffM_Configuration.getConfigurationByName(editedConfigurationSet, currentConfigurationName);

         rankFormat.updateTarget(this.selectedConfiguration.builtins.rank);
         countFormat.updateTarget(this.selectedConfiguration.builtins.count);
      }

   }
   BuiltinVariableGroup.prototype = new GroupBox();


  // ---------------------------------------------------------------------------------------------------------
  // This Dialog controls the update of a configurationSet, starting at a current configuration.
  // The configurationSet may be modified or not, at the end the caller must get
  // the configurationSet and new current configuration name properties from the dialog to get the current
  // state.
  // There is currently no way to add, delete or rename configurations
  function ConfigurationDialog( parentDialog) {
     this.__base__ = Dialog;
     this.__base__();
     var that = this;

     // Model -
     // Keeps track of configurationSet and current configuration selected
     // in a copy, so in case of cancel nothing is changed
     // Done in 'configure'
     this.editedConfigurationSet = null;
     this.currentConfigurationName = null;

     this.newVariableCounter = 0;

     this.windowTitle = Text.T.REMAPPING_SECTION_PART_TEXT;


     this.sizer = new VerticalSizer;
     this.sizer.margin = 6;
     this.sizer.spacing = 4;


     // -- GUI actions callbacks
     // configuration changed (also used in initialization)
     var configurationSelectedCallback = function(configurationName) {
#ifdef DEBUG
        debug("ConfigurationDialog: configurationSelectedCallback - ConfigurationSet selected:",configurationName);
#endif
        that.selectedConfiguration = ffM_Configuration.getConfigurationByName(that.editedConfigurationSet, configurationName);
        if (that.selectedConfiguration == null) {
           throw "PROGRAM ERROR - Invalid configuration set name '" + configurationName +"'";
        }

        // Update model
        that.currentConfigurationName = configurationName;
        // Update UI
        that.variableUI.updateVariableList(that.selectedConfiguration.variableList);
        that.builtinVariable_Group.configure(that.editedConfigurationSet, that.currentConfigurationName);
        // Update the description text
        that.configurationLayoutControl.configure(that.editedConfigurationSet, that.currentConfigurationName);
     }

     var deleteConfigurationCallback = function(currentConfigurationName) {
#ifdef DEBUG
        debug("ConfigurationDialog: deleteConfigurationCallback");
#endif
        var newConfigurationName = ffM_Configuration.removeConfigurationByName(that.editedConfigurationSet, currentConfigurationName);
        if (newConfigurationName == null) {
           Console.writeln("Configuration '" + currentConfigurationName  +"' not deleted, it is the last one");
        } else {
           configurationSelectedCallback(newConfigurationName);
        }
     }

     // We duplicate the current configuration
     var duplicateConfigurationCallback = function(configurationName) {
#ifdef DEBUG
        debug("ConfigurationDialog: duplicateConfigurationCallback");
#endif
        var selectedConfiguration = ffM_Configuration.getConfigurationByName(that.editedConfigurationSet, configurationName);
        var newConfiguration = deepCopyData(selectedConfiguration);
        newConfiguration.name = createUniqueName(configurationName,ffM_Configuration.getAllConfigurationNames(that.editedConfigurationSet));
        that.editedConfigurationSet.push(newConfiguration);
        configurationSelectedCallback(newConfiguration.name);
        Console.writeln("Configuration '" + newConfiguration.name  +"' created");
     }

     // -- Model call backs
     // New variable requested in current configuration, define one with default values
     var variableDefinitionFactory = function() {
#ifdef DEBUG
          debug("ConfigurationDialog: variableDefinitionFactory - create new variable");
#endif
          that.newVariableCounter++;
          return ffM_Configuration.defineVariable("new_" + that.newVariableCounter,'','Text');
     }

     // -- Build the top level pane

     // Top pane - select configuration to operate upon
     this.configurationLayoutControl = new ConfigurationGroupBox(this, configurationSelectedCallback, duplicateConfigurationCallback, deleteConfigurationCallback);
     this.sizer.add(this.configurationLayoutControl);


     // Middle pane - define variables, their resolvers and the resolver's parameters
     this.variableUI = new VariableUIControl(this, variableDefinitionFactory);
     this.sizer.add(this.variableUI);

     this.builtinVariable_Group = new BuiltinVariableGroup(this);
     this.sizer.add(this.builtinVariable_Group);

     // Bottom pane - buttons
     var okCancelButtons = makeOKCancel(this);
     this.sizer.add(okCancelButtons);

     this.setVariableSize();
     //this.adjustToContents();

     // -- Configure before executing
     this.configure = function configure(originalConfigurationSet, configurationNameToEdit) {
        this.editedConfigurationSet = deepCopyData(originalConfigurationSet);
        this.selectedConfiguration = ffM_Configuration.getConfigurationByName(this.editedConfigurationSet, configurationNameToEdit);
        if (this.selectedConfiguration === null) {
           throw "Internal Error - current configuration not found:'" + configurationNameToEdit + "'";
        }
        this.currentConfigurationName = configurationNameToEdit;
        var configurationNames = ffM_Configuration.getAllConfigurationNames(this.editedConfigurationSet);
        // Initialize content
        this.configurationLayoutControl.configure(this.editedConfigurationSet, this.currentConfigurationName);
        this.builtinVariable_Group.configure(this.editedConfigurationSet, this.currentConfigurationName);
        configurationSelectedCallback(this.currentConfigurationName);
     }



  }
  ConfigurationDialog.prototype = new Dialog;



  return {
     makeDialog: function(parent, configurationSet, configurationNameToEdit) {
        return new ConfigurationDialog(parent, configurationSet, configurationNameToEdit);
     },
     ConfigurationSelection_ComboBox: ConfigurationSelection_ComboBox,

  }

}) ();

//==== end of include ============================================
//================================================================
// include "FITSFileManager-gui.jsh"
//================================================================
// FITSFileManager-gui.js

// This file is part of FITSFileManager, see copyright in FITSFileManager.js




// All texts are in the module FITSFileManager-text.jsh


// Constants
var CompletionDialog_doneContinue = 0;
var CompletionDialog_doneKeep = 1;
var CompletionDialog_doneRemove = 2;
var CompletionDialog_doneLeave= 3;

// These sizes show depend on the screen size and font used

// Tree box for file lists minimum sizes, do not set too large to avoid problem on small screens
var TreeboxWindowMinWidth = 700;
var InputTreeboxMinHeight = 200;
var TransformTreeBoxMinHeight = 200;

// Main dialog minimum size, required on Linux to avoid overlap on small window
// Should be derived from sizes of tree box or autoadjust
var MainDialogMinimumWidth = 600;
var MainDialogMinimumHeight = 600;

// Prefered size may be larger than screen, should be reduced by PI,
// although this is usually not pretty
var MainDialogPreferedHeight = 700;
var MainDialogPreferedWidth = 800;

// Fits keys window:
// To make sure something reasonable is visible,
// if too large this may be a problem in small screens
var FitsKeysDialogMinimumHeight = 300;
var FitsKeysDialogMinimumWidth = 400;
// Make default more reasonable
var FitsKeysDialogPreferedHeight = 600;
var FitsKeysDialogPreferedWidth = 800;


// ------------------------------------------------------------------------------------------------------------------------
// Section Group - support to switch between SectionBar and a group box
// ------------------------------------------------------------------------------------------------------------------------


#ifdef USE_SECTION_BAR
function makeSectionGroup(parent, content, title, initialyCollapsed) {
   var section = new SectionBar( parent,  initialyCollapsed);
   section.setTitle( title );
   section.setSection( content );
   return section;
}
#else
var dummySectionPrototype = {
    setCollapsedTitle: function(title) { return; },
    setSection: function(section) { this.section = section; },
    addSection: function(sizer, weight) {
      sizer.add(this.section, weight);
   },
}
// The content is a group box
function makeSectionGroup(parent, content, title, initialyCollapsed) {
   // Return an object that fake a section bar, without any functionality,
   // show just the GroupBox with title
   var section = Object.create(dummySectionPrototype);
   content.title = title;
   section.setSection(content);
   return section;
}
#endif


// ------------------------------------------------------------------------------------------------------------------------
// SectionBar Control adapted from Juan: http://pixinsight.com/forum/index.php?topic=4610.msg32012#msg32012
// This does not work well on Mac, unfortunately
// ------------------------------------------------------------------------------------------------------------------------

#define contract_icon   new Bitmap( ":/images/icons/contract_v.png" )
#define expand_icon     new Bitmap( ":/images/icons/expand_v.png" )

function SectionBar( parent, initialyCollapsed ) {
   this.__base__ = Control;
   if ( parent )
      this.__base__( parent );
   else
      this.__base__();

   this.section = null;
   this.collapsedTitle = "";
   this.expandedTitle = "";

#ifgteq __PI_BUILD__ 854
   var bgColor = Settings.readGlobal( "InterfaceWindow/SectionBarColor", DataType_UInt32 );
   var fgColor = Settings.readGlobal( "InterfaceWindow/SectionBarTextColor", DataType_UInt32 );
#else
   // PJSR access to global settings is broken in PI 1.7
   var bgColor = Color.rgbaColor( 192, 192, 168, 255 );
   var fgColor = Color.rgbaColor(   0,   0, 255, 255 );
#endif

   this.backgroundColor = bgColor;
   this.focusStyle = FocusStyle_NoFocus;

   this.label = new Label( this );
   this.label.textAlignment = TextAlign_Left|TextAlign_VertCenter;
   this.label.styleSheet =
      "QLabel { color: " + Color.rgbColorToHexString( fgColor ) + "; " +
               "background: " + Color.rgbColorToHexString( bgColor ) + "; }" +
      "QLabel:disabled { color: gray; }";

   this.button = new ToolButton( this );
   this.button.icon = contract_icon;
   this.button.setFixedSize( 17, 17 );
   this.button.focusStyle = FocusStyle_NoFocus;
   this.button.onClick = function()
   {
      this.parent.toggleSection();
   };

   var hSizer = new HorizontalSizer;
   hSizer.addSpacing( 4 );
   hSizer.add( this.label );
   hSizer.addStretch();
   hSizer.add( this.button );
   hSizer.addSpacing( 4 );

   this.sizer = new VerticalSizer;
   this.sizer.addSpacing( 1 );
   this.sizer.add( hSizer );
   this.sizer.addSpacing( 1 );

   this.adjustToContents();
   this.setFixedHeight();

   // -- Private implementation

   this.onMousePress = function( x, y, button, buttonState, modifiers )
   {
      if ( button == MouseButton_Left )
         this.button.onClick();
   };

   this.onShow = function()
   {
      this.updateTitle();
      this.updateIcon();
   };

   this.toggleSection = function()
   {
      if ( this.section )
      {
         this.setFixedWidth();

         if ( this.section.visible ) {
            this.section.hide();
         } else {
            this.section.show();
         }
         this.updateTitle();
         this.updateIcon();

         // Original adjustToContent
         this.dialog.adjustToContents();

         this.setVariableWidth();

      }
   };

   this.updateTitle = function() {
      if ( this.section &&  this.section.visible ) {
         this.label.text = this.expandedTitle;
      } else {
         this.label.text = this.collapsedTitle;
      }
   }

   this.updateIcon = function()
   {
      if ( this.section )
         if ( this.section.visible )
            this.button.icon = contract_icon;
         else
            this.button.icon = expand_icon;
   };



   //---  Public interface

   this.setTitle = function( title )
   {
      this.expandedTitle = title;
      this.collapsedTitle = title;
      this.updateTitle();
   };
   this.setExpandedTitle = function( title )
   {
      this.expandedTitle = title;
      this.updateTitle();
   };
   this.setCollapsedTitle = function( title )
   {
      this.collapsedTitle = title;
      this.updateTitle();
   };

   this.setSection = function( section )
   {
      this.section = section;
      if (initialyCollapsed) {
          this.section.hide();
      }
      this.updateIcon();
   };

   this.addSection = function(sizer, weight) {
      sizer.add(this);
      sizer.add(this.section, weight);
   }


}

SectionBar.prototype = new Control;




// ========================================================================================================================
// GUI Main Dialog
// ========================================================================================================================

function MainDialog(engine, guiParameters) {
   this.__base__ = Dialog;
   this.__base__();
   this.engine = engine;
   this.guiParameters = guiParameters;

   var that = this;

   this.showFullPath = false;

   var labelWidth = this.font.width( "MMMMMMMMMMM" ) ;

   this.setMinWidth(MainDialogMinimumWidth);
   this.setMinHeight(MainDialogMinimumHeight);
   this.width = MainDialogPreferedWidth;
   this.height = MainDialogPreferedHeight;

   // -- FITSKeyword Dialog (opened as a child on request)
   this.fitsKeysDialog = new FITSKeysDialog( this, engine );


   // -- CompletionDialog Dialog (opened as a child on request, when the requested file operation is completed)
   this.completionDialog = new CompletionDialog( this, engine );



   //----------------------------------------------------------------------------------
   // -- HelpLabel
   //----------------------------------------------------------------------------------
   var helpLabel = new Label( this );
   helpLabel.frameStyle = FrameStyle_Box;
   helpLabel.margin = 4;
   helpLabel.wordWrapping = true;
   helpLabel.useRichText = true;
   helpLabel.text = Text.H.HELP_LABEL;


   //----------------------------------------------------------------------------------
   //--  Input file list section
   //----------------------------------------------------------------------------------
   this.filesTreeBox = new TreeBox( this );

   this.filesTreeBox.rootDecoration = false;
   this.filesTreeBox.numberOfColumns = 1;
   this.filesTreeBox.multipleSelection = true;
   this.filesTreeBox.headerVisible = true;
   this.filesTreeBox.headerSorting = true;
   this.filesTreeBox.setHeaderText(0, "Filename");
   this.filesTreeBox.sort(0,true);
   this.filesTreeBox.setMinSize( TreeboxWindowMinWidth, InputTreeboxMinHeight );
   this.filesTreeBox.toolTip = Text.H.FILES_TREEBOX_TOOLTIP;


   // Assume that 'check' is the only operation that update the nodes,
   // this may not be true...
   this.filesTreeBox.onNodeUpdated = function( node, column )  {
   // Invert CheckMark
#ifdef DEBUG_EVENTS
         debug("filesTreeBox: onNodeUpdated("+node+","+column+")");
#endif
      // Check box was likely changed
      this.dialog.updateTotal();
      this.dialog.refreshTargetFiles();
   };


   // -- Actions for input file list ---------------------------------------------------------------------------------------

   // -- Open FITS keyword dialog
   this.keyButton = new ToolButton( this );
   this.keyButton.icon = new Bitmap( ":/images/icons/text.png" );
   this.keyButton.toolTip = Text.H.KEY_BUTTON_TOOLTIP;
   this.keyButton.onClick = function() {
   if (this.dialog.engine.keywordsSet.size()) {
         this.dialog.fitsKeysDialog.execute();
         this.dialog.showOrHideFITSkey();
      }
   }



   // --  Add files
   this.filesAdd_Button = new ToolButton( this );
   this.filesAdd_Button.icon = new Bitmap( ":/images/image_container/add_files.png" );
   this.filesAdd_Button.toolTip = "Add files";
   this.filesAdd_Button.onClick = function() {
         var ofd = new OpenFileDialog;
         ofd.multipleSelections = true;
         ofd.caption = "Select FITS Files";
         ofd.filters = [["FITS Files", "*.fit", "*.fits", "*.fts"]];
         if ( ofd.execute() ) {
            this.dialog.addFilesAction(ofd.fileNames);
         }
      }



   // -- Add Directory
   this.dirAdd_Button = new ToolButton( this );
   this.dirAdd_Button.icon = new Bitmap( ":/images/icons/folders.png" );
   this.dirAdd_Button.toolTip =Text.H.DIRADD_BUTTON_TOOLTIP;
   this.dirAdd_Button.onClick = function()
      {
         var gdd = new GetDirectoryDialog;
         //gdd.initialPath = outputDirectory;
         gdd.caption = "Select Input Directory";
         if ( gdd.execute() )
         {
#ifdef DEBUG
            debug("Start searching FITS file in SubFolders");
#endif
            // TODO Make configurable
            var fileNames = searchDirectory(gdd.directory+"/*.fit" ,true)
            .concat(searchDirectory(gdd.directory+"/*.fits",true))
            .concat(searchDirectory(gdd.directory+"/*.fts",true));
#ifdef DEBUG
            debug("Finish searching FITS file in SubFolders");
#endif
            this.dialog.addFilesAction(fileNames);
         }
      }


   // -- Check selected files
   this.checkSelected_Button = new ToolButton( this );
   this.checkSelected_Button.icon = new Bitmap( ":/images/process_explorer/expand_all.png" );
   this.checkSelected_Button.toolTip = Text.H.CHECK_SELECTED_BUTTON_TOOLTIP;
   this.checkSelected_Button.onClick = function() {
#ifdef DEBUG
      debug("checkSelected_Button: onClick");
#endif
      for (var i=0; i < this.dialog.filesTreeBox.selectedNodes.length; i++) {
            this.dialog.filesTreeBox.selectedNodes[i].checked = true;
      }
      this.dialog.updateTotal();
      this.dialog.refreshTargetFiles();
   }

   // -- uncheck selected files
   this.uncheckSelected_Button = new ToolButton( this );
   this.uncheckSelected_Button.icon = new Bitmap( ":/images/process_explorer/collapse_all.png" );
   this.uncheckSelected_Button.toolTip = Text.H.CHECK_UNSELECTED_BUTTON_TOOLTIP;
   this.uncheckSelected_Button.onClick = function() {
#ifdef DEBUG
      debug("uncheckSelected_Button: onClick");
#endif
      for (var i=0; i < this.dialog.filesTreeBox.selectedNodes.length; i++) {
            this.dialog.filesTreeBox.selectedNodes[i].checked = false;
      }
      this.dialog.updateTotal();
      this.dialog.refreshTargetFiles();
   }

   // -- Remove selected files
   this.remove_files_Button = new ToolButton( this );
   this.remove_files_Button.icon = new Bitmap( ":/images/close.png" );
   this.remove_files_Button.toolTip = Text.H.REMOVE_FILES_BUTTON_TOOLTIP;
   this.remove_files_Button.onClick = function() {
#ifdef DEBUG
      debug("remove_files_Button: onClick");
#endif

      for ( var iTreeBox = this.dialog.filesTreeBox.numberOfChildren; --iTreeBox >= 0; ) {
         if ( this.dialog.filesTreeBox.child( iTreeBox ).selected ) {
            //var nameInTreeBox = this.dialog.filesTreeBox.child(iTreeBox).text(0);
            var nameInTreeBox = this.dialog.filesTreeBox.child(iTreeBox).fullFileName;

            this.dialog.engine.removeFiles(nameInTreeBox);
            this.dialog.filesTreeBox.remove( iTreeBox );
         }
      }
      this.dialog.updateTotal();
      this.dialog.updateButtonState();
      // Refresh the generated files
      this.dialog.refreshTargetFiles();
   }


   // -- Remove all files
   this.remove_all_files_Button = new ToolButton( this );
   this.remove_all_files_Button.icon = new Bitmap( ":/images/close_all.png" );
   this.remove_all_files_Button.toolTip = Text.H.REMOVE_ALL_FILES_BUTTON_TOOLTIP;
   this.remove_all_files_Button.onClick = function() {
#ifdef DEBUG
      debug("remove_all_files_Button: onClick");
#endif
      if (this.dialog.filesTreeBox.numberOfChildren > 0) {
          if (new MessageBox( "Do you really want to remove all " +
             this.dialog.filesTreeBox.numberOfChildren + " files from input list ?",
              TITLE, StdIcon_Question, StdButton_No, StdButton_Yes ).execute() != StdButton_Yes )
            {
            return;
         }
      }

      // TODO We can probably clear in one go
      for ( var i = this.dialog.filesTreeBox.numberOfChildren; --i >= 0; ) {
            this.dialog.filesTreeBox.remove( i );
      }
      this.dialog.engine.reset();
      this.dialog.updateTotal();
      this.dialog.updateButtonState();
      // Refresh the generated files
      this.dialog.refreshTargetFiles();

   }

   this.fullPath_CheckBox = new CheckBox(this);
   this.fullPath_CheckBox.state = this.showFullPath ? 1 : 0;
   this.fullPath_CheckBox.onCheck = function(checked) {
      this.dialog.showFullPath = checked;
      this.dialog.rebuildFilesTreeBox();
   }

   this.fullPath_Label = new Label(this);
   this.fullPath_Label.text = "Full path";
   this.fullPath_Label.textAlignment	= TextAlign_Left | TextAlign_VertCenter;

   // -- Total file Label ---------------------------------------------------------------------------
   this.inputSummaryLabel = new Label( this );
   this.inputSummaryLabel.textAlignment = TextAlign_Right|TextAlign_VertCenter;

   // -- Sizer for Input Files Section

   this.fileButonSizer = new HorizontalSizer;
   this.fileButonSizer.margin = 6;
   this.fileButonSizer.spacing = 4;
   this.fileButonSizer.add( this.keyButton );
   this.fileButonSizer.addSpacing( 5 );
   this.fileButonSizer.add( this.filesAdd_Button );
   this.fileButonSizer.add( this.dirAdd_Button );
   this.fileButonSizer.add( this.checkSelected_Button );
   this.fileButonSizer.add( this.uncheckSelected_Button );
   this.fileButonSizer.addSpacing( 5 );
   this.fileButonSizer.add( this.remove_files_Button );
   this.fileButonSizer.add( this.remove_all_files_Button );
   this.fileButonSizer.add( this.fullPath_CheckBox );
   this.fileButonSizer.add( this.fullPath_Label );
   this.fileButonSizer.addSpacing( 3 );
   this.fileButonSizer.add( this.inputSummaryLabel );
   this.fileButonSizer.addStretch();


   this.inputFiles_GroupBox = new GroupBox( this );
   this.inputFiles_GroupBox.sizer = new VerticalSizer;
   this.inputFiles_GroupBox.sizer.margin = 4;
   this.inputFiles_GroupBox.sizer.spacing = 2;
   this.inputFiles_GroupBox.sizer.add( this.filesTreeBox,100 );
   this.inputFiles_GroupBox.sizer.add( this.fileButonSizer );



   this.barInput = makeSectionGroup(this,this.inputFiles_GroupBox,"Input",false);
   this.barInput.setCollapsedTitle( "Input - No file" );




   //----------------------------------------------------------------------------------
   // -- Configuration section
   //----------------------------------------------------------------------------------

   // This section shows current configuration and allow access to the configuration dialog

   var configurationSelectedCallback = function(configurationName) {
#ifdef DEBUG
      debug("MainDialog: configurationSelectedCallback - ConfigurationSet selected:",configurationName);
#endif

      var selectedConfiguration = ffM_Configuration.getConfigurationByName(ffM_Configuration.getConfigurationTable(), configurationName);
      if (selectedConfiguration == null) {
         throw "PROGRAM ERROR - Invalid configuration set name '" + configurationName +"'";
      }
      // Update the configuration - THIS REBUILD MOST OF THE VARIABLES AND FILE LISTS
      ffM_Configuration.setActiveConfigurationName(configurationName);
      that.configurationDescription_Label.text	=  ffM_Configuration.getActiveConfigurationElement().description;

      that.engine.setConfiguration(ffM_Configuration.createWorkingConfiguration());
      that.engine.rebuildAll();


      // TODO - Merge with action on add files and manage configurations
      that.rebuildFilesTreeBox();
      that.updateButtonState();
      that.updateTotal();
      that.refreshTargetFiles();

   }
   this.configurationSelection_ComboBox = new ffM_GUI_config.ConfigurationSelection_ComboBox(this, [], configurationSelectedCallback);
   this.configurationSelection_ComboBox.configure(
         ffM_Configuration.getAllConfigurationNames(ffM_Configuration.getConfigurationTable()),
         ffM_Configuration.getActiveConfigurationName());

   this.configuration_Button = new PushButton( this );
   this.configuration_Button.text = Text.T.CONFIGURE_BUTTON_TEXT;
   this.configuration_Button.toolTip = Text.H.CONFIGURE_BUTTON_TOOLTIP;
   this.configuration_Button.enabled = true;
   this.configurationDialog = ffM_GUI_config.makeDialog(this);

   this.configurationDescription_Label = new Label();
   this.configurationDescription_Label.text =  ffM_Configuration.getActiveConfigurationElement().description;
   this.configurationDescription_Label.textAlignment	= TextAlign_Left | TextAlign_VertCenter;
   this.configurationDescription_Label.toolTip = "The description of the current configuration (define the synthetic variables)";

   this.configuration_Button.onClick = function() {
      var configurationName = ffM_Configuration.getActiveConfigurationName();
      var configurationDialog = this.dialog.configurationDialog;
      configurationDialog.configure(ffM_Configuration.getConfigurationTable(), configurationName);

      var result =  configurationDialog.execute();

      if (result) {
         // Update the configuration - THIS REBUILD MOST OF THE VARIABLES AND FILE LISTS
         ffM_Configuration.replaceConfigurationTable(configurationDialog.editedConfigurationSet,configurationDialog.selectedConfiguration.name)


         this.dialog.configurationSelection_ComboBox.configure(
            ffM_Configuration.getAllConfigurationNames(configurationDialog.editedConfigurationSet),
            configurationDialog.selectedConfiguration.name);

         this.dialog.configurationDescription_Label.text	=  ffM_Configuration.getActiveConfigurationElement().description;

         engine.setConfiguration(ffM_Configuration.createWorkingConfiguration());
         engine.rebuildAll();


         // TODO - Merge with action on add files and select configuration
         this.dialog.rebuildFilesTreeBox();
         this.dialog.updateButtonState();
         this.dialog.updateTotal();
         this.dialog.refreshTargetFiles();
      }
   }



   // Target template --------------------------------------------------------------------------------------

   this.targetFileTemplate_ComboBox = new ComboBox( this );
   this.targetFileTemplate_ComboBox.toolTip = Text.H.TARGET_FILE_TEMPLATE_TOOLTIP;
   this.targetFileTemplate_ComboBox.enabled = true;
   this.targetFileTemplate_ComboBox.editEnabled = true;
   for (var it = 0; it<guiParameters.targetFileItemListText.length; it++) {
      this.targetFileTemplate_ComboBox.addItem("'" + guiParameters.targetFileItemListText[it] +  "' - " + guiParameters.targetFileItemListComment[it]);
   }
   this.targetFileTemplate_ComboBox.editText = guiParameters.targetFileItemListText[0];


   this.targetFileTemplate_ComboBox.onEditTextUpdated = function() {
#ifdef DEBUG
      debug("targetFileTemplate_ComboBox: onEditTextUpdated " + this.editText);
#endif

      var text = this.editText;
      var templateErrors = [];

      var templateErrors = [];
      var t = ffM_template.analyzeTemplate(templateErrors, text);
      if (templateErrors.length === 0) {
         this.textColor = 0x000000;
         guiParameters.targetFileNameCompiledTemplate  = t;
         this.dialog.refreshTargetFiles();
      } else {
         this.textColor = 0xFF0000;
      }
   }

   this.targetFileTemplate_ComboBox.onItemSelected = function() {
#ifdef DEBUG
      debug("targetFileTemplate_ComboBox: onItemSelected " + this.currentItem);
#endif
      if (this.currentItem >= guiParameters.targetFileItemListText.length) {
         return;  // protect as when a 'CR' is typed in the field, currentItem may be outside of array
      }
      var text =guiParameters.targetFileItemListText[this.currentItem];
      this.dialog.targetFileTemplate_ComboBox.editText = text;
      var templateErrors = [];
      var t = ffM_template.analyzeTemplate(templateErrors, text);
      if (templateErrors.length === 0) {
         this.textColor = 0x000000;
         guiParameters.targetFileNameCompiledTemplate  = t;
         this.dialog.refreshTargetFiles();
      } else {
         this.textColor = 0xFF0000;
      }
   }


   // Regular expression on source file --------------------------------------------------------------------------------------
   this.regexp_ComboBox = new ComboBox( this );
   this.regexp_ComboBox.toolTip = Text.H.SOURCE_FILENAME_REGEXP_TOOLTIP;
   this.regexp_ComboBox.enabled = true;
   this.regexp_ComboBox.editEnabled = true;
   for (var it = 0; it<guiParameters.regexpItemListText.length; it++) {
      this.regexp_ComboBox.addItem("'" + guiParameters.regexpItemListText[it] +  "' - " + guiParameters.regexpItemListComment[it]);
   }
   this.regexp_ComboBox.editText = guiParameters.regexpItemListText[0];


   this.regexp_ComboBox.onEditTextUpdated = function() {
#ifdef DEBUG
      debug("regexp_ComboBox: onEditTextUpdated " + this.editText);
#endif

      var re = this.editText.trim();
      if (re.length === 0) {
         guiParameters.sourceFileNameRegExp = null;
#ifdef DEBUG
         debug("sourceTemplate_Edit: onTextUpdated:- cancel regexp");
#endif
      } else {
         try {
            guiParameters.sourceFileNameRegExp = regExpFromUserString(re);
            this.textColor = 0x000000;
#ifdef DEBUG
            debug("sourceTemplate_Edit: onTextUpdated: regexp: " + guiParameters.sourceFileNameRegExp);
#endif
         } catch (err) {
            guiParameters.sourceFileNameRegExp = null;
            this.textColor = 0xFF0000;
#ifdef DEBUG
            debug("sourceTemplate_Edit: onTextUpdated:  bad regexp - err: " + err);
#endif
         }
      }
   }

   this.regexp_ComboBox.onItemSelected = function() {
#ifdef DEBUG
      debug("regexp_ComboBox: onItemSelected " + this.currentItem);
#endif
      if (this.currentItem >= guiParameters.regexpItemListText.length) {
         return;  // protect as when a 'CR' is typed in the field, currentItem may be outside of array
      }
      var text = regExpToString(guiParameters.regexpItemListText[this.currentItem]);
      this.dialog.regexp_ComboBox.editText = text;
      var re = text.trim();
      if (re.length === 0) {
         guiParameters.sourceFileNameRegExp = null;
#ifdef DEBUG
         debug("sourceTemplate_Edit: onTextUpdated:- cancel regexp");
#endif
      } else {
         try {
            guiParameters.sourceFileNameRegExp = regExpFromUserString(re);
            this.textColor = 0x000000;
#ifdef DEBUG
            debug("sourceTemplate_Edit: onTextUpdated: regexp: " + guiParameters.sourceFileNameRegExp);
#endif
         } catch (err) {
            guiParameters.sourceFileNameRegExp = null;
            this.textColor = 0xFF0000;
#ifdef DEBUG
            debug("sourceTemplate_Edit: onTextUpdated:  bad regexp - err: " + err);
#endif
         }
      }
   }


   // Group template --------------------------------------------------------------------------------------

   this.groupTemplate_ComboBox = new ComboBox( this );
   this.groupTemplate_ComboBox.toolTip = Text.H.GROUP_TEMPLATE_TOOLTIP;
   this.groupTemplate_ComboBox.enabled = true;
   this.groupTemplate_ComboBox.editEnabled = true;
   for (var it = 0; it<guiParameters.groupItemListText.length; it++) {
      this.groupTemplate_ComboBox.addItem("'" + guiParameters.groupItemListText[it] +  "' - " + guiParameters.groupItemListComment[it]);
   }
   this.groupTemplate_ComboBox.editText = guiParameters.groupItemListText[0];


   this.groupTemplate_ComboBox.onEditTextUpdated = function() {
#ifdef DEBUG
      debug("groupTemplate_ComboBox: onEditTextUpdated " + this.editText);
#endif

      var text = this.editText;
      var templateErrors = [];

      var t = ffM_template.analyzeTemplate(templateErrors,text);
      if (templateErrors.length === 0) {
         this.textColor = 0x000000;
         guiParameters.groupByCompiledTemplate  = t;
         this.dialog.refreshTargetFiles();
      } else {
         this.textColor = 0xFF0000;
      }
   }

   this.groupTemplate_ComboBox.onItemSelected = function() {
#ifdef DEBUG
      debug("groupTemplate_ComboBox: onItemSelected " + this.currentItem);
#endif
      if (this.currentItem >= guiParameters.groupItemListText.length) {
         return;  // protect as when a 'CR' is typed in the field, currentItem may be outside of array
      }
      var text = guiParameters.groupItemListText[this.currentItem];
      this.dialog.groupTemplate_ComboBox.editText = text;
      var templateErrors = [];
      var t = ffM_template.analyzeTemplate(templateErrors,text);
      if (templateErrors.length === 0) {
         this.textColor = 0x000000;
         guiParameters.groupByCompiledTemplate  = t;
         this.dialog.refreshTargetFiles();
      } else {
         this.textColor = 0xFF0000;
      }
   }


   // Sizers

   var configurationNameMinWidth = this.font.width( "Target file template:  " );

   this.configurationSelection_sizer = new HorizontalSizer;
   this.configurationSelection_sizer.margin = 4;
   this.configurationSelection_sizer.spacing = 2;
   var label = new Label();
   label.setFixedWidth(labelWidth);
   label.text		= "Configuration: ";
   label.textAlignment	= TextAlign_Right | TextAlign_VertCenter;


   this.configurationSelection_sizer.add( label );
   this.configurationSelection_sizer.add( this.configurationSelection_ComboBox );
   this.configurationSelection_sizer.addSpacing(5);
   this.configurationSelection_ComboBox.minWidth = configurationNameMinWidth;


   this.configurationSelection_sizer.add( this.configurationDescription_Label );
   this.configurationSelection_sizer.addStretch( );
   this.configurationSelection_sizer.add( this.configuration_Button );


   this.targetFileTemplate_Edit_sizer = new HorizontalSizer;
   this.targetFileTemplate_Edit_sizer.margin = 4;
   this.targetFileTemplate_Edit_sizer.spacing = 2;
   var label = new Label();
   label.setFixedWidth(labelWidth);
   label.text		= "Target file template: ";
   label.textAlignment	= TextAlign_Right | TextAlign_VertCenter;

   this.targetFileTemplate_Edit_sizer.add( label );
   this.targetFileTemplate_Edit_sizer.add( this.targetFileTemplate_ComboBox,100 );



   this.regexp_ComboBox_sizer = new HorizontalSizer;
   this.regexp_ComboBox_sizer.margin = 4;
   this.regexp_ComboBox_sizer.spacing = 2;
   var label = new Label();
   label.setFixedWidth(labelWidth);;
   label.text		= "File name RegExp: ";
   label.textAlignment	= TextAlign_Right | TextAlign_VertCenter;

   this.regexp_ComboBox_sizer.add( label );
   this.regexp_ComboBox_sizer.add( this.regexp_ComboBox,100 );


   this.groupTemplate_ComboBox_sizer = new HorizontalSizer;
   this.groupTemplate_ComboBox_sizer.margin = 4;
   this.groupTemplate_ComboBox_sizer.spacing = 2;
   var label = new Label();
   label.setFixedWidth(labelWidth);
   label.text		= "Group template: ";
   label.textAlignment	= TextAlign_Right | TextAlign_VertCenter;

   this.groupTemplate_ComboBox_sizer.add( label );
   this.groupTemplate_ComboBox_sizer.add( this.groupTemplate_ComboBox,100);


   this.rules_GroupBox = new GroupBox( this );

   this.rules_GroupBox.sizer = new VerticalSizer;
   this.rules_GroupBox.sizer.margin = 4;
   this.rules_GroupBox.sizer.spacing = 2;

   this.rules_GroupBox.sizer.add( this.configurationSelection_sizer );
   this.rules_GroupBox.sizer.add( this.targetFileTemplate_Edit_sizer, 100);
   this.rules_GroupBox.sizer.add( this.regexp_ComboBox_sizer );
   this.rules_GroupBox.sizer.add( this.groupTemplate_ComboBox_sizer );

   // More or less futile attempt to avoid bad small layout on unix
   this.rules_GroupBox.adjustToContents();
   this.rules_GroupBox.setFixedHeight();

   this.barRules = makeSectionGroup(this, this.rules_GroupBox,"Rules",false);





   //----------------------------------------------------------------------------------
   // -- Output directory section
   //----------------------------------------------------------------------------------

   //Output Dir --------------------------------------------------------------------------------------
   this.outputDir_Edit = new Edit( this );
   this.outputDir_Edit.readOnly = true;
   this.outputDir_Edit.text = this.engine.outputDirectory;
   this.outputDir_Edit.toolTip = Text.H.OUTPUTDIR_TOOLTIP;

   this.outputDirSelect_Button = new ToolButton( this );
   this.outputDirSelect_Button.icon = new Bitmap( ":/images/icons/select.png" );
   this.outputDirSelect_Button.toolTip = Text.H.OUTPUTDIR_SELECT_TOOLTIP;
   this.outputDirSelect_Button.onClick = function() {
      var gdd = new GetDirectoryDialog;
      gdd.initialPath = engine.outputDirectory;
      gdd.caption = Text.T.GET_DIRECTORY_DIALOG_CAPTION;
      if ( gdd.execute() ) {
         this.dialog.engine.outputDirectory = gdd.directory;
         this.dialog.outputDir_Edit.text = this.dialog.engine.outputDirectory;
         this.dialog.barOutput.setCollapsedTitle( Text.T.OUPUT_SECTION_TEXT_PART + " - " +  this.dialog.engine.outputDirectory);
         this.dialog.updateButtonState();
      }
   }

   this.outputDir_GroupBox = new GroupBox( this );
   this.outputDir_GroupBox.sizer = new HorizontalSizer;
   this.outputDir_GroupBox.sizer.margin = 4;
   this.outputDir_GroupBox.sizer.spacing = 2;
   this.outputDir_GroupBox.sizer.add( this.outputDir_Edit, 100 );
   this.outputDir_GroupBox.sizer.add( this.outputDirSelect_Button );

   // More or less futile attempt to avoid bad small layout on unix
   this.outputDir_GroupBox.adjustToContents();
   this.outputDir_GroupBox.setFixedHeight();

   this.barOutput = makeSectionGroup(this, this.outputDir_GroupBox, Text.T.OUPUT_SECTION_TEXT_PART, false);


   //----------------------------------------------------------------------------------
   // -- Operation list and action section
   //----------------------------------------------------------------------------------


   // Result operations --------------------------------------------------------------------------------------
   this.transform_TreeBox = new TreeBox( this );

   this.transform_TreeBox.rootDecoration = false;
   this.transform_TreeBox.numberOfColumns = 1;
   this.transform_TreeBox.multipleSelection = false;
   this.transform_TreeBox.headerVisible = false;
   this.transform_TreeBox.headerSorting = false;
   this.transform_TreeBox.setHeaderText(0, "Filename");
   //this.transform_TreeBox.sort(0,false);
   this.transform_TreeBox.setMinSize( TreeboxWindowMinWidth, TransformTreeBoxMinHeight );
   this.transform_TreeBox.toolTip = Text.H.TRANSFORM_TREEBOX_TOOLTIP;

   // Select the corresponding source file when the target file is selected
   // Note that the source files may not all be in the list, so
   // a lookup is required.
   this.transform_TreeBox.onCurrentNodeUpdated = function(newCurrentNode) {
      if (newCurrentNode !== null) {
//         var index = this.childIndex(newCurrentNode);
//         index = Math.floor(index / 2 );
         var index = newCurrentNode.inputFileIndex;
         this.dialog.filesTreeBox.currentNode = this.dialog.filesTreeBox.child(index);
      }
   }

   this.outputSummaryLabel = new Label( this );
   this.outputSummaryLabel.textAlignment = TextAlign_Left|TextAlign_VertCenter;
   this.outputSummaryLabel.text = "No operation";

   this.outputFiles_GroupBox = new GroupBox( this );
   this.outputFiles_GroupBox.sizer = new VerticalSizer;
   this.outputFiles_GroupBox.sizer.margin = 4;
   this.outputFiles_GroupBox.sizer.spacing = 2;
   this.outputFiles_GroupBox.sizer.add( this.transform_TreeBox, 100);
   this.outputFiles_GroupBox.sizer.add( this.outputSummaryLabel );


   this.barResult = makeSectionGroup(this, this.outputFiles_GroupBox, Text.T.RESULT_SECTION_PART_TEXT, false);
   this.barResult.setCollapsedTitle( Text.T.RESULT_SECTION_PART_TEXT + " - None" );



   // -- Action buttons --------------------------------------------------------------------------------------

   this.check_Button = new PushButton( this );
   this.check_Button.text = Text.T.CHECK_BUTTON_TEXT;
   this.check_Button.toolTip = Text.H.CHECK_BUTTON_TOOLTIP;
   this.check_Button.enabled = true;
   this.check_Button.onClick = function() {
      var listOfFiles = this.parent.makeListOfCheckedFiles();
      var errors = this.parent.engine.checkValidTargets(listOfFiles);
      if (errors.length > 0) {
         var msg = new MessageBox( errors.join("\n"),
                   "Check failed", StdIcon_Error, StdButton_Ok );
         msg.execute();
      } else if (this.parent.engine.targetFiles.length === 0) {
         var msg = new MessageBox(
            "There is no file to move or copy", "Check irrelevant", StdIcon_Information, StdButton_Ok );
         msg.execute();
      } else {
         var text = "" + this.parent.engine.targetFiles.length + " files checked" ;
            if (this.parent.engine.nmbFilesTransformed >0) {
               text += ", " + this.parent.engine.nmbFilesTransformed + " to copy/move";
            }

         //      this.engine.nmbFilesSkipped;
         if (this.parent.engine.nmbFilesInError >0) {
            // Should not happens
            text += ", " + this.parent.engine.nmbFilesInError + " IN ERROR";
         }

         if (!this.parent.engine.outputDirectory) {
            text += ",\nbut output directory is not defined";
         }

         var msg = new MessageBox(text,
            "Check successful", StdIcon_Information, StdButton_Ok );
         msg.execute();
      }
   }

   this.refresh_Button = new PushButton( this );
   this.refresh_Button.text = Text.T.REFRESH_BUTTON_TEXT;
   this.refresh_Button.toolTip = Text.H.REFRESH_BUTTON_TOOLTIP;
   this.refresh_Button.enabled = true;
   this.refresh_Button.onClick = function() {
      this.parent.removeDeletedFiles();
      this.parent.refreshTargetFiles();
   }



   this.move_Button = new PushButton( this );
   this.move_Button.text = Text.T.MOVE_BUTTON_TEXT;
   this.move_Button.toolTip = Text.H.MOVE_BUTTON_TOOLTIP;
   this.move_Button.enabled = false;
   this.move_Button.onClick = function() {
      var listOfFiles = this.parent.makeListOfCheckedFiles();
      var errors = this.parent.engine.checkValidTargets(listOfFiles);
      if (errors.length > 0) {
         var msg = new MessageBox( errors.join("\n"),
                   "Check failed", StdIcon_Error, StdButton_Ok );
         msg.execute();
         return;
      }
      try {
         console.show();
         this.dialog.cursor = new Cursor( StdCursor_Hourglass );
         var resultText = this.parent.engine.executeFileOperations(0);
         this.parent.removeDeletedFiles();
         this.parent.refreshTargetFiles();
      } finally {
         this.dialog.cursor = new Cursor( StdCursor_Arrow );
         //console.hide();
      }
      this.dialog.completionDialog.setResultText(resultText + "\nMoved files were removed from the input list");
      this.dialog.completionDialog.setResultModeMove();
      var completionCode =  this.dialog.completionDialog.execute();
      switch (completionCode) {
         case CompletionDialog_doneContinue:
         case CompletionDialog_doneKeep:
         case CompletionDialog_doneRemove:
            // Nothing to do, only Continue makes sense above
         break;
         case CompletionDialog_doneLeave:
            this.dialog.ok();
         break;
      }
   }


   this.copy_Button = new PushButton( this );
   this.copy_Button.text = Text.T.COPY_BUTTON_TEXT;
   this.copy_Button.toolTip = Text.H.COPY_BUTTON_TOOLTIP;
   this.copy_Button.enabled = false;
   this.copy_Button.onClick = function() {
      var listOfFiles = this.parent.makeListOfCheckedFiles();
      var errors = this.parent.engine.checkValidTargets(listOfFiles);
      if (errors.length > 0) {
            var msg = new MessageBox( errors.join("\n"),
                   "Check failed", StdIcon_Error, StdButton_Ok );
            msg.execute();
            return;
      }
      try {
         console.show();
         this.dialog.cursor = new Cursor( StdCursor_Hourglass );
         var resultText =  this.parent.engine.executeFileOperations(1);
      } finally {
         this.dialog.cursor = new Cursor( StdCursor_Arrow );
         //console.hide();
      }

      this.dialog.completionDialog.setResultText(resultText + "\nCopied files are still checked in the input list");
      this.dialog.completionDialog.setResultModeCopy();
      var completionCode =  this.dialog.completionDialog.execute();
      switch (completionCode) {
         case CompletionDialog_doneContinue:
         case CompletionDialog_doneKeep:
            return;
         break;
         case CompletionDialog_doneRemove:
            for ( var iTreeBox = this.dialog.filesTreeBox.numberOfChildren; --iTreeBox >= 0; ) {
            if ( this.dialog.filesTreeBox.child( iTreeBox ).checked ) {
//               var nameInTreeBox = this.dialog.filesTreeBox.child(iTreeBox).text(0);
               var nameInTreeBox = this.dialog.filesTreeBox.child(iTreeBox).fullFileName;
               this.dialog.engine.removeFiles(nameInTreeBox);
               this.dialog.filesTreeBox.remove( iTreeBox );
            }
         }
         this.dialog.updateTotal();
         this.dialog.updateButtonState();
         this.dialog.refreshTargetFiles();
         break;
         case CompletionDialog_doneLeave:
            this.dialog.ok();
         break;
      }
   }



   this.loadSave_Button = new PushButton( this );
   this.loadSave_Button.text = Text.T.LOADSAVE_BUTTON_TEXT;
   this.loadSave_Button.toolTip = Text.H.LOADSAVE_BUTTON_TOOLTIP;
   this.loadSave_Button.enabled = false;
   this.loadSave_Button.onClick = function() {
      var listOfFiles = this.parent.makeListOfCheckedFiles();
      var errors = this.parent.engine.checkValidTargets(listOfFiles);
      if (errors.length > 0) {
            var msg = new MessageBox( errors.join("\n"),
                   "Check failed", StdIcon_Error, StdButton_Ok );
            msg.execute();
            return;
      }
      try {
         console.show();
         this.dialog.cursor = new Cursor( StdCursor_Hourglass );
         var resultText = this.parent.engine.executeFileOperations(2);
      } finally {
         this.dialog.cursor = new Cursor( StdCursor_Arrow );
         //console.hide();
      }

      this.dialog.completionDialog.setResultText(resultText+ "\nLoad/saved files are still checked in the input list");
      this.dialog.completionDialog.setResultModeCopy();
      var completionCode =  this.dialog.completionDialog.execute();
      switch (completionCode) {
         case CompletionDialog_doneContinue:
         case CompletionDialog_doneKeep:
            return;
         break;
         case CompletionDialog_doneRemove:
            for ( var iTreeBox = this.dialog.filesTreeBox.numberOfChildren; --iTreeBox >= 0; ) {
            if ( this.dialog.filesTreeBox.child( iTreeBox ).checked ) {
               //var nameInTreeBox = this.dialog.filesTreeBox.child(iTreeBox).text(0);
               var nameInTreeBox = this.dialog.filesTreeBox.child(iTreeBox).fullFileName;
               this.dialog.engine.removeFiles(nameInTreeBox);
               this.dialog.filesTreeBox.remove( iTreeBox );
            }
         }
         this.dialog.updateTotal();
         this.dialog.updateButtonState();
         this.dialog.refreshTargetFiles();

         break;
         case CompletionDialog_doneLeave:
            this.dialog.ok();
         break;
      }
   }


#ifdef IMPLEMENTS_FITS_EXPORT
// Export FITS values button
   this.txt_Button = new PushButton( this );
   this.txt_Button.text = "Export FITS.txt";
   this.txt_Button.toolTip = "For Checked files write FitKeywords value to file FITS.txt in output directory";
   this.txt_Button.enabled = false;
   this.txt_Button.onClick = function() {
      this.parent.engine.exportFITSKeywords();
   }
#endif

   this.exit_Button = new PushButton( this );
   this.exit_Button.text = "Exit";
   this.exit_Button.toolTip = "Exit (no action)";
   this.exit_Button.enabled = true;
   this.exit_Button.onClick = function() {
      if (this.dialog.filesTreeBox.numberOfChildren == 0 ||
          ((new MessageBox( "Do you really want to exit " + TITLE + " ?",
              TITLE, StdIcon_Question, StdButton_No, StdButton_Yes )).execute() == StdButton_Yes )
      ) {
         this.dialog.cancel();
      }
   }

   // Help buton
   this.helpButton = new ToolButton( this );
   this.helpButton.icon = new Bitmap( ":/images/interface/browseDocumentationButton.png" );
   this.helpButton.toolTip = Text.H.HELP_BUTTON_TOOLTIP;
   this.helpDialog = new HelpDialog(this);
   this.helpButton.onClick = function() {
      this.dialog.helpDialog.execute();
   }




   // Sizer for Operation List and Actions section

   this.buttonSizer = new HorizontalSizer;
   this.buttonSizer.margin = 4;
   this.buttonSizer.spacing = 2;
   this.buttonSizer.add( this.refresh_Button);
   this.buttonSizer.add( this.check_Button);
   this.buttonSizer.add( this.move_Button);
   this.buttonSizer.add( this.copy_Button);
   this.buttonSizer.add( this.loadSave_Button);
#ifdef IMPLEMENTS_FITS_EXPORT
   this.buttonSizer.add( this.txt_Button);
#endif
   this.buttonSizer.addStretch();
   this.buttonSizer.add( this.exit_Button);
   this.buttonSizer.add( this.helpButton);




   // --------------------------------------------------------------------------------------------
   // Sizer for dialog

   this.sizer = new VerticalSizer;
   this.sizer.margin = 4;
   this.sizer.spacing = 6;
   this.sizer.add( helpLabel );
   this.barInput.addSection(this.sizer, 50);
   this.barRules.addSection(this.sizer,0);
   this.barOutput.addSection(this.sizer,0);
   this.barResult.addSection(this.sizer, 50);
   this.sizer.add( this.buttonSizer );






   //----------------------------------------------------------------------------------
   // -- Support methods
   //----------------------------------------------------------------------------------


   // -- Set visibility of synthetic and FITS keywords columns (called to apply changes)
   this.showOrHideFITSkey = function () {
      var allFITSKeyNames = this.engine.keywordsSet.allValueKeywordNameList;
      for (var i = 0; i<ffM_Configuration.syntheticVariableNames.length;i++) {
        // +1 as the file name column is always visible
         var c = i + 1;
         this.filesTreeBox.showColumn( c, this.engine.isVariableVisible(ffM_Configuration.syntheticVariableNames[i]));
      }
      for (var i = 0; i<allFITSKeyNames.length;i++) {
        // + 1 to skip file name and then skip the synthetic variables columns (they are present even if not shown))
         var c = i + 1 + ffM_Configuration.syntheticVariableNames.length;
         this.filesTreeBox.showColumn( c, this.engine.shownFITSKeyNames.hasOwnProperty(allFITSKeyNames[i]));
      }
   }


   // -- Rebuild the TreeBox content
   this.rebuildFilesTreeBox = function () {
      var i, keys, node, name, iKeyOfFile, k,displayFileName;

#ifdef DEBUG
      debug("rebuildFilesTreeBox: rebuilding filesTreeBox - " + this.engine.inputFiles.length + " input files");
#endif

      this.filesTreeBox.clear();
      this.filesTreeBox.numberOfColumns = 1; // Filename

      this.engine.keywordsSet = ffM_FITS_Keywords.makeKeywordsSet(); // clear

      // Add the synthetic keys columns
      for (var iSynthKey = 0; iSynthKey<ffM_Configuration.syntheticVariableNames.length; iSynthKey++) {
         var name = ffM_Configuration.syntheticVariableNames[iSynthKey];
         this.filesTreeBox.numberOfColumns++;// add new column
         this.filesTreeBox.setHeaderText(this.filesTreeBox.numberOfColumns-1, name);//set name of new column
#ifdef DEBUG
         debug("rebuildFilesTreeBox: added synthetic key header '" + name +  "' as col " + this.filesTreeBox.numberOfColumns);
#endif
      }



      // Add all lines (nodes), one for each file
      // The node will contain the file name, then the synthethic keys, then the FITS keys.
      // All synthethic and FITS keys are added, even if they are not shown or used (this simplifies some code,
      // but maybe could be optimized).
      // The list of synthethic keys is currently fixed.
      // The list of FITS key is dynamic and FITS keys coluns are added as needed when an image
      // has a FITS keyword not yet mapped to a column. The mapping of key name to column index
      // is built on the fly.
      // Only the keys with a value (not of type isNull) are considered (this skip comments)
      var longestFileName = "          "; // File name column will have at least 10 characters

      for (var i = 0; i < this.engine.inputFiles.length; ++i) {

         if (this.showFullPath) {
            displayFileName = this.engine.inputFiles[i];
         } else {
            displayFileName = File.extractNameAndExtension(this.engine.inputFiles[i]);
         }

         if (displayFileName.length>longestFileName.length) {longestFileName = displayFileName;}

#ifdef DEBUG
         debug("rebuildFilesTreeBox: adding file '" +this.engine.inputFiles[i] + "' to row " + i);
#endif

         var imageKeywords = this.engine.inputFITSKeywords[i]; // all FITS keywords/Values of the current file
         var keys=imageKeywords.fitsKeywordsList;
         var syntheticKeywords = this.engine.inputVariables[i]; // Map of all synthethic keywords and values of the current file

         // Create TreeBoxNode (line) for the current file
         var node = new TreeBoxNode( this.filesTreeBox );
         // put name of the file int the first column
         node.setText( 0, displayFileName );
         node.checked = true;
         // Keep full file name for retrievial by action methods (as the rows may be sorted)
         node.fullFileName = this.engine.inputFiles[i];
         // Reserve column for file name
         var colOffset = 1;

         // Add synthethic keyword columns (based on fixed list of syntethic keywords)
#ifdef DEBUG
         debug("rebuildFilesTreeBox: adding " + Object.keys(syntheticKeywords) + " synthetics keys, " + keys.length + " FITS keys to row " + i);
#endif
         for (var iSynthKey = 0; iSynthKey<ffM_Configuration.syntheticVariableNames.length; iSynthKey++) {
            var name = ffM_Configuration.syntheticVariableNames[iSynthKey];
            var textSynthKey = syntheticKeywords[name];
            node.setText(iSynthKey+colOffset, textSynthKey ? textSynthKey : "");
         }
         // Skip next columns
         colOffset += ffM_Configuration.syntheticVariableNames.length;

         // Adding FITS keyword columns (based on FITS keywords in current file and map of keywords to column index)
#ifdef DEBUG
         debug("rebuildFilesTreeBox: setting " + keys.length + " FITS keys to row " + i + ", colOffset=" +colOffset);
#endif
         for ( var iKeyOfFile = 0; iKeyOfFile<keys.length; iKeyOfFile++) {
            var key = keys[iKeyOfFile];

            // Only show the value keywords (not the comment keywords)
            if (! keys[iKeyOfFile].isNull) {
               var name = key.name; //name of Keyword from file
               var allFITSKeyNames = this.engine.keywordsSet.allValueKeywordNameList;
               var indexOfKey = allFITSKeyNames.indexOf(name);// find index of "name" in allFITSKeyNames
               if (indexOfKey < 0)  {
               // new FITS keyword, not yet mapped to a column, add new value keyword
#ifdef DEBUG_COLUMNS
                  debug("rebuildFilesTreeBox: Creating new column " + this.filesTreeBox.numberOfColumns + " for value keywords '"  + name + "', total col len " + this.engine.allFITSKeyNames.size());
#endif
                  allFITSKeyNames.push(name);//add keyword name to table
                  this.filesTreeBox.numberOfColumns++;// add new column
                  this.filesTreeBox.setHeaderText(this.filesTreeBox.numberOfColumns-1, name);//set name of new column
                  indexOfKey = this.filesTreeBox.numberOfColumns-colOffset-1;
               }
               // Set column content to value of keyword
#ifdef DEBUG_COLUMNS
               debug("rebuildFilesTreeBox: Set column, colOffset " + colOffset + ", index "  + indexOfKey + ", value " + keys[iKeyOfFile].value);
#endif
               var formattedValue = key.strippedValue;

              if (key.isNumeric) {
                  // Remove leading 0 and trailing 0 of decimal values to use less space
                  var n = key.numericValue;
                  if (n % 1 === 0) {
                     // Will be formatted without decimal point and leading 0
                     formattedValue = n.toString();
                  } else {
                     // TODO Supress possible leading zero of floating (leave 1 before decimal point)
                     formattedValue = key.strippedValue;
                  }
               } else {
                  formattedValue = ffM_FITS_Keywords.unquote(key.value);
               }
               node.setText(colOffset + indexOfKey, formattedValue);
            }
         }
      }


      // hide the columns of unchecked FITS keywords
      this.showOrHideFITSkey();

      // Keep the File name column reasonably sized
      if (longestFileName.length > 80) {
         longestFileName=longestFileName.substr(0,80);
      }
      this.filesTreeBox.setColumnWidth(0,this.font.width(longestFileName + "MMMM") );

   }


   // -- enable/disable operation buttons depending on context
   this.updateButtonState = function()
   {
      var enabled = this.dialog.engine.canDoOperation();
      this.dialog.move_Button.enabled = enabled;
      this.dialog.copy_Button.enabled = enabled;
      this.dialog.loadSave_Button.enabled = enabled;
#ifdef IMPLEMENTS_FITS_EXPORT
      this.dialog.txt_Button.enabled = enabled;
#endif
   }


   // -- Add a list of files to the TreeBox, refresh the TreeBox
   this.addFilesAction = function (fileNames)
   {
#ifdef DEBUG
      debug("addFilesAction - adding " + fileNames.length + " files");
#endif
      this.engine.addFiles(fileNames);

      this.rebuildFilesTreeBox();
      //this.adjustToContents();
      this.dialog.updateButtonState();
      this.dialog.updateTotal();

      this.refreshTargetFiles();
#ifdef DEBUG
      debug("addFilesAction - added " + fileNames.length + " files");
#endif
   }


   // -- Update the input file list total after each add/remove/check toggle
   this.updateTotal = function() {
      // Should be same as this.engine.inputFiles.length
      var countTotal = this.filesTreeBox.numberOfChildren;
      var countChecked = 0;
      for (var iTreeBox = 0; iTreeBox < this.filesTreeBox.numberOfChildren; ++iTreeBox) {
         if ( this.filesTreeBox.child(iTreeBox).checked ) {
            countChecked += 1;
         }
      }
      var countText;
      if (countTotal === 0 && countChecked ===0) {
        countText = "No file";
      } else if (countChecked===0) {
         countText = "No checked file / " + countTotal + " file" + (countTotal>1 ? "s" : "");
      } else {
         countText = "" + countChecked + " checked file" +  (countChecked>1 ? "s" : "") + " / " + countTotal + " file" + (countTotal>1 ? "s" : "");
      }
      this.inputSummaryLabel.text = countText;
      this.barInput.setCollapsedTitle("Input - " + countText);
   }


   // -- Return an array of the files that with chekd box ticked
   this.makeListOfCheckedFiles = function() {
      var listOfFiles = [];

      for (var iTreeBox = 0; iTreeBox < this.filesTreeBox.numberOfChildren; ++iTreeBox) {

         if ( this.filesTreeBox.child(iTreeBox).checked ) {
            // Select name in tree box, find corresponding file in inputFiles
//            var nameInTreeBox = this.filesTreeBox.child(iTreeBox).text(0);
            var nameInTreeBox = this.filesTreeBox.child(iTreeBox).fullFileName;
            listOfFiles.push(nameInTreeBox);
         }
      }

      return listOfFiles;
   }


   // -- Update the output operations indications
   this.refreshTargetFiles = function() {

#ifdef DEBUG
      debug("refreshTargetFiles() called");
#endif
      var listOfFiles = this.makeListOfCheckedFiles();

      this.engine.buildTargetFiles(listOfFiles);


      // List of text accumulating the transformation rules for display
      var listsOfTransforms = this.engine.makeListsOfTransforms();

      this.transform_TreeBox.clear();
      var firstNode = null;
      for (var i=0; i<listsOfTransforms.inputFiles.length; i++) {
         var inputFile = listsOfTransforms.inputFiles[i];
         var inputFileIndex =  listsOfTransforms.inputFileIndices[i];
         //Console.writeln(i + " " + inputFileIndex + " " + inputFile);
         var targetFile =  listsOfTransforms.targetFiles[i];
         var errorMessage = listsOfTransforms.errorMessages[i];


         var sourceNode = new TreeBoxNode( this.transform_TreeBox );
         if (i===0) { firstNode = sourceNode}
         sourceNode.setText(0,"File ".concat(inputFile));
         sourceNode.inputFileIndex = inputFileIndex;

         var resultNode = new TreeBoxNode( this.transform_TreeBox );
         if (targetFile) {
            resultNode.setText(0,"  to .../".concat(targetFile));
         } else {
            resultNode.setText(0,"     Error ".concat(errorMessage));
            resultNode.setTextColor(0,0x00FF0000);
         }
         resultNode.inputFileIndex = inputFileIndex;
      }
      if (firstNode) {
         // Scroll to beginning
         this.transform_TreeBox.currentNode = firstNode;
         // But do not select
         firstNode.selected = false;
      }

      var nmbFilesExamined = this.engine.targetFiles.length;

      var barResultTitle = "";
      if (nmbFilesExamined === 0) {
          barResultTitle += "No operation";
      } else {
         barResultTitle += "" + nmbFilesExamined + " files checked" ;
         if (this.engine.nmbFilesTransformed >0) {
            barResultTitle += ", " + this.engine.nmbFilesTransformed + " to copy/move";
         }

         //      this.engine.nmbFilesSkipped;
         if (this.engine.nmbFilesInError >0) {
            barResultTitle += ", " + this.engine.nmbFilesInError + " IN ERROR";
         }
      }
      this.outputSummaryLabel.text = barResultTitle;
      this.barResult.setCollapsedTitle("Resulting operations - " + barResultTitle);


    }


    // -- Support for refresh and move, remove all input files that are not present anymore in the file system
    this.removeDeletedFiles = function() {
      for ( var iTreeBox = this.dialog.filesTreeBox.numberOfChildren; --iTreeBox >= 0; ) {

         //var nameInTreeBox = this.dialog.filesTreeBox.child(iTreeBox).text(0);
         var nameInTreeBox = this.dialog.filesTreeBox.child(iTreeBox).fullFileName;
         if (!File.exists(nameInTreeBox)) {
#ifdef DEBUG
            debug("File '" + nameInTreeBox + "' removed from input list as not present on file system anymore.");
#endif
            this.dialog.engine.removeFiles(nameInTreeBox);
            this.dialog.filesTreeBox.remove( iTreeBox );
         }

         this.dialog.updateTotal();
         this.dialog.updateButtonState();
         // Caller must refresh the generated files
       }
    }

}
MainDialog.prototype = new Dialog;




// ========================================================================================================================
// Documentation dialog
// ========================================================================================================================

// See http://pixinsight.com/developer/pcl/doc/20120901/html/classpcl_1_1Console.html
// for formatting instructions

function HelpDialog( parentDialog, engine ) {
   this.__base__ = Dialog;
   this.__base__();

   this.windowTitle = "FITSFileManager Help";

   this.helpBox = new TextBox( this );
   this.helpBox.readOnly = true;
   this.helpBox.text = Text.H.HELP_TEXT;
   this.helpBox.setMinSize( 800, 400 );
   this.helpBox.caretPosition = 0;

   this.sizer = new HorizontalSizer;
   this.sizer.margin = 6;
   this.sizer.add( this.helpBox );
   this.setVariableSize();
   this.adjustToContents();
}

HelpDialog.prototype = new Dialog;


// ========================================================================================================================
// Completion dialog
// ========================================================================================================================


function CompletionDialog( parentDialog, engine ) {
   this.__base__ = Dialog;
   this.__base__();

   this.windowTitle = Text.T.COMPLETION_TITLE;

   this.resultBox = new TextBox( this );
   this.resultBox.readOnly = true;
   this.resultBox.text = "TEXT NOT INITIALIZED";
   this.resultBox.setMinSize( 600, 200 );
   this.resultBox.caretPosition = 0;

   this.continue_Button = new PushButton( this );
   this.continue_Button.text = Text.T.COMPLETION_CONTINUE_BUTTON_TEXT;
   this.continue_Button.toolTip = Text.H.COMPLETION_CONTINUE_BUTTON_TOOLTIP;
   this.continue_Button.enabled = true;
   this.continue_Button.onClick = function() {
      this.dialog.done(CompletionDialog_doneContinue);
   }
   this.keep_Button = new PushButton( this );
   this.keep_Button.text = Text.T.COMPLETION_KEEP_BUTTON_TEXT;
   this.keep_Button.toolTip = Text.H.COMPLETION_KEEP_BUTTON_TOOLTIP;
   this.keep_Button.enabled = true;
   this.keep_Button.onClick = function() {
      this.dialog.done(CompletionDialog_doneKeep);
   }
   this.remove_Button = new PushButton( this );
   this.remove_Button.text = Text.T.COMPLETION_REMOVE_BUTTON_TEXT;
   this.remove_Button.toolTip = Text.H.COMPLETION_REMOVE_BUTTON_TOOLTIP;
   this.remove_Button.enabled = true;
   this.remove_Button.onClick = function() {
      this.dialog.done(CompletionDialog_doneRemove);
   }
   this.leave_Button = new PushButton( this );
   this.leave_Button.text = Text.T.COMPLETION_LEAVE_BUTTON_TEXT;
   this.leave_Button.toolTip = Text.H.COMPLETION_LEAVE_BUTTON_TOOLTIP;
   this.leave_Button.enabled = true;
   this.leave_Button.onClick = function() {
      this.dialog.done(CompletionDialog_doneLeave);
   }



   // Sizer for Operation List and Actions section

   this.buttonSizer = new HorizontalSizer;
   this.buttonSizer.spacing = 2;
   this.buttonSizer.add( this.continue_Button);
   this.buttonSizer.add( this.keep_Button);
   this.buttonSizer.add( this.remove_Button);
   this.buttonSizer.add( this.leave_Button);
   this.buttonSizer.addStretch();


   this.sizer = new VerticalSizer;
   this.sizer.margin = 6;
   this.sizer.add( this.resultBox );
   this.sizer.add( this.buttonSizer );
   this.setVariableSize();
   this.adjustToContents();

   this.setResultText = function (text) {
      this.resultBox.text = text;
   }
   this.setResultModeMove = function() {
      this.continue_Button.enabled = true;
      this.keep_Button.enabled = false;
      this.remove_Button.enabled = false;
   }
   this.setResultModeCopy = function () {
      this.continue_Button.enabled = false;
      this.keep_Button.enabled = true;
      this.remove_Button.enabled = true;
   }
}

CompletionDialog.prototype = new Dialog;





// ========================================================================================================================
// FITS and synthetic keys dialog
// ========================================================================================================================
// Present a dialog with:
//   A selection of the files (drop down)
//   A list of  FITS keywords (selection box, keyword, value)  of the selected file, as a TreeBox
// ---------------------------------------------------------------------------------------------------------
function FITSKeysDialog( parentDialog, engine) {
   this.__base__ = Dialog;
   this.__base__();
   this.windowTitle = "Select FITS keywords to show in main window";

   // ComboBox to select the file to display values
   this.file_ComboBox = new ComboBox( this );

   // A file was selected (also called to initialize)
   this.file_ComboBox.onItemSelected = function( indexInComboBox ) {
      var fileName = this.itemText(indexInComboBox);
      var indexInFiles = engine.inputFiles.indexOf(fileName);
#ifdef DEBUG
      debug("FITSKeysDialog: file_ComboBox: onItemSelected(" + indexInComboBox + ") -  indexInFiles = " + indexInFiles + ", keywordsSet.size() = " + engine.keywordsSet.size());
#endif

     this.dialog.populate(indexInFiles);

   }

   // TreeBox to display list of FITS keywords
   this.keyword_TreeBox = new TreeBox( this );
   this.keyword_TreeBox.toolTip = "Synthetic and value keywords of the image selected in the drop box at top,\n" +
      "Tick the check mark to include the keyword value in the input file table\n"+
      "The red color indicates that the keyword is not in the selected image, but appears in some other loaded images,\n"+
      "or that its value is space (it will be considered missing).";
   this.keyword_TreeBox.rootDecoration = false;
   this.keyword_TreeBox.numberOfColumns = 3;
   this.keyword_TreeBox.setHeaderText(0, "name");
   this.keyword_TreeBox.setHeaderText(1, "value");
   this.keyword_TreeBox.setHeaderText(2, "type");
   this.keyword_TreeBox.setHeaderText(3, "comment");
   this.keyword_TreeBox.setColumnWidth(0,150);
   this.keyword_TreeBox.setColumnWidth(1,200);
   this.keyword_TreeBox.setColumnWidth(2,50);
   this.keyword_TreeBox.setColumnWidth(3,600);


   // Buttons
   this.cancel_Button = new PushButton( this );
   this.cancel_Button.text = "Cancel";
   this.cancel_Button.enabled = true;
   this.cancel_Button.onClick = function() {
      this.dialog.cancel();
   }
   this.ok_Button = new PushButton( this );
   this.ok_Button.text = "OK";
   this.ok_Button.enabled = true;
   this.ok_Button.onClick = function() {
#ifdef DEBUG
      debug("FITSKeysDialog: ok_Button: onClick - save check/uncheck status in shownFITSKeyNames ");
#endif
      var fitsRoootNode = this.parent.keyword_TreeBox.child(1);
      var allFITSKeyNames = engine.keywordsSet.allValueKeywordNameList;
      // Clear and rebuild list of keywords to show
      engine.shownFITSKeyNames = {};
      for (var i = 0; i< allFITSKeyNames.length; i++) {
         var checked = fitsRoootNode.child(i).checked; // List and rows are in same order
         var name = allFITSKeyNames[i];
         if (checked) {
             engine.shownFITSKeyNames[name] = true;
         }
      }
      var variableRoootNode = this.parent.keyword_TreeBox.child(0);
      // TODO Delegate operation to engine
      engine.shownSyntheticKeyNames = {};
      for (var i = 0; i< ffM_Configuration.syntheticVariableNames.length; i++) {
         var checked = variableRoootNode.child(i).checked; // List and rows are in same order
         var name = ffM_Configuration.syntheticVariableNames[i];
         if (checked) {
             engine.shownSyntheticKeyNames[name] = true;
         }
      }
      this.dialog.ok();
   }

   this.buttonsSizer = new HorizontalSizer;
   this.buttonsSizer.spacing = 2;
   this.buttonsSizer.addStretch();
   this.buttonsSizer.add( this.cancel_Button);
   this.buttonsSizer.add( this.ok_Button);


   // Assemble FITS keyword Dialog
   this.sizer = new VerticalSizer;
   this.sizer.margin = 4;
   this.sizer.spacing = 4;
   this.sizer.add( this.file_ComboBox );
   this.sizer.add( this.keyword_TreeBox );
   this.sizer.add(this.buttonsSizer);
   this.adjustToContents();

   // ------------------------------------------------------------
   // Recreate the content when the dialog is showns
   // This method recreate the columnsm key names, populate the list of file and call initializer for data
   this.onShow = function()
   {
#ifdef DEBUG
      debug("FITSKeysDialog: file_ComboBox: onShow()");
#endif
      // -- Locate dialog
      var p = new Point( parentDialog.position );
      p.moveBy( 16,16 );
      this.position = p;

      // -- Update the DropDown box - Fill list of files from parent list of files, to be in the same order
      // Note - this requires finding the proper index in inputFile[] when a selection is done, as order may be different
      this.file_ComboBox.clear();
      for (i = 0; i<parentDialog.filesTreeBox.numberOfChildren; i++) {
//         this.file_ComboBox.addItem(parentDialog.filesTreeBox.child(i).text(0));
        this.file_ComboBox.addItem(parentDialog.filesTreeBox.child(i).fullFileName);
      }
////    OLD: This is filling the list in the loaded file order
//      for (i = 0; i< engine.inputFiles.length; i++) {
//         this.file_ComboBox.addItem(engine.inputFiles[i]);
//      }

      var selectedFileIndex = 0;
      var fileIndex = 0;
      if (parentDialog.filesTreeBox.selectedNodes.length >0) {
         // Show first selected node
         var firstSelectedNode = parentDialog.filesTreeBox.selectedNodes[0];
//         var fileName = firstSelectedNode.text(0);
         var fileName = firstSelectedNode.fullFileName;
         fileIndex = engine.inputFiles.indexOf(fileName);
         selectedFileIndex = parentDialog.filesTreeBox.childIndex(firstSelectedNode);
#ifdef DEBUG
         debug("FITSKeysDialog: file_ComboBox: onShow - fileName="+fileName+", fileIndex=" + fileIndex+ ", selectedFileIndex=" + selectedFileIndex);
#endif
      }
      // Select the file in the combo box
      this.file_ComboBox.currentItem = selectedFileIndex;


      // -- Rebuild the 3 lists of keywords in the tree box

      this.keyword_TreeBox.clear();

      // Create list of synthetic keywords as a fist subtree
      var synthRootNode = new TreeBoxNode(this.keyword_TreeBox);
      synthRootNode.expanded = true;
      synthRootNode.setText(0,"Synthetic keywords");


      // Fill the name column from the list of all synthetic keywords names
      for (var i =0; i<ffM_Configuration.syntheticVariableNames.length; i++) {
         var node = new TreeBoxNode(synthRootNode);
         node.setText( 0, ffM_Configuration.syntheticVariableNames[i] );
         node.checked = engine.isVariableVisible(ffM_Configuration.syntheticVariableNames[i]);;
      }

      // Create list of FITS keywords used as variables as a second subtree
      var fitsVarRootNode = new TreeBoxNode(this.keyword_TreeBox);
      fitsVarRootNode.expanded = true;
      fitsVarRootNode.setText(0,"Valued FITS keywords");


      // Fill the name columns from the from allFITSKeyNames (accumulated names of all keywords)
      var allFITSKeyNames = engine.keywordsSet.allValueKeywordNameList;
      for (var i =0; i<allFITSKeyNames.length; i++) {
         var node = new TreeBoxNode(fitsVarRootNode);
         node.setText( 0, allFITSKeyNames[i] );
         node.checked = engine.shownFITSKeyNames.hasOwnProperty(allFITSKeyNames[i]);
      }


      // Populate with default file
      this.populate(fileIndex);

      // Ensure that something reasonable is visible so the user can
      // figure out what to do, but try to make it large to be convenient,
      // hopefully this should work on small screens.
      // adjustToContents result in too small a window.
      this.setMinSize(FitsKeysDialogMinimumWidth,FitsKeysDialogMinimumHeight);
      this.width = FitsKeysDialogPreferedWidth;
      this.height = FitsKeysDialogPreferedHeight;
   };

   this.getTypeString = function (fitsKey) {
      if (fitsKey.isBoolean) {
         return "bool";
      } else if (fitsKey.isNumeric) {
         return "num";
      } else if (fitsKey.isString) {
         return "str";
      } else if (fitsKey.isNull) {
         return "null";
      }
      return "";

   }

   // Populate from information of inputFile[index] and inputFITSKeywords[index]
   this.populate = function (index) {

#ifdef DEBUG
      debug("FITSKeysDialog: file_ComboBox: populate() - index=" + index + " for file " + engine.inputFiles[index] );
#endif

      // Update the values of the synthethic keywords from a predefined list and engine values
      var synthRootNode = this.keyword_TreeBox.child(0);

      for (var i =0; i<ffM_Configuration.syntheticVariableNames.length; i++) {
         var keyName = ffM_Configuration.syntheticVariableNames[i];
         var variables = engine.inputVariables[index];
         var variable = variables[keyName];
         if (variable !== null) {
            synthRootNode.child(i).setTextColor(0,0x00000000);
            // TODO ADHOC TEMPORARY TEST
            if (typeof variable !== 'string') {
               variable = '?[' + typeof variable +  ']';
            }

            synthRootNode.child(i).setText(1,variable);
            synthRootNode.child(i).setText(2,'');
            synthRootNode.child(i).setText(3, ffM_Configuration.syntheticVariableComments[i]);
         } else {
            synthRootNode.child(i).setTextColor(0,0x00FF0000);
            synthRootNode.child(i).setText(1,'');
            synthRootNode.child(i).setText(2,'');
            synthRootNode.child(i).setText(3,'');
         }
      }

      // Update FITS key words values from engine information
      var fitsVarRootNode = this.keyword_TreeBox.child(1);

      var allFITSKeyNames = engine.keywordsSet.allValueKeywordNameList;
      var imageKeywords = engine.inputFITSKeywords[index];
      for (var i = 0; i<allFITSKeyNames.length; i++) {
         var keyName = allFITSKeyNames[i];
         var keyValue = imageKeywords.getValueKeyword(keyName);
#ifdef DEBUG_FITS
         debug("FITSKeysDialog: populate(): FITS  keyName=" + keyName + ",  keyValue=" + keyValue );
#endif
         if (keyValue === null) {
            // No value at all
            fitsVarRootNode.child(i).setTextColor(0,0x00FF0000); // Red
            fitsVarRootNode.child(i).setText(1,'');
            fitsVarRootNode.child(i).setText(2,'');
            fitsVarRootNode.child(i).setText(3,'');
         } else {
            var nullOrValue = ffM_variables.filterFITSValue(keyValue.value);
            if (nullOrValue === null) {
               // All spaces value
               fitsVarRootNode.child(i).setTextColor(0,0x00FF0000);  // Red
               fitsVarRootNode.child(i).setText(1,keyValue.value);
               fitsVarRootNode.child(i).setText(2,this.getTypeString(keyValue));
               fitsVarRootNode.child(i).setText(3,keyValue.comment);
            } else {
               // Non empty string, number or boolean
               fitsVarRootNode.child(i).setTextColor(0,0x00000000);
               fitsVarRootNode.child(i).setText(1,keyValue.value);
               fitsVarRootNode.child(i).setText(2,this.getTypeString(keyValue));
               fitsVarRootNode.child(i).setText(3,keyValue.comment);
            }
         }
      }

   }

};


FITSKeysDialog.prototype = new Dialog;



//==== end of include ============================================




// -------------------------------------------------------------------------------------
function ffM_main() {
#ifdef DEBUG
   Console.show();
#endif
   var guiParameters = new FFM_GUIParameters();
   guiParameters.loadSettings();

   var engine = new FFM_Engine(guiParameters);
   engine.setConfiguration(ffM_Configuration.createWorkingConfiguration());

   var dialog = new MainDialog(engine, guiParameters);
   dialog.execute();
   guiParameters.saveSettings();
}

ffM_main();

